#!/bin/bash
# This script fixes CPiag Scheduler and registry issues

. /opt/CPshared/5.0/tmp/.CPprofile.sh
. `dirname $0`"/DDR_shared_functions"

GAIA_VERSION=""
CPDIAG=$DIAGDIR/bin/cpdiag
GREP=/bin/grep
LS=/bin/ls
CPD_SCHED_BIN=$CPDIR/bin/cpd_sched_config
DIAGNOSTICS_FILE=/tmp/cpdiag_to_cpd_sched.csv
# Diagnostics file will contain 2 columns: timestamp of the script run and the script result
# Values of the script result are
CPDIAG_WAS_IN_CPD_SCHED=51
CPDIAG_ADDED_TO_CPD_SCHED=52
CPDIAG_WAS_SET_ON_THE_REGISTRY=57
FINISHED_UPDATING_THE_REGISTRY=58
MISSING_CPDIAG_FOLDER=59
FAILED_TO_ADD_CPDIAG_TO_CPD_SCHED=53
FAILED_TO_FIND_CPDIAG_BINARY=54
SCRIPT_NOT_MATCHING_REQUIRED_OS=55
RPM_IS_NOT_INSTALLED=56
IS_RUN=/var/log/CPDIAG_FIXER_TOKEN
LOG_FILE=/var/log/cpdiag_to_cpd_sched.log

if [[ -f $IS_RUN ]]; then
	endCheck 0 "Fail"
fi

function lock() {
	touch $IS_RUN 2>&1 >> $LOG_FILE
}

if [ -f $DIAGNOSTICS_FILE ];
then
	echo "$(date) : $DIAGNOSTICS_FILE already exists" >> $LOG_FILE
else
	echo "$(date) : Create $DIAGNOSTICS_FILE" >> $LOG_FILE
	# Add column names to the diagnostics csv file
	echo "script_run_ts,script_result" >> $DIAGNOSTICS_FILE
fi

ret_val=0

echo "$(date) : cpdiag to cpd_sched script started" >> $LOG_FILE

if cpprod_util CPPROD_GetLastMinorVersion 2>&1 >> $LOG_FILE; then
	TMP_CURRENT_VERSION=`cpprod_util CPPROD_GetLastMinorVersion | tr -d " " 2>/dev/null`
	if [[ "$TMP_CURRENT_VERSION" == "<nil>" || "$TMP_CURRENT_VERSION" == "0" ]]; then
		CURRENT_VERSION=`cpprod_util CPPROD_GetVerText CPshared | tr -d " " 2>/dev/null`
	else 
		CURRENT_VERSION=$TMP_CURRENT_VERSION
	fi
else
	CURRENT_VERSION=`cpprod_util CPPROD_GetVerText CPshared | tr -d " " 2>/dev/null`
fi

if [[ "$CURRENT_VERSION" == "R80.10" ]]; then 
	CURRENT_VERSION_TEXT=R80
else
	CURRENT_VERSION_TEXT=$CURRENT_VERSION
fi

if [[ "R80.10" == "${CURRENT_VERSION}" ]]; then 
	echo "$(date) : Version required match this version" >> $LOG_FILE
else 
	echo "$(date) : Version required doesn't match this version" >> $LOG_FILE
	echo "$(date +%s),$SCRIPT_NOT_MATCHING_REQUIRED_OS" >> $DIAGNOSTICS_FILE
	lock
	endCheck 0 "Fail"
fi

if rpm -qa | grep CPdiag 2>&1 >> $LOG_FILE; then 
	echo "$(date) : CPDiag RPM is installed" >> $LOG_FILE
else 
	echo "$(date) : CPDiag RPM is not installed" >> $LOG_FILE
	echo "$(date +%s),$RPM_IS_NOT_INSTALLED" >> $DIAGNOSTICS_FILE
	lock
	endCheck 0 "Fail"
fi

echo "$(date) : Find cpdiag installation folder under /opt" >> $LOG_FILE

LS_PRINT=$($LS /opt | $GREP CPdiag 2> /dev/null)
if [ ! -z "$LS_PRINT" ];
then
	echo "$(date) : Found cpdiag installation folder" >> $LOG_FILE
	CPDIAG=/opt/$LS_PRINT/bin/cpdiag
	echo "$(date) : Installation folder: ${CPDIAG}" >> $LOG_FILE
else
	echo "$(date) : Did not find the cpdiag installation folder" >> $LOG_FILE
	echo "$(date +%s),$MISSING_CPDIAG_FOLDER" >> $DIAGNOSTICS_FILE
	lock
	endCheck 0 "Fail"
fi

if [ -f $CPDIAG ];
then
	echo "$(date) : Found cpdiag binary" >> $LOG_FILE
	
	ret_val=1
	
	CPD_PRINT=""
	for i in 1 2 3
	do
		CPD_PRINT=$($CPD_SCHED_BIN print | $GREP CPDiag 2> /dev/null)
		if [ -z "$CPD_PRINT" ];
		then
			# Install CPDiag as a service, i.e cpdiag will be added to cpd_sched
			echo "$(date) : cpdiag is not registered in cpd_sched, register it" >> $LOG_FILE
			$CPDIAG --install 2>&1 >> $LOG_FILE
		else
			echo "$(date) : cpdiag is registered in cpd_sched" >> $LOG_FILE
			if [[ "$i" -eq 1 ]];
			then
				echo "$(date) : cpdiag was already registered in cpd_sched" >> $LOG_FILE
				echo "$(date +%s),$CPDIAG_WAS_IN_CPD_SCHED" >> $DIAGNOSTICS_FILE
			else
				echo "$(date) : cpdiag was added to the cpd_sched" >> $LOG_FILE
				echo "$(date +%s),$CPDIAG_ADDED_TO_CPD_SCHED" >> $DIAGNOSTICS_FILE
			fi
			
			ret_val=0
			break
		fi
	done
	
	# Don't re-run script if a successful operation occured or too much operation occured
	OPERATION_NUMBER=`cat ${DIAGNOSTICS_FILE} | wc -l 2>/dev/null`
	OPERATION_NUMBER=$((OPERATION_NUMBER-1))
	if [[ ! -z "$CPD_PRINT" || $OPERATION_NUMBER -gt 20 ]]; then 
		lock
	fi
	
	if [[ "$ret_val" -ne 0 ]];
	then
		echo "$(date) : Failed addeding CPDIAG to the cpd_sched" >> $LOG_FILE
		echo "$(date +%s),$FAILED_TO_ADD_CPDIAG_TO_CPD_SCHED" >> $DIAGNOSTICS_FILE
	fi
	
	echo "$(date) : Try to fix the registry" >> $LOG_FILE
	PRODUCTNAME=DIAG
	VER=6.0
	if cpprod_util CPPROD_GetValue Products $PRODUCTNAME 0 2>&1 >> $LOG_FILE; then
		echo "$(date) : CPDiag already set on the registry" >> $LOG_FILE
		echo "$(date +%s),$CPDIAG_WAS_SET_ON_THE_REGISTRY" >> $DIAGNOSTICS_FILE
	else 
		cpprod_util CPPROD_SetValue Products $PRODUCTNAME 1 1 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue $PRODUCTNAME CurrentVersion 1 ${VER} 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER} CurrentLabel 1 `cpprod_util CPPROD_GetVerText CPshared | tr -d " " 2>/dev/null` 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER} CurrentSP 1 `cpprod_util CPPROD_GetCurrentServicePack CPshared | tr -d " " 2>/dev/null` 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER} PRODUCTDIR 1 /opt/$LS_PRINT 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER} PRODDIR 1 /opt/$LS_PRINT 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER} AutoStart 1 1 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER} IsConfigured 1 1 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER} ProdActive 1 0 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER}\\SP`cpprod_util CPPROD_GetCurrentServicePack CPshared | tr -d " " 2>/dev/null` CurrentMSP 1 `cpprod_util CPPROD_GetCurrentMSP CPshared | tr -d " " 2>/dev/null` 1 2>&1 >> $LOG_FILE
		cpprod_util CPPROD_SetValue ${PRODUCTNAME}\\${VER}\\SP`cpprod_util CPPROD_GetCurrentServicePack CPshared | tr -d " " 2>/dev/null`\\MSP`cpprod_util CPPROD_GetCurrentMSP CPshared | tr -d " " 2>/dev/null` SilentUninstall 1 "rpm -e ${LS_PRINT}"  1 2>&1 >> $LOG_FILE
		echo "$(date) : Finished updatding the registry." >> $LOG_FILE 
		echo "$(date +%s),$FINISHED_UPDATING_THE_REGISTRY" >> $DIAGNOSTICS_FILE
	fi
	# Update DIAG if not matching folder
	if [[ $DIAGDIR != /opt/$LS_PRINT ]]; then
		echo "$(date) : Update enviroment" >> $LOG_FILE
		sed -i '/^\_cpprof\_add DIAGDIR \"\/opt\/CPdiag\" 1 1$/d' /opt/CPshared/5.0/tmp/.CPprofile.sh 2>&1 >> $LOG_FILE
		sed -i '/^\_cpprof\_add DIAGDIR \"\/opt\/CPdiag-${CURRENT_VERSION_TEXT}\" 1 1$/d' /opt/CPshared/5.0/tmp/.CPprofile.sh 2>&1 >> $LOG_FILE
		/opt/$LS_PRINT/CPinstall/XInstall AddToCPprofile DIAGDIR /opt/$LS_PRINT 1 1 2>&1 >> $LOG_FILE
	fi
else
	echo "$(date) : cpdiag binary is not present" >> $LOG_FILE
	echo "$(date +%s),$FAILED_TO_FIND_CPDIAG_BINARY" >> $DIAGNOSTICS_FILE
fi

echo "$(date) : cpdiag to cpd_sched script exits with rc=$ret_val" >> $LOG_FILE
endCheck 0 "Fail"