#!/bin/bash
# This script fixes CPiag Scheduler and registry issues

. /opt/CPshared/5.0/tmp/.CPprofile.sh
. `dirname $0`"/DDR_shared_functions"

GAIA_VERSION=""
CPDIAG=$DIAGDIR/bin/cpdiag
GREP=/bin/grep
LS=/bin/ls
TR=/usr/bin/tr
CPD_SCHED_BIN=$CPDIR/bin/cpd_sched_config
DIAGNOSTICS_FILE=/tmp/cpdiag_to_cpd_sched.csv
# Diagnostics file will contain 2 columns: timestamp of the script run and the script result
# Values of the script result are
MISSING_CPDIAG_FOLDER=68
MISSING_CPDIAG_EXECUTABLE=66
RPM_IS_NOT_INSTALLED=67
PATH_DEFINED_IN_SCHED_DOESNT_MATCH_FILE_SYSTEM=69
PATH_DEFINED_IN_SCHED_MATCH_FILE_SYSTEM=70
IS_RUN=/var/log/CPDIAG_FIXER_TOKEN_3
LOG_FILE=/var/log/cpdiag_fix_not_matching_cpsched.log

if [[ -f $IS_RUN ]]; then
	endCheck 0 "Fail"
fi

function lock() {
	touch $IS_RUN 2>&1 >> $LOG_FILE
}

function end_and_lock() {
	lock
	endCheck 0 "Fail"
}


echo "$(date) : scheduler validator script started" >> $LOG_FILE

################################################################################
# Define the CSV file for collecting the script status by CPDiag               #

if [ -f $DIAGNOSTICS_FILE ];
then
	echo "$(date) : $DIAGNOSTICS_FILE already exists" >> $LOG_FILE
else
	echo "$(date) : Create $DIAGNOSTICS_FILE" >> $LOG_FILE
	# Add column names to the diagnostics csv file
	echo "script_run_ts,script_result" >> $DIAGNOSTICS_FILE
fi

##############################################################
# Check cpdiag is Installed on the system

if rpm -qa | grep CPdiag 2>&1 >> $LOG_FILE; then 
	echo "$(date) : CPDiag RPM is installed" >> $LOG_FILE
else 
	echo "$(date) : CPDiag RPM is not installed" >> $LOG_FILE
	echo "$(date +%s),$RPM_IS_NOT_INSTALLED" >> $DIAGNOSTICS_FILE
	end_and_lock
fi

##############################################################
# Find the install folder of CPDiag

echo "$(date) : Find cpdiag installation folder under /opt" >> $LOG_FILE

LS_PRINT=$($LS /opt | $GREP CPdiag 2> /dev/null)
if [ ! -z "$LS_PRINT" ];
then
	echo "$(date) : Found cpdiag installation folder" >> $LOG_FILE
	CPDIAG=/opt/$LS_PRINT/bin/cpdiag
	echo "$(date) : Installation folder: ${CPDIAG}" >> $LOG_FILE
else
	echo "$(date) : Did not find the cpdiag installation folder" >> $LOG_FILE
	echo "$(date +%s),$MISSING_CPDIAG_FOLDER" >> $DIAGNOSTICS_FILE
	end_and_lock
fi

if [ ! -f $CPDIAG ]; then
	echo "$(date) : Did not find the cpdiag executable." >> $LOG_FILE
	echo "$(date +%s),$MISSING_CPDIAG_EXECUTABLE" >> $DIAGNOSTICS_FILE
	end_and_lock
fi

#####################################################################
# Verify 2 Conditions:
# 1) The CPDiag scheduler task has a valid command on the registry
# 2) Use grep on the output of the cpd_sched_config to see the task is defined live
# If both are invalid - reset the scheduler definition of CPDIAG

echo "$(date) : Verifying Scheduler is looking ok." >> $LOG_FILE

COMMAND_IN_SCHEDULER_REG_DEFINE=$($CPDIR/bin/cpprod_util CPPROD_GetValue CPshared//$($CPDIR/bin/cpprod_util CPPROD_GetCurrentVersion CPshared 2>/dev/null | $TR -d "[[:space:]]")//reserved//CPD_Scheduler//CPDiag Cmd 0 2>/dev/null | $TR -d "[[:space:]]")
CPDIAG_FILESYS_PATH_IN_SCHED_OUTPUT=$($CPDIR/bin/cpd_sched_config print 2>/dev/null | $GREP "Command: ${CPDIAG}")
if [[ "X${COMMAND_IN_SCHEDULER_REG_DEFINE}" == "X${CPDIAG}" ]] && [[ ! -z $CPDIAG_FILESYS_PATH_IN_SCHED_OUTPUT ]]; then
		echo "$(date) : Scheduler is matching to version." >> $LOG_FILE
		echo "$(date +%s),$PATH_DEFINED_IN_SCHED_MATCH_FILE_SYSTEM" >> $DIAGNOSTICS_FILE
		end_and_lock
fi

lock

echo "$(date) : Fixing the Mismatch." >> $LOG_FILE
echo "$(date +%s),$PATH_DEFINED_IN_SCHED_DOESNT_MATCH_FILE_SYSTEM" >> $DIAGNOSTICS_FILE
$CPDIR/bin/cpd_sched_config add CPDiag -c "${CPDIAG}" -v "--periodic" -e 86400 -r -s 2>&1 >> $LOG_FILE

#####################################################################

endCheck 0 "Fail"