#!/bin/bash
. `dirname $0`"/DDR_shared_functions"

# THIS SCRIPT IS FOR R80.20X ONLY !!
# FOR OTHER VERSIONS  TEST / MODIFY THE SCRIPT



CPPROD_UTIL=$CPDIR/bin/cpprod_util
ECHO="/bin/echo"
CAT=/bin/cat
PS=/bin/ps
LOCAL_LOG_FILE=/opt/CPInstLog/collect_mgmtpuv_output.log
LOCAL_PUV_OUTPUT=/var/log/puv_output.elg
LOCAL_PUV_EXIT_CODE=/var/log/puv_exit_code.elg
#parameter received is 'target_mgmt_version'
parseArgs "$@"

internal_failure_msg="Failed to collect verification parameters. Please try again."

#dump all stderr to log file
exec 2>$LOCAL_LOG_FILE

getFatherProcessID (){    
		fatherPpidReturn=`ps -p $1 -o ppid=`
		if [ $? -ne 0 -o "X$fatherPpidReturn" = "X" ]; then
			$ECHO "Failed to get process $1 parent PID..." >> $LOCAL_LOG_FILE 
			$ECHO "verify_output=\"$internal_failure_msg\""
			$ECHO "fatherPpidReturn=$fatherPpidReturn" >> $LOCAL_LOG_FILE
			exit 0
		fi
}
#initialization of global varibale fatherPpidReturn
fatherPpidReturn=-1

#run mgmt puv only on management machine
#Don't run puv in case of automatic verify of DA (SKIP_PUV=1)
IsMgmt=`$CPPROD_UTIL FwIsFirewallMgmt`
if [ ${IsMgmt} != 1 -o "X$SKIP_PUV" = "X1" ]; then 
	$ECHO "verify_output=\"success\""
	exit 0
fi 

#make sure we receive the full output of ps commands
COLUMNS=300
$ECHO "Running $0 " `date` >> $LOCAL_LOG_FILE 
#find the path to conditions_set_is_ngm_upgrade.json


getFatherProcessID $$
ppid=$fatherPpidReturn

getFatherProcessID $ppid
gppid=$fatherPpidReturn

gppidArgs=$($PS -o args= -p $gppid)
if [ $? -ne 0 -o "X$gppidArgs" = "X" ]; then
	$ECHO "$PS -o args= -p $gppid" >> $LOCAL_LOG_FILE  
	$ECHO "verify_output=\"$internal_failure_msg\""
	exit 0
fi
metadataPathArg=$($ECHO $gppidArgs | $CUT -f3 -d\ | $SED "s/--path_to_metadata=//" )
if [ $? -ne 0 -o "X$metadataPathArg" = "X" ]; then
	$ECHO "$ECHO $gppidArgs | $CUT -f3 -d\ " >> $LOCAL_LOG_FILE 
	$ECHO "verify_output=\"$internal_failure_msg\""
	exit 0
fi
jsonDirPath=$(dirname "$metadataPathArg" )
if [ $? -ne 0 -o "X$jsonDirPath" = "X" ]; then
	$ECHO "dirname $metadataPathArg" >> $LOCAL_LOG_FILE  
	$ECHO "verify_output=\"$internal_failure_msg\""
	exit 0
fi

version=${target_mgmt_version:1:2}
#check the version of upgrade
if [ $version -lt "81" ]; then
#run DDR to determine whether we are using ngm upgrade or old upgrade
	hotfixID=`$ECHO ${jsonDirPath} | cut -d/ -f 6 | cut -d# -f 7 `
	/opt/DDR/bin/DDRService --action=collect_and_verify --path_to_metadata=${jsonDirPath}/conditions_set_is_ngm_upgrade.json --output_file=/tmp/DDR_validation_$$ --build_num=0 --hotfix_id=$hotfixID 2>/dev/null
is_ngm_upgrade=$?
else
	is_ngm_upgrade=0
fi

#run mgmt puv only for ngm_upgrade
if  [ $is_ngm_upgrade -ne 0 ]; then
		$ECHO "verify_output=\"success\""
		exit 0
fi

launcher_cmd=`$CPPROD_UTIL CPPROD_GetValue FW1 UpgradeLauncher 6.0`
if [[ "X$launcher_cmd" == "XFailed"* ]]; then
		$ECHO "verify_output=\"Failed to get UpgradeLauncher command from registry, please update the Upgrade Tools package. To download the package manually, refer to sk135172.\""
		exit 0
fi

verify_output=$(<$LOCAL_PUV_OUTPUT)
verify_result=$(<$LOCAL_PUV_EXIT_CODE)

  #if [ "$verify_output" == "" ];then
  if [ $verify_result -eq 0 ]; then
		$ECHO "verify_output=\"success\""
  else
		verify_output=$($ECHO "$verify_output"|sed ':a;N;$!ba;s/\n/\\n/g')
		$ECHO "verify_output='$verify_output'"
  fi


exit 0
