#!/bin/bash
. `dirname $0`"/DDR_shared_functions"
#machine collects comptability of grub and kernel files with secureboot keys
# return "success" in case verification is good.
LOCAL_LOG_FILE=/opt/CPInstLog/verify_secureboot.log

echo "===================================" >> $LOCAL_LOG_FILE
date >> $LOCAL_LOG_FILE

#Do we have uefi?
if [ ! -d "/sys/firmware/efi" ]; then
	echo "UEFI is not supported" >> $LOCAL_LOG_FILE
	echo "success"
	exit 0
fi

#Skip test only if secureboot is disabled
is_sb_enabled=`cp_secureboot --state`
if [ $? -eq 0 ] && [ ${is_sb_enabled} == "0" ] ; then
        echo "secureboot is disabled" >> $LOCAL_LOG_FILE
	echo "success"
	exit 0
fi

getFatherProcessID (){    
		fatherPpidReturn=`ps -p $1 -o ppid=`
		if [ $? -ne 0 -o "X$fatherPpidReturn" = "X" ]; then
			echo "Failed to get process $1 parent PID..." >> $LOCAL_LOG_FILE 
			echo "fatherPpidReturn=$fatherPpidReturn" >> $LOCAL_LOG_FILE
                        endCheck 0 "collect_err_msg='Could not verify secureboot comptability'"
		fi
}
#initialization of global varibale fatherPpidReturn
fatherPpidReturn=-1

getFatherProcessID $$
ppid=$fatherPpidReturn

getFatherProcessID $ppid
gppid=$fatherPpidReturn

gppidArgs=$(ps -o args= -p $gppid)
if [ $? -ne 0 -o "X$gppidArgs" = "X" ]; then
	echo "ps -o args= -p $gppid" >> $LOCAL_LOG_FILE  
        endCheck 0 "collect_err_msg='Could not verify secureboot comptability'"
fi

metadataPathArg=$(echo $gppidArgs | cut -f3 -d\ | sed "s/--path_to_metadata=//" )
if [ $? -ne 0 -o "X$metadataPathArg" = "X" ]; then
	echo "echo $gppidArgs | $CUT -f3 -d\ " >> $LOCAL_LOG_FILE 
        endCheck 0 "collect_err_msg='Could not verify secureboot comptability'"
fi

if [ -f ${metadataPathArg}/secureboot_grub_sigs.full ] ; then
	grub_sigs_file="${metadataPathArg}/secureboot_grub_sigs.full"
elif [ -f `dirname ${metadataPathArg}`/secureboot_grub_sigs.full ] ; then
	grub_sigs_file="`dirname ${metadataPathArg}`/secureboot_grub_sigs.full"
else
	echo "Grub sigs info file does not exist" >> $LOCAL_LOG_FILE
	endCheck 0 "collect_err_msg='Could not verify secureboot comptability'"
fi

echo "grub sig: $grub_sigs_file" >> $LOCAL_LOG_FILE
echo "Checking grub sigs" >> $LOCAL_LOG_FILE
cp_secureboot --verify-sigs-file $grub_sigs_file >> $LOCAL_LOG_FILE
if [ $? -ne 0 ] ; then
	endCheck 0 "collect_err_msg='Package is not compatible to secureboot keys.'"
fi


if [ -f ${metadataPathArg}/secureboot_kernel_sigs.full ] ; then
	kernel_sigs_file="${metadataPathArg}/secureboot_kernel_sigs.full"
elif [ -f `dirname ${metadataPathArg}`/secureboot_kernel_sigs.full ] ; then
	kernel_sigs_file="`dirname ${metadataPathArg}`/secureboot_kernel_sigs.full"
else
        echo "Kernel sigs info file does not exist" >> $LOCAL_LOG_FILE
        endCheck 0 "collect_err_msg='Could not verify secureboot comptability'"
fi

echo "kernel sig: $kernel_sigs_file" >> $LOCAL_LOG_FILE
echo "Checking kernel sigs" >> $LOCAL_LOG_FILE
cp_secureboot --verify-sigs-file $kernel_sigs_file >> $LOCAL_LOG_FILE
if [ $? -ne 0 ] ; then
	endCheck 0 "collect_err_msg='Package is not compatible to secureboot keys.'"
fi

echo "success"
exit 0
