#!/bin/bash
. `dirname $0`"/DDR_shared_functions"

GREP=/bin/grep
ECHO=/bin/echo
LSPCI=/sbin/lspci
CPPROD_UTIL=$CPDIR/bin/cpprod_util

uppak_should_enable=1

parseArgs "$@"


#USFW must be enabled
IS_USFW_ENABLED=`$CPPROD_UTIL FwIsUsermode 2>/dev/null | awk '{print $1}'`
if [ "$IS_USFW_ENABLED" == "0" ] ; then
	uppak_should_enable=0
fi

if [ "$vsx_supported" != "1" ]; then
	#If unsupported, VSX must be disabled
	IS_VSX=`$CPPROD_UTIL FwIsVSX 2>/dev/null | awk '{print $1}'`
	if [ "$IS_VSX" == "1" ]; then
		uppak_should_enable=0
	fi
fi

if [ "$sp_supported" != "1" ]; then
	#If unsupported, SP must be disabled
	IS_SP=`/usr/bin/is_scalable_platform.sh 2>/dev/null`
	if [ "$IS_SP" == "1" ] ; then
		uppak_should_enable=0
	fi
fi

if [ "$ipv6_supported" != "1" ]; then
	#If unsupported, IPv6 must be disabled
	IS_IPV6_ENABLED=`$CPPROD_UTIL FwIsFireWallIPv6 2>/dev/null | awk '{print $1}'`
	if [ "$IS_IPV6_ENABLED" == "1" ] ; then
		uppak_should_enable=0
	fi
fi

if [ "$nics_supported" != "all" ]; then
	#If all nics are not supported, check against
	#the given list of PCI device IDs
	found_supported_nic=0
	IFS=","
	for nic in $nics_supported; do
		lspci_output=`$LSPCI -d $nic 2>/dev/null`
		if [ "$lspci_output" != "" ] ; then
			found_supported_nic=1
		fi
	done
	if [ "$found_supported_nic" == "0" ]; then
		uppak_should_enable=0
	fi
fi

#UPPAK must not already be enabled
SXL_IS_USERMODE=`$CPPROD_UTIL SxlIsUsermode 2>/dev/null | awk '{print $1}'`
if [ "$SXL_IS_USERMODE" == "1" ] ; then
	uppak_should_enable=0
fi

#Check if UPPAK has already been automatically enabled during a previous upgrade
SXL_WAS_AUTOENABLED=`$CPPROD_UTIL SxlWasAutoEnabled 2>/dev/null | awk '{print $1}'`
if [ "$SXL_WAS_AUTOENABLED" == "1" ] ; then
	uppak_should_enable=0
fi

$ECHO "uppak_should_enable=\"$uppak_should_enable\""

exit 0
