#!/bin/bash
. `dirname $0`"/DDR_shared_functions"

#script verifies that the required partition size is between 'min' and 'max'
#parameters received:
# 1) 'min', minimum allowed partition size, in GB
# 2) 'max', maximum allowed partition size, in GB
# 3) 'partition_size' the required partition size, in GB

BC=/usr/bin/bc
ECHO=/bin/echo

rv=0
internal_err_str="Internal error: "

#parameters received are 'min', 'max', and 'partition_size'
parseArgs "$@"

# remove this section once DDR handles errors from collect
if [ "X$collect_err_msg" != "X" ]
then
	endCheck 1 "$collect_err_msg" 
fi


#-------------------------
#MAIN
#-------------------------

if [ "X$min" = "X" ] && [ "X$max" = "X" ]
then
	endCheck 2 "$internal_err_str 'min' and 'max' in conditions_set.json are both empty"
fi

if [ "X$partition_size" = "X" ]
then
	endCheck 2 "$internal_err_str partition_size recieved from collect is empty."
fi

if [ "X$partition" = "X" ]
then
	endCheck 2 "$internal_err_str partition recieved from collect is empty."
fi

if [ "$partition" == "lv_current" ]
then
	partition="root (lv_current)"
fi

error_message="The size of partition $partition is ${partition_size}GB"

if [ ! -z $min ]
then
	if (( $($ECHO "$partition_size" '<' "$min" | $BC -l) )) 
	then
			error_message="${error_message}. At least ${min}GB is required."
			rv=1
	fi
fi

if [ ! -z $max ]
then
	if (( $($ECHO "$partition_size" '>' "$max" | $BC -l) )) 
	then
			error_message="${error_message}. At most ${max}GB is required."
			rv=1
	fi
fi


endCheck $rv "$error_message"