#!/bin/bash
############################## Utilites ###############################

CPPROD_UTIL=$CPDIR/bin/cpprod_util

function cpprod_true() {
	eval "$1 | grep 1 > /dev/null"
}

RPM_FAILED=2
RPM_SUCCESS=1

# Machine Roles : GW,MGMT,SA,LOGSERVER,MDS,VSX, SME, SECONDERY_MGMT, SMARTCENTER

ROLE=NONE
PRODUCTS_TO_RMOVE=()
REMOVED_PACKAGES=0

#Error Codes
# 2 - Invalid Machine Role
# 3 - Failed Uninstalling an RPM

############################### Helpers ###################################

function isMDS() {
	if $CPPROD_UTIL CPPROD_GetValue PROVIDER-1 MDSDIR 1 > /dev/null; then
		if [[ $? ]] ; then
			true
		else 
			false
		fi
	else 
		false
	fi
}

function get_machine_role() {
	local IS_GW=false
	local IS_MGMT=false
	if cpprod_true "${CPPROD_UTIL} FwIsFirewallModule"; then
		IS_GW=true
	fi
	if cpprod_true "${CPPROD_UTIL} FwIsFirewallMgmt"; then
		IS_MGMT=true
	fi
	if [[ "$IS_GW" == true ]] && [[ "$IS_MGMT" == true ]]; then
		ROLE=SA
	elif [[ "$IS_GW" == true ]]; then 
		if cpprod_true "${CPPROD_UTIL} FwIsVSX"; then
			ROLE=VSX
		else
			ROLE=GW
		fi
	elif [[ "$IS_MGMT" == true ]]; then
		if isMDS; then
			ROLE=MDS
		elif cpprod_true "${CPPROD_UTIL} RtIsAnalyzerServer"; then
			ROLE=SME
		elif cpprod_true "${CPPROD_UTIL} FwIsLogServer"; then
			ROLE=LOGSERVER
		elif ! cpprod_true "${CPPROD_UTIL} FwIsPrimary"; then
			ROLE=SECONDERY_MGMT
		else 
			ROLE=SMARTCENTER
		fi
	else 
		return 1
	fi
	return 0
}

function get_products_to_remove() {
	case "${ROLE}" in 
		VSX)
			;;
		GW)
			;;
		SME)
			PRODUCTS_TO_RMOVE+=('CVPN' 'PPACK')
			;;
		SECONDERY_MGMT)
			PRODUCTS_TO_RMOVE+=('CVPN' 'PPACK')
			;;
		SMARTCENTER)
			PRODUCTS_TO_RMOVE+=('CVPN' 'PPACK')
			;;
		MDS)
			PRODUCTS_TO_RMOVE+=('CVPN' 'PPACK')
			;;
		LOGSERVER)
			PRODUCTS_TO_RMOVE+=('CVPN' 'PPACK')
			;;
		SA)
			;;
		*)
			;;
	esac
}

function remove_product() {
	local PRODUCT_NAME=$1
	local INSTALLED_PRODUCTS=$($CPPROD_UTIL CPPROD_GetInstalledProducts)
	if [[ $INSTALLED_PRODUCTS =~ $PRODUCT_NAME ]]; then
        local PRODUCT_VERSION=$($CPPROD_UTIL CPPROD_GetCurrentVersion $PRODUCT)
        PRODUCT_VERSION=$(echo -e "${PRODUCT_VERSION}" | tr -d '[:space:]')
        local PRODUCT_SP=$($CPPROD_UTIL CPPROD_GetCurrentServicePack $PRODUCT | tr -d '[:space:]')
        local PRODUCT_MSP=$($CPPROD_UTIL CPPROD_GetCurrentMSP $PRODUCT | tr -d '[:space:]')
        local PRODUCT_SU=$($CPPROD_UTIL CPPROD_GetValue ${PRODUCT}//${PRODUCT_VERSION}//SP${PRODUCT_SP}//MSP${PRODUCT_MSP} SilentUninstall 1)
        echo "Removing ${PRODUCT} From ${ROLE}, Command: '${PRODUCT_SU}'."
        eval ${PRODUCT_SU}
        if [[ $? ]]; then
            ((REMOVED_PACKAGES++))
            true
        else
            false
        fi
    fi
	true
}

function reset_packages_indexing() {
    echo "Reading packages from database."
    PACKAGE_LIST=$(dbget -v installer:packages | awk '{print $1}' | awk -F":"  '{print $3}')
    for PACKAGE in $PACKAGE_LIST
    do
        echo "Reseting $PACKAGE indexing."
        echo "Reseting key installer:packages:$PACKAGE:package_indexed"
        $(dbset installer:packages:$PACKAGE:package_indexed)
    done
}

################## Main ####################################

if ! get_machine_role; then 
	echo "Error: Cannot get machine role."
	exit 2
fi

touch $DADIR/bin/ROLE_FIXER_SCRIPT_MARKER

get_products_to_remove

if [[ ${#PRODUCTS_TO_RMOVE[@]} == 0 ]]; then
	exit 0
fi

echo "The role of the machine is ${ROLE}, removing un-needed packages."

for PRODUCT in "${PRODUCTS_TO_RMOVE[@]}"; do
  if ! remove_product $PRODUCT; then
       exit 3
  fi
done

if [[ $REMOVED_PACKAGES -gt 0 ]]; then
    reset_packages_indexing
fi




	

