#!/bin/sh

if [ -x /bin/echo ] ; then
	ECHO=/bin/echo
else
	ECHO=echo
fi

if [ -x /bin/cp ] ; then
	CP=/bin/cp
else
	CP=cp
fi

if [ -x /bin/sed ] ; then
	SED=/bin/sed
else
	SED=sed
fi

LOGFILE="/var/log/LastMinorVersionFix.elg"

cpprod_cmd()
{
	CPPROD_BIN=$CPDIR/bin/cpprod_util
	$ECHO "Runinng: $CPPROD_BIN $*" >>$LOGFILE
	CPPROD_VAL=$($CPPROD_BIN "$@" 2>>$LOGFILE)
	CPPROD_RET=$?
	$ECHO "Command return value is: $CPPROD_RET" >>$LOGFILE
	$ECHO "Value is: $CPPROD_VAL" >>$LOGFILE
	
	#we treat <nil> and "" as error 
	if [ $CPPROD_RET -ne 0 ] || [ "$CPPROD_VAL" = "<nil> " ] ; then
		CPPROD_VAL=""
	fi
	if [ "$CPPROD_VAL" = "" ] ; then
		CPPROD_RET=1
	fi
	$ECHO "$CPPROD_VAL" | $SED 's/ //g'
	
	$ECHO "Function return value is: $CPPROD_RET" >>$LOGFILE
	return $CPPROD_RET
}

$ECHO "Fixing LastMinorVerion" >>$LOGFILE

. /opt/CPshared/5.0/tmp/.CPprofile.sh

if [ ! -d "$CPDIR" ] ; then
	$ECHO "Can't find directory CPDIR ($CPDIR), exiting" >>$LOGFILE
	exit 1
fi

#backup current registry file
REG="$CPDIR/registry/HKLM_registry.data"
$ECHO "backup registry file: $REG to $REG.BeforeLastMinorVersionFix" >>$LOGFILE
$CP "$REG" "$REG.BeforeLastMinorVersionFix" >>$LOGFILE 2>&1

# get current LastMinorVersion
CPPROD_OUT=`cpprod_cmd CPPROD_GetLastMinorVersion`
CPPROD_RET=$?
if [ $CPPROD_RET -eq 0 ] ; then
	$ECHO "LastMinorVersion exist with value: $CPPROD_OUT no need to fix registry, exiting." >>$LOGFILE
	exit 0
fi

# Get VersionText
CPPROD_OUT=`cpprod_cmd CPPROD_GetVerText CPshared`
CPPROD_RET=$?
if [ $CPPROD_RET -ne 0 ] ; then
	$ECHO "failed to get VersionText, exiting." >>$LOGFILE
	exit 1
fi

$ECHO "VersionText before manipulation is: $CPPROD_OUT " >>$LOGFILE
VersionText=${CPPROD_OUT%%.*}
$ECHO "VersionText after manipulation is: $VersionText " >>$LOGFILE

# Get LastHFAIndex of FW1
CPPROD_OUT=`cpprod_cmd CPPROD_GetLastHFAIndex FW1`
CPPROD_RET=$?
LastHFAIndex=$CPPROD_OUT
$ECHO "LastHFAIndex is: $LastHFAIndex" >>$LOGFILE

if [ $CPPROD_RET -eq 0 ] ; then
	LastMinorVersion=$VersionText"."$LastHFAIndex
else
	$ECHO "LastHFAIndex is not set (NULL or ZERO), no need to fix registry, exiting." >>$LOGFILE
	exit 0
fi

# Set the correct value of LastMinorVersion
CPPROD_OUT=`cpprod_cmd CPPROD_SetLastMinorVersion $LastMinorVersion`

exit 0
