#!/usr/bin/tclsh

# DeploymentAgent wrapper for SMO(Scalable Platform) #

source $env(SMODIR)/bin/glib.tcl
source $env(SMODIR)/bin/g_proc.tcl
source /usr/lib/smo/libsmo.tcl
source /usr/lib/smo/libcluster.tcl
 
# DEBUG / TESTING FLAGS #
set DISPLAY_REAL_DACLIENT_OUTPUT 0  ;# Don't convert the execution output from DAClient into user friendly output
set DEBUG              0
set DEBUG_TRACE        0
set ALWAYS_LOG_DEBUG   0 ;# Append to log DEBUG messages even if not in DEBUG MODE
set DEBUG_PRINT_DB     0 ;# Print GAIA DB query in case in debug mode
set TEST_MODE          0 ;# When the screen is updated, Append the update on the screen instead the last screen
set SKIP_AUDIT         0
set TRACE_HEADER       [string repeat " " 30]

# members id's and status array for table printing #
set member_ids [list ]
array set members_status {}
  
# keys of the array returned by  get_local_pkg_action_information function
set PKG_INFORMATION_NAME            "name"
set PKG_INFORMATION_DISPLAY_NAME    "display_name"
set PKG_INFORMATION_STATE           "state"
set PKG_INFORMATION_KEY             "key"
set PKG_INFORMATION_IS_VALID_ACTION "is_valid_action"
set PKG_INFORMATION_QUERY_OUTPUT    "query_output"
  
# Current terminal size
set console_columns 80
set console_lines   24

# Line count for the header and table (Important for clearing all terminal content after the table)
set table_lines     0
set header_lines    4 ;# time_date_line + new_line + header_line + new_line
 
# Table columns width
set TABLE_COLUMN_WIDTH_MEMBER_ID     13
set TABLE_COLUMN_WIDTH_MEMBER_STATUS 64

# DA build conf file
set REQUIRED_DA_BUILD_FILE "/etc/sp_upgrade_min_da_build"
set MAX_DA_BUILD_NUMBER 200

# General 'global' variables
array set ping {
        
    # DEFINES matching to Basic_common.h and ISbase.h of DeploymentAgent module
    BIND_PING                               "installer"
    BIND_PACKAGES                           "packages" 
    BIND_REBOOT                             "requires_reboot" 
    BIND_VOLATILE                           "volatile"
    BIND_SUPPRESS_REBOOT	            "suppress_reboot"
    BIND_PROGRESS                           "progress"
    BIND_REMAINING_TIME                     "remaining_time"
    BIND_STATUS                             "status_text"
    BIND_CLISH_PROGRESS_TYPE                "action_type"
    BIND_CLISH_PROGRESS_PKG                 "action_pkg"
    BIND_CLISH_PROGRESS_RESULT              "action_result"
    BIND_SELF_UPDATE_IN_PROGRESS            "self_update_in_progress"
    BIND_PACKAGE_DISPLAY_NAME               "display_name"
    BIND_PACKAGE_DESCRIPTION                "description"
    BIND_AGENT_INSTALL_MSG                  "agent_install_msg"
    
    BIND_CLISH_INSTALL_ID_STR               "install_id_str"
    
    
    BIND_CLISH_IMPORT_ID_STR                "import_id_str"
    BIND_IMPORT_LOCAL                       "local"
    
    BIND_VALIDATE_PACKAGE                   "validate_package"
    BIND_VALIDATE_PACKAGE_ALL_MEMBERS       "package_validation_all_members"
    BIND_VALIDATE_PACKAGE_NAME              "File name"
    BIND_VALIDATE_PACKAGE_DISPLAY_NAME      "Display name"
    BIND_VALIDATE_PACKAGE_KEY               "Package key"
    BIND_VALIDATE_PACKAGE_STATE             "Package state"
    BIND_VALIDATE_PACKAGE_ACTION_VALIDATION "Valid for action"
    
    INVALID_PKG_ACTION  0
    VALID_PKG_ACTION    1

    

    # INSTALLER_ACTIONS
    BIND_IMPORT                     "import_pkg"    
    BIND_INSTALL                    "install_pkg"
    BIND_DELETE                     "delete"
    BIND_DOWNLOAD                   "download_pkg"
    BIND_DOWNLOAD_AND_INSTALL       "download_and_install"
    BIND_PAUSE_DOWNLOAD             "pause_download"
    BIND_RESUME_DOWNLOAD            "resume_download"

    # Action_Types
    ACTION_AGENT_DISABLE            "disable"
    ACTION_AGENT_ENABLE             "enable"
    ACTION_AGENT_START              "start"
    ACTION_AGENT_STOP               "stop"
    ACTION_UPDATE_DA                "update_da"
    ACTION_INSTALL_DA               "agent_install"
    ACTION_CHECK_FOR_UPDATES        "check-for-updates"
    ACTION_DELETE                   "delete"
    ACTION_DOWNLOAD                 "download"
    ACTION_DOWNLOAD_AND_INSTALL     "download_install"
    ACTION_IMPORT                   "import"
    ACTION_IMPORT_CLOUD             "add_private_package"
    ACTION_INSTALL                  "install"
    ACTION_REINSTALL                "reinstall"
    ACTION_UNINSTALL                "uninstall"
    ACTION_UPGRADE                  "upgrade"
    ACTION_VERIFY                   "verify"
    ACTION_PAUSE                    "pause"
    ACTION_RESUME                   "resume"

    # GCLISH_COMMAND_ACTIONS
    GCLISH_ARG_IMPORT                "import"
    GCLISH_ARG_DOWNLOAD_AND_INSTALL  "download-and-install"
    GCLISH_ARG_AGENT                 "agent"
    GCLISH_ARG_AGENT_UPDATE          "update"
    GCLISH_ARG_AGENT_INSTALL         "install"
    GCLISH_IMPORT_CLOUD              "import cloud"
    GCLISH_MEMBER_IDS                "member_ids"
    GCLISH_COMPLETELY                "completely"
    # PKG_STATES
    PKG_STATE_NO_STATE                      0
    PKG_STATE_PENDING_DOWNLOAD_STATE        1
    PKG_STATE_DOWNLOADING_STATE             2
    PKG_STATE_PENDING_INSTALL_STATE         3
    PKG_STATE_INSTALLING_STATE              4
    PKG_STATE_INSTALLED_STATE               5
    PKG_STATE_DOWNLOAD_FAILED_STATE         6
    PKG_STATE_INSTALL_FAILED_STATE          7
    PKG_STATE_UNINSTALLING_STATE            8
    PKG_STATE_UNINSTALL_FAILED_STATE        9
    PKG_STATE_PARTIALLY_DOWNLOADED_STATE   10
    PKG_STATE_INSTALL_SKIPPED_STATE        11
    PKG_STATE_INSTALL_CONTAINED            12
       
    
    DACLIENT_WRAPPER_FLAG  "-DAClient"
    
    # FILES_PATH
    IMPORT_PATH         "/var/log/upload"
    DAWRAPPER_LOG_PATH  "/var/log/da_smo.log"
    TEST_MODE_FILE      "/home/admin/TEST_MODE" 
    REPO_PATH           "/var/log/CPda/repository"
    WRAPPER_PATH        "/opt/CPda/bin/cli_sp_dawrapper.tcl"

    DBSET_BINARY        "dbset"
    DBGET_BINARY        "dbget"
    
    
    # CLI_FLAGS_sent_to_the_wrapper by DAclish_utils.c
    DACLISH_UTILS_CMD_SWITCH        "-cmd"
    DACLISH_UTILS_PKG_NAME_SWITCH   "-pkg_name"

    
    # Commands
    CMD_PREFIX                          ". \$CPDIR/tmp/.CPprofile.sh"
    CMD_GET_LOCAL_MEMBER_ID             "cphaprob chassis -i local_id"
    CMD_GET_ALL_CLUSTER_MEMBER_IDS      "cphaprob chassis -i all_ids"    
    CMD_GET_ACTIVE_CLUSTER_MEMBER_IDS   "cphaprob chassis -i active_ids"

    # OTHER
    DELAYED_DISPLAY         500
    PROGRESS_REFRESH_TIME   500

    
    members_str      "all"        ;# Members to perform_action
    LOCAL_MEMBER_STR "local"
    ALL_MEMBERS_STR  "all"
    
    members_status_changed 1            ;# Flag indicating whether or not members_status has changed since last table_printing
    
    local_member_id      ""
    package_display_name ""
    package_key          ""
    members_cant_do_action ""
    
    is_uninstall_completely 0              ;#for default we want to uninstall takes one by one. 
    
      
    gclish_cmd ""            ;# GCLISH command_line arguments
    
    # Log_types
    LOG_INFO   "INFO"
    LOG_ERROR  "ERROR"
    LOG_DEBUG  "DEBUG"
    
    # OUTPUT_MSG_FOR_THE_USER IN_CASE_COMMAND_EXECUTION_ON_REMOTE_MEMBER    HAS_FAILED
    MSG_COMMAND_EXECUTION_FAILED "Failed to execute command."
    
    NA_STR                        "NA"
    PKG_DISPLAY_NAME_ESCAPED_CHAR "\\|"
    
    MSG_SEARCH_IN_CLOUD      "Checking package availability on cloud..."
    MSG_AGENT_IS_UP_TO_DATE  "Deployment Agent is up to date"
    MSG_DA_ISNT_RUNNING      "Deployment Agent is not running."
    WAIT_FOR_RESULT_TIME_OUT_SECONDS 1000
} 

set F_BIND_SUPRESS_REBOOT "$ping(BIND_PING):$ping(BIND_VOLATILE):$ping(BIND_SUPPRESS_REBOOT)"

# List of actions which might require reboot after action is completed
set REBOOT_ACTIONS [list                \
    $ping(ACTION_DOWNLOAD_AND_INSTALL)  \
    $ping(ACTION_INSTALL)               \
    $ping(ACTION_REINSTALL)             \
    $ping(ACTION_UNINSTALL)             \
    $ping(ACTION_UPGRADE)               \
]


# Flag for skipping installation on download-and-install action in case local_member_id wasn't included by the use
set skip_installation_on_local_member 0

# For package_action verification
set PKG_FAILED_RETRIEVING_INFO 1
set PKG_DOESNT_EXIST           2
set PKG_INVALID_FOR_ACTION     4
set PKG_KEY_MISMATCH           8
set PKG_OLD_DA_VERSION         16

     
set NO_ACTIVE_MEMBERS            0
set ALL_MEMBERS_ARE_ACTIVE       2
set SEVERAL_MEMBERS_ARE_INACTIVE 1

set ALL_MEMBERS_ACTIONS_WERE_COMPLETED 1

# Channel id for wrapper log file
set log_file_channel_id ""
 
# INSTALLER KEYS #
set KEY_SELF_UPDATE_IN_PROGRESS "$ping(BIND_PING):$ping(BIND_SELF_UPDATE_IN_PROGRESS)"
set KEY_CLISH_PROGRESS_PKG      "$ping(BIND_PING):$ping(BIND_CLISH_PROGRESS_PKG)"
set KEY_CLISH_PROGRESS_RESULT   "$ping(BIND_PING):$ping(BIND_CLISH_PROGRESS_RESULT)"
set KEY_AGENT_INSTALL_MSG       "$ping(BIND_PING):$ping(BIND_AGENT_INSTALL_MSG)"

# '2' according to 'wait_for_result' function in DAclish_utils.c
set STATE_SELF_UPDATE_IN_PROGRESS "2"
     
# "installer agent <action>" actions
set ping(AGENT_ACTIONS) [list $ping(ACTION_INSTALL_DA) $ping(ACTION_UPDATE_DA) $ping(ACTION_AGENT_START) $ping(ACTION_AGENT_STOP) $ping(ACTION_AGENT_DISABLE) $ping(ACTION_AGENT_ENABLE)]


# A list of DAClient actions in which we don't
# Need to wait and constantly pull their progress/result from Gaia DB
set SKIP_PROGRESS_DISPLAY [list $ping(ACTION_AGENT_START) $ping(ACTION_AGENT_STOP) $ping(ACTION_AGENT_DISABLE) $ping(ACTION_AGENT_ENABLE)]


set PRINT_EXCLUTION_LIST          [list "*Import Failed*" "*official documentation*" "*Uninstall the package and try again*"    ]
set ACTION_VERIFICATION_EXCLUTION [list     $ping(ACTION_IMPORT)  $ping(ACTION_IMPORT_CLOUD)  {*}$ping(AGENT_ACTIONS) $ping(ACTION_CHECK_FOR_UPDATES)  "show" ]
                                                            
set SHOW_CMD_EXCLUTION_LIST          [list "*\\*\\*\\**" "\\*\\*"  "Display name*"        ]
set DISPLAY_PROGRESS_EXCULSION_LIST  [list $ping(ACTION_IMPORT) $ping(ACTION_DELETE) ]


set ACTIONS_FOR_LOCAL_MEMBER_ONLY [list  $ping(ACTION_CHECK_FOR_UPDATES)  $ping(ACTION_IMPORT_CLOUD) ]

 
set MEMBER_STR_SEPERATOR ","

set ACTION_STATE_MATCH    "match"
set ACTION_STATE_MISMATCH "mismatch"

# match/mismatch  -  action_name  -   package_state    -   output-message 
set ACTION_VALIDATION_ERROR_MESSAGES [list \
    $ACTION_STATE_MATCH     $ping(ACTION_INSTALL)               "$ping(PKG_STATE_INSTALLED_STATE)"                    "Package is already installed"            \
    $ACTION_STATE_MISMATCH  $ping(ACTION_UNINSTALL)             "$ping(PKG_STATE_INSTALLED_STATE)"                    "Package isn't installed"                    \
    $ACTION_STATE_MISMATCH  $ping(ACTION_REINSTALL)             "$ping(PKG_STATE_INSTALLED_STATE)"                    "Package isn't installed"                    \
    $ACTION_STATE_MATCH     $ping(ACTION_DOWNLOAD)              "$ping(PKG_STATE_PENDING_INSTALL_STATE)"        "Package was already imported"        \
    $ACTION_STATE_MATCH     $ping(ACTION_DOWNLOAD_AND_INSTALL)  "$ping(PKG_STATE_PENDING_INSTALL_STATE)"        "Package was already imported"        \
]

# wdebug
# Wrapping proc for debugging #
if {$DEBUG} {
    rename proc _proc
    _proc proc {name arglist body} {
        set ignore_list [list fn putsd pvars print_time_and_date_header \
                         remove_excluded_lines puts_table_body_ open_log_file_for_writing \
                         isDebugMode close_log_file get_current_time get_current_date writeLog \
                         log_last_action_status get_current_time_and_date fit_status_to_cell strip_HTML_tags \
                         clean_execution_output screen_clear_after_table InitTZData get_blades_ids convert_to_formal_member_ids \
                         convert_to_formal_member_id list_to_csv screen_clear screen_update_console_size \
                         screen_reset_with_blank_str get_empty_line_str tlog]
        if {[lsearch $ignore_list $name] == -1 && [string match "*::*" $name] == 0} { 
            _proc $name $arglist [concat "proc_start;" $body ";proc_end"]
        } else {
            _proc $name $arglist $body
        }
        
    }
    _proc proc_start {} {
        global DEBUG_TRACE TRACE_HEADER
        puts stderr "$TRACE_HEADER>>>> ENTER PROC [lindex [info level -1] 0]()"
        if {$DEBUG_TRACE} {
            for {set level [expr [info level] -1]} {$level > 0} {incr level -1} {
                puts stderr "$TRACE_HEADER     LEVEL $level: [info level $level]"
            }
        }
        
        puts stderr ""
    }
    _proc proc_end {} {
        global TRACE_HEADER
        puts stderr "$TRACE_HEADER<<<< LEAVE PROC [lindex [info level -1] 0]\n"
    }
}

# Print variable values in case in Debug mode
proc pvars {args} {
    global DEBUG
    if {$DEBUG} {
        set res ""
        foreach i $args {
             if [uplevel info exists $i] {
                 append res "$i='[uplevel set $i]'\n"
             } else {
                 append res "$i=NULL\n"
             }
        }
        # puts stderr $res
        putsd "Variables values:\n$res"
    }
}

# Return the current function name
proc fn {} {
    return [lindex [split [info level [expr [info level] - 1]]] 0]
}

# for debugging: print line number
proc printLine {frame_info} {
    # Getting value of the key 'line' from the dictionary 
    # returned by 'info frame'
    set result [dict get [info frame $frame_info]  line]
}

# ###################################### Functions  ###################################### #

# Clear current stdin buffer(Useful before asking and reading user's input)
proc clear_stdin {} {
    fconfigure stdin -blocking 0
    exec stty raw
    read stdin
    exec stty -raw
    fconfigure stdin -blocking 1
}

proc get_user_input {user_input} {
    upvar $user_input loc_user_input
    set loc_user_input ""
    clear_stdin
    set loc_user_input [string tolower [gets stdin] ]
}

#
# Convert member_status (e.g. DAService status) in order to fit one line / friendlier status
#
proc convert_member_status {status} {
    global ping
    
    # NOTE: Don't consolidate with the next IF_STATEMENT because further conversion need to be done
    if { [regexp  -nocase -- "\nIf the package is not visible"  $status] } {
        regsub  "\nIf the package is not visible.*\$" $status  "" status
    } 
    
    if {[regexp -nocase -- "Deployment agent: Disabling the Deployment Agent service"  $status]} {
        return "Disabling the Deployment Agent service"
    } elseif {[regexp -nocase -- "Deployment agent: Enabling the Deployment Agent service"  $status]} {
        return "Enabling the Deployment Agent service"
    } elseif {[regexp -nocase -- "Error: Deployment Agent is already running"  $status]} {
        return "Deployment Agent is already running"
    } elseif {[regexp -nocase -- "Deployment agent: Stopping the Deployment Agent process"  $status]} {
        return "Stopping the Deployment Agent process"
    } elseif {[regexp -nocase -- "Error: Deployment Agent is not running"  $status]} {
        return $ping(MSG_DA_ISNT_RUNNING)
    } elseif {[regexp -nocase -- "Deployment agent: Starting the Deployment Agent process"  $status]} {
        return "Starting the Deployment Agent process"
    } elseif {[regexp -nocase -- "Deployment Agent: agent is up to date"  $status]} {
        return $ping(MSG_AGENT_IS_UP_TO_DATE)
    } elseif { [regexp -nocase -- "Initiating import of"  $status]} {
        regsub  "Initiating import of" $status  "Importing" status
        return $status
    } elseif { [regexp -nocase -- "Initiating verify of"  $status]} {
        regsub  "Initiating verify of" $status  "Initiating verification of" status
        return $status
    } elseif { [regexp -nocase -- "Initiating install of"  $status]} {
        regsub  "Initiating install of" $status  "Initiating installation of" status
        return $status
    } elseif { [regexp -nocase -- "Checking for new available packages"  $status]} {
        return "Checking for new available packages..."
    } elseif {[regexp "Downloading:\\s+\$"  $status]} {
        return "Downloading..."
    } elseif {[regexp -nocase -- "was successfully removed"  $status]} {
        return "The package was successfully removed from the machine."
    } elseif {[regexp -nocase -- "Package.*was downloaded successfully."  $status]} {
        # Convert "Package <pkg_name> was downloaded successfully.   Download time: 3 seconds"  --> "The package <pkg_name> was successfully imported."
        regsub -nocase -- "Download time:.*" $status  "" status ;# # Removes the time it took to download the package in order for the status to fit in one line
        regsub -nocase -- "Package " $status  "The package " status
        regsub -nocase -- " was downloaded successfully." $status  " was successfully imported." status
        return $status
    } elseif {[regexp  -nocase -- "The package you are trying to import already exists in your"  $status]} {
        return "The package you are trying to import already exists."
    } elseif {[regexp  -nocase -- " Additional Info:\$"  $status]} {
        # TODO: best thing would be to remove it from Main-Train in case there is no additional info.
        regsub " Additional Info:\$" $status  "" status
        return $status
    }
    
    return $status
}

proc putsd {msg} {
    global DEBUG

    if {$DEBUG} {
        puts "\[DEBUG\]: $msg"
    }
}

# Description: Minimize member status length by shortening it's package name so it would fit the status cell width 
#                     (e.g.  'initiating installation of package_number_one.tgz' will be converted to 'initiating installation of 'package_num...' ')
#
# Return: 
#        0    - In case shortening succeeded
#       -1    - In case shortening failed
#
proc fit_status_to_cell {status  new_status} {
    global TABLE_COLUMN_WIDTH_MEMBER_STATUS
    
    upvar $new_status loc_new_status
    set loc_new_status $status                 ;# Default new status is the original status
    
    #  Minimum required package name
    set MIN_LENGTH_PKG_NAME 5

    set PKG_NAME_PREFIX         '
    set PKG_NAME_PREFIX_LENGTH  [string length $PKG_NAME_PREFIX]
    set PKG_NAME_POSTFIX        "..'"
    set PKG_NAME_POSTFIX_LENGTH [string length $PKG_NAME_POSTFIX]
    
    # Extract the package name from the status
    if {[get_pkg_name_from_status $status pkg_name]} {
        return -1
    }
    
    # Trim
    set pkg_name [string trim $pkg_name]
    
    set status_length       [string length $status]
    set pkg_name_length     [string length $pkg_name]
    set max_length_pkg_name [expr $TABLE_COLUMN_WIDTH_MEMBER_STATUS - ($status_length - $pkg_name_length)]

    if {$pkg_name_length <= $max_length_pkg_name} {
        # Status already fit the cell width
        set loc_new_status $status
        if {[isDebugMode]} {  puts "fit_status_to_cell_CASE_1"}
        return 0
    } elseif {$pkg_name_length > $max_length_pkg_name  && $max_length_pkg_name < [expr $MIN_LENGTH_PKG_NAME + $PKG_NAME_POSTFIX_LENGTH + $PKG_NAME_PREFIX_LENGTH]} {
        if {[isDebugMode]} {  puts "fit_status_to_cell_CASE_2"}
        # PKG_NAME will be shorter than required in order to fit one line (MIN_LENGTH_PKG_NAME)  or status is way too long for one line anyway
        return -1
    } 
    
    # Make package name shorter in order to fit the cell width
    set short_pkg_name_last_index [expr $max_length_pkg_name-1-$PKG_NAME_POSTFIX_LENGTH-$PKG_NAME_PREFIX_LENGTH]
    set short_pkg_name            [string range $pkg_name 0 $short_pkg_name_last_index]
    set short_pkg_name            [string trim $short_pkg_name]
    set short_pkg_name            $PKG_NAME_PREFIX$short_pkg_name$PKG_NAME_POSTFIX
    
    
    # Replace pkg_name with short_pkg_name in new_status
    if {[regsub -- $pkg_name $status $short_pkg_name loc_new_status] < 1} {
        # Failed to replace pkg_name with short_pkg_name in status
        if {[isDebugMode]} {  puts "fit_status_to_cell_CASE_3"}
        return -1
    }
    
    return 0
}


proc get_pkg_name_from_status {status pkg_name} {
    upvar $pkg_name loc_pkg_name
    
    set loc_pkg_name ""
    
    # '?' in "*.?" for non-greedy
    if {[regexp -nocase -- "(?:package )(.+?)(?: was )" $status - loc_pkg_name] && $loc_pkg_name != ""} {
        return 0
    } elseif {[regexp -nocase -- "(?:Importing )(.+?)(?:\\.\\.+?$)" $status - loc_pkg_name] && $loc_pkg_name != ""} {
        return 0
    } elseif {[regexp -nocase -- "(?: of )(.+?)(?:\\.\\.+?$)" $status - loc_pkg_name] && $loc_pkg_name != ""} {
        return 0
    }    
    return -1
}


proc action_to_status {action} {
    putsd "[fn]: action=$action"
    global ping
    
    if {$action == $ping(ACTION_AGENT_ENABLE)} {
        return "Enabling Deployment Agent..." 
    } elseif {$action == $ping(ACTION_AGENT_DISABLE)} {
        return "Disabling Deployment Agent..." 
    } elseif {$action == $ping(ACTION_AGENT_START)} {
        return "Starting Deployment Agent..." 
    } elseif {$action == $ping(ACTION_AGENT_STOP)} {
        return "Stopping Deployment Agent..." 
    } elseif {    $action == $ping(ACTION_UPDATE_DA)} {
        return "Initiating Deployment Agent update..." 
    } elseif {    $action == $ping(ACTION_INSTALL_DA)} {
        return "Installing Deployment Agent..."
    } elseif {    $action == $ping(ACTION_DOWNLOAD)} {
        return "Initiating $action"
    } elseif {    $action == $ping(ACTION_DOWNLOAD_AND_INSTALL)} {
        return "Initiating download and install..." 
    } elseif {    $action == $ping(ACTION_CHECK_FOR_UPDATES)} {
        return "Checking for new available packages. This operation may take a few moments" 
    } elseif {    $action == $ping(ACTION_IMPORT)} {
        return "Initiating package import..." 
    } elseif {    $action == $ping(ACTION_IMPORT_CLOUD)} {
        return $ping(MSG_SEARCH_IN_CLOUD) 
    } elseif {    $action == $ping(ACTION_INSTALL)} {
        return "Initiating package installation..." 
    } elseif {    $action == $ping(ACTION_VERIFY)} {
        return "Initiating package verification..." 
    } elseif {    $action == $ping(ACTION_DELETE)} {
        return "Initiating package delete..." 
    } elseif {    $action == $ping(ACTION_REINSTALL)} {
        return "Initiating package reinstallation..." 
    } elseif {    $action == $ping(ACTION_UNINSTALL)} {
        return "Initiating package uninstallation..." 
    } else   {
        return "Initiating $action..." 
    }
}

proc set_action_initiation_status {action members_str} {
    global ping
     
    if { $action == $ping(ACTION_DOWNLOAD) } {
        set_member_status $members_str ""
    }

    set_member_status $members_str [action_to_status $action]
}


proc get_validation_error_message {action_name package_state} {
    global ACTION_VALIDATION_ERROR_MESSAGES        ACTION_STATE_MATCH    ACTION_STATE_MISMATCH

    set DEFAULT_ERROR_MESSAGE  "Package is not ready for $action_name"
    
    foreach {cur_match_str cur_action cur_pkg_state cur_msg} $ACTION_VALIDATION_ERROR_MESSAGES {
        if {$action_name == $cur_action} {
            if {($cur_match_str == $ACTION_STATE_MATCH            &&    $package_state == $cur_pkg_state) 
                || 
                ($cur_match_str == $ACTION_STATE_MISMATCH         &&    $package_state != $cur_pkg_state)   } {
                # Message was found -> Return current message
                return $cur_msg
            } 
        }
    }
    
    # Message wasn't found, Return default message
    return $DEFAULT_ERROR_MESSAGE
}

 
proc isDebugMode {} {
    global DEBUG
    return $DEBUG
}


proc isTestingMode {} {
    global TEST_MODE
    return $TEST_MODE
}


proc open_log_file_for_writing {} {
    global ping
    global log_file_channel_id

    if [catch {set log_file_channel_id [open $ping(DAWRAPPER_LOG_PATH) a] } error_msg ] {
        #puts "Failed to open log file $ping(DAWRAPPER_LOG_PATH)"
        #puts $error_msg
        return -1
    }
    
    return 0
}

proc close_log_file {} {
    global log_file_channel_id

    catch {close $log_file_channel_id}
    
    return 0
}

proc get_current_time {} {
    # Get system time
    set systemTime     [clock seconds]
    # Convert system time into required format
    set time             [clock format $systemTime -format %H:%M:%S]
    
    return $time
}

proc get_current_date {} {
    # Get system time
    set systemTime     [clock seconds]
    # Convert system time into required format
    set date             [clock format $systemTime -format {%b %d}]
    
    return $date
}

proc get_current_time_and_date {} {
    set time [get_current_time]
    set date [get_current_date]
    
    # Concatenate current time and date
    set current_time_and_date "\[$date $time\]"

    return $current_time_and_date
}

 
proc writeLog {log_type function_name msg {display_log 1} {log_function_name 1}} {
    global ping 
    global log_file_channel_id
    global ALWAYS_LOG_DEBUG
    
    
    if {[open_log_file_for_writing] == 0 && 
        ($log_type != $ping(LOG_DEBUG) ||  ([isDebugMode] || $ALWAYS_LOG_DEBUG))     } {
            set func_name_in_log ""
            if {$log_type != $ping(LOG_INFO) && $log_function_name} {
                    set func_name_in_log "\[$function_name\]"
            }
            
            # Write to log file
            puts $log_file_channel_id  "[get_current_time_and_date] \[$log_type\] $func_name_in_log $msg"

            # Flush to file
            flush    $log_file_channel_id
        
            close_log_file
    }
    
    
    if {$display_log} {
        # Print log to the screen
        if {$log_type == $ping(LOG_INFO)} {
            puts "[get_current_time_and_date] $msg"
        } elseif {$log_type == $ping(LOG_DEBUG)} {
            if {[isDebugMode]} {
                puts "[get_current_time_and_date] \[$log_type\] \[$function_name\] $msg"
            }
        } elseif {$log_type == $ping(LOG_ERROR) } {
            puts "[get_current_time_and_date] \[$log_type\] $msg"
        } 
    }
    
}
 
proc clean_execution_output {execution_output} {
    # Remove lines containing 'License has expired'
    regsub -all -line -nocase ".*License has expired.*(?:\n|$)" $execution_output "" execution_output

    # Remove empty lines from execution output by replacing "\n\n...\n" with "\n"
    regsub -all "(\n)+" $execution_output "\n" execution_output

    # Trim
    set execution_output [string trim $execution_output]
    return $execution_output
}

proc clean_errors_from_gexec_output {execution_output} {
    
    # Remove lines containing 'License has expired'
    regsub -all -line -nocase ".*License has expired.*(?:\n|$)" $execution_output "" execution_output

    # Remove empty lines from execution output by replacing "\n\n...\n" with "\n"
    regsub -all "(\n)+" $execution_output "\n" execution_output

    # Trim
    set execution_output [string trim $execution_output]
    return $execution_output
}





# Description: Execute bash command from GCLISH locally
#
# Return: 
#         0 - Command execution succeeded
#        -1 - Command execution failed
#
#        <execution_output> - execution output 
#
proc gclish_exec {command execution_output} {
    global ping
    global DEBUG_PRINT_DB
    
    putsd "[fn]()"
    
    upvar $execution_output loc_execution_output
    set loc_execution_output ""

    # Inject 'bash -c' into the command string and redirect stderr into stdout
    set command "|bash -c \"$command 2>&1\" "

    writeLog $ping(LOG_DEBUG) "[fn]" "command=$command"

    # Execute command
    if {[catch {set fd [open $command r+ ] } error_msg]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to execute command: $command"
        writeLog $ping(LOG_ERROR) "[fn]" "$error_msg" 0
        return -1
    }
    
    # Read execution output
    set loc_execution_output [read $fd] 

    # Close execution output file channel
    catch {close $fd}
    
    putsd "[fn](): output before clean: $loc_execution_output"
    
    # Clean execution output from unnecessary content 
    set loc_execution_output [clean_execution_output $loc_execution_output]
    
    if {$DEBUG_PRINT_DB || [string first "$ping(BIND_PING):" $loc_execution_output]==-1} {
        writeLog $ping(LOG_DEBUG) "[fn]" "Result, after output was cleaned:\n$loc_execution_output"
    } else {
        writeLog $ping(LOG_DEBUG) "[fn]" "SKIPPING OUTPUT PRINTING(DEBUG_PRINT_DB FLAG is OFF)"
    }
    return 0
}


# Description: Execute bash command from GCLISH shell on remote cluster members and parse result per member id into array
#
# Return: 
#         0 - Command execution succeeded
#        -1 - Command execution failed
#
#        <execution_output_array>    - execution output per member_id
#
proc gclish_exec_remote_members {members_str cmd execution_output_array {timeout 60} {retries 3}} {
    global  ping
    upvar $execution_output_array loc_execution_output_array
    set TIME_BETWEEN_ATTEMPTS 1000
    
    # Initialize execution_output array
    if {[info exist loc_execution_output_array]} {unset loc_execution_output_array}
    array set loc_execution_output_array {}
    
    if {$members_str == "" } {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to execute command."
        writeLog $ping(LOG_ERROR) "[fn]" "members_str is empty." 0
        return -1
    }
        
    # Build command string #
    set f_cmd  "gexec -a "
    if {$timeout > 0} {
        append f_cmd "-x $timeout "
    }
    # (echo is used in order to make sure gexec will always return some output per member as an indication whether or not the command was successfully executed)
    append f_cmd "-b $members_str -c \\\"$cmd ; echo \\\"  "
    
    writeLog $ping(LOG_DEBUG) "[fn]" "command=$f_cmd"

    # Try to execute command and parse its output per member
    for {set cur_attempt 1} {$cur_attempt <= $retries} {incr cur_attempt} {
        if {$cur_attempt > 1} {
            # Sleep between attempts
            after $TIME_BETWEEN_ATTEMPTS
        }
        # Execute command
        if {[gclish_exec $f_cmd execution_output]} {
            # Failed to execute command
            writeLog $ping(LOG_ERROR) "[fn]" "Failed to execute command, members_str=$members_str , command=\n$cmd" 0
            continue
        } 
        # putsd "[fn](): output before parsing: $execution_output"
        if {[gexec_parse_results $members_str $execution_output loc_execution_output_array]} {
            # Failed to parse result
            writeLog $ping(LOG_ERROR) "[fn]" "Failed to parse execution output members_str='$members_str' command='$cmd' output='$execution_output'" 0
            continue
        }
        return 0 ;# Execution and parsing succeeded
    }
    return -1 ;# Failed to execute or parse execution result
}


proc parse_execution_output_per_member {execution_output   parsed_execution_output_array} {
    global ping
    set GEXEC_ERROR_FAILED_TO_EXECUTE_COMMAND "No blade will be affected, please check your blade string"
    upvar $parsed_execution_output_array loc_parsed_execution_output_array

    # Initialize parsed_execution_output_array and failed_ids list #
    if {[info exist loc_parsed_execution_output_array]} {unset loc_parsed_execution_output_array}
    array set loc_parsed_execution_output_array {}

    
    # TODO: add regex description 
    set REGEX_MEMBER_ID    {^(?:\[)(.*?)(?:\])}   ;# e.g. match on [1_01]  where subMatchVar_1 = 1_01
    set REGEX_LINE_CONTENT {(?:\[.*?\])(.*$)}     ;# e.g. 'Remove' [1_01]  from line where in subMatchVar_1 = line without [1_01]

    # Split execution output into lines
    set execution_output_lines [split $execution_output "\n"]
    
    set prasing_res 0
    
    # Build output per member_id and add it to parsed_execution_output_array #
    foreach {line} $execution_output_lines {
        set member_id    ""
        set line_content ""
        
        if {[regexp $GEXEC_ERROR_FAILED_TO_EXECUTE_COMMAND $line]} {
            writeLog $ping(LOG_ERROR) "[fn]" "Several member id's are down"         0
            writeLog $ping(LOG_ERROR) "[fn]" "Execution output:\n$execution_output" 0
            return -1
        # extract member_id and and execution_output content from <line>
        } elseif { ![regexp $REGEX_MEMBER_ID $line - member_id]  || $member_id == ""  ||   ![regexp $REGEX_LINE_CONTENT $line - line_content] } {
            # Failed to parse current line
            
                writeLog $ping(LOG_ERROR) "[fn]" "Failed to parse line:\n$line"                                0
            
            set prasing_res -1
            continue
        }

        # Append current line to member's output
        if {![info exists loc_parsed_execution_output_array($member_id) ]} {
            set loc_parsed_execution_output_array($member_id) $line_content
        } else {
            append loc_parsed_execution_output_array($member_id) "\n$line_content"
        }
    }
    
    if {$prasing_res != 0} {
        # Failed to parse several lines
        writeLog $ping(LOG_DEBUG) "[fn]" "Execution output:\n$execution_output"
        return -1
    }
    
    # Remove last empty line which was added by "echo" command in gclish_exec
    foreach {member_id} [array name loc_parsed_execution_output_array] {
        if { [string index $loc_parsed_execution_output_array($member_id) end] == "\n" } {
            set loc_parsed_execution_output_array($member_id) [string range $loc_parsed_execution_output_array($member_id) 0 end-1]
        }
    }
    return 0
}

# Verify that each member_id output exist in execution_output_array
proc verify_execution_output {members_str  parsed_execution_output_array} {
    global ping
    upvar $parsed_execution_output_array loc_parsed_execution_output_array
    
    # Convert members_str to member_ids LIST
    set member_ids [member_str_to_lst $members_str]
    
    # Verify that the command was executed on all members in $member_ids
    set found_missing_member_id 0
    
    foreach {member_id} $member_ids {
        if { ![info exists loc_parsed_execution_output_array($member_id)]  } {
            set found_missing_member_id    1
            # Add ERROR_MSG to parsed_execution_output_array
            set loc_parsed_execution_output_array($member_id) $ping(MSG_COMMAND_EXECUTION_FAILED)
            writeLog $ping(LOG_ERROR) "[fn]" "Failed to execute command on: $member_id" 0
        }
    }
    
    if {$found_missing_member_id} {
        return -1
    }
    return 0
}

# input: members_str , execution_output
# output: output_array
# Parse execution output result and fill <parsed_execution_output_array> per member_id
proc gexec_parse_results {members_str execution_output parsed_execution_output_array} {
    global ping
    global DEBUG_PRINT_DB    
    
    upvar $parsed_execution_output_array loc_parsed_execution_output_array

    # Parse gexec_execution_output into loc_parsed_execution_output_array per member
    if {[parse_execution_output_per_member $execution_output loc_parsed_execution_output_array]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Parsing execution output has failed" 0
        return -1
    }
    
    set result 0
    
    # Verify that all member ids exist in loc_parsed_execution_output_array
    if {[verify_execution_output $members_str loc_parsed_execution_output_array]} {
        set result -1
    }
    
    if {[isDebugMode]} {
        puts "arrays:"
        if {$DEBUG_PRINT_DB || [string first "$ping(BIND_PING):" [array get loc_parsed_execution_output_array]]==-1} {
            puts [array get loc_parsed_execution_output_array]
            puts [array get loc_failed_ids_array]
        } else {
            puts "[fn](): DONT PRINT OUTPUT, DEBUG_PRINT_DB is OFF"
        }
    }
    return $result
}



# Description: Get current cluster blade ids 
#
# Return: 
#         0  - Command execution succeeded
#        -1  - Command execution failed
#       
#        <blade_ids> - Blades_ids (all or only the active ones) (e.g. "1_01,1_02,2_01")
#
proc get_blades_ids {blade_ids {all_ids 0}} {
    global ping
    
    upvar $blade_ids loc_blade_ids
    
    # Build command #
    set cmd    "$ping(CMD_PREFIX) ; "

    if { $all_ids == 1 } {    
        append cmd $ping(CMD_GET_ALL_CLUSTER_MEMBER_IDS)
    } else {    
        append cmd $ping(CMD_GET_ACTIVE_CLUSTER_MEMBER_IDS)
    }    
    
    
    # Execute command
    set res [gclish_exec $cmd loc_blade_ids]
    return $res
}


# Copy local file to other cluster members using "asg_cp2blades" command
#
# NOTE: asg_cp2blades output will be  "Operation completed successfully" in case one of the members is down, so further verification is required 
proc cp2blades_exec {member_str src dst} {
    global  ping

    set REQUIRED_OUTPUT "Operation completed successfully"
    set FAILED_OUTPUT   "Operation succeeded on"

    set cmd "asg_cp2blades -a -b $member_str $src $dst"    

    # Execute asg_cp2blades #
    set execution_output ""
    
    if {[gclish_exec $cmd execution_output]} {
        # Failed to execute command
        writeLog $ping(LOG_ERROR) "[fn]" "File transfer has failed."
        writeLog $ping(LOG_DEBUG) "[fn]" "Command: '$cmd' "
        return -1
    }
    
    # Verify asg_cp2blades output #
    if {![regexp -nocase -- $REQUIRED_OUTPUT  $execution_output] || 
         [regexp -nocase -- $FAILED_OUTPUT  $execution_output] } {
            writeLog $ping(LOG_ERROR) "[fn]" "File transfer has failed."
            writeLog $ping(LOG_ERROR) "[fn]" "Execution output: \n$execution_output" 0
            writeLog $ping(LOG_DEBUG) "[fn]" "Command: '$cmd' "
            return -1
    }
    return 0
}

# Perform installer action according to execution arguments
proc perform_installer_action {action pkg_str pkg_display_name members_str {should_audit_action 1}} {
    global ping members_cant_do_action
    global PKG_INFORMATION_NAME  PKG_INFORMATION_DISPLAY_NAME PKG_INFORMATION_KEY
    global skip_installation_on_local_member 

    pvars action pkg_str pkg_display_name members_str
    
    set original_pkg_str    $pkg_str
    
    writeLog $ping(LOG_INFO) "[fn]"  "ACTION=$action , pkg_str=$pkg_str , members_str=$members_str" 0
    
    array set local_pkg_action_information {}
    set       pkg_key    ""

    # Display action_initiation_status
    if {$action == $ping(ACTION_DOWNLOAD_AND_INSTALL) &&
        $skip_installation_on_local_member} {
            set members_str_without_local_id [get_members_str_without_local_id $members_str]    
            set_action_initiation_status $action $members_str_without_local_id ;# Display download-and-install on all except the local member
            set_action_initiation_status $ping(ACTION_DOWNLOAD) $ping(local_member_id)              ;# Display download on the local member
    } else {
            set_action_initiation_status $action $members_str  
    }

    # Print member status
    print_members_status
    after $ping(DELAYED_DISPLAY) ;#Sleep
     
     
     
    # In case pkg_str is a numeric value(pkg index) and not full_pkg_name -->  get full_pkg_name from local member in order to avoid pkg_index mismatch when executing commands on remote members
    # In case ACTION=import , pull package information from local member in order to import the package on remote members with identical pkg_display_name as on the local member
    if { $action != $ping(ACTION_IMPORT_CLOUD) &&
         ([string is integer -strict $pkg_str] ||
         $action == $ping(ACTION_IMPORT)       ||
         $action == $ping(ACTION_INSTALL)      ||
         $action == $ping(ACTION_DOWNLOAD)     ||
         $action == $ping(ACTION_DOWNLOAD_AND_INSTALL) )} {
        if { $action == $ping(ACTION_UNINSTALL) } {
            if { [get_local_pkg_action_information $action $pkg_str local_pkg_action_information 1] } {
                return -1 ; # Failed to execute package query
            }
        } else {
            if { [get_local_pkg_action_information $action $pkg_str local_pkg_action_information 0] } {
                return -1 ; # Failed to execute package query
            } 
        }    
        if { $local_pkg_action_information($PKG_INFORMATION_NAME) != $ping(NA_STR) } {
            # Update pkg action values for verification and action according to package state on local member #
        
            if {$action != $ping(ACTION_IMPORT)} {
                # Update pkg_str with query result # (pkg_str might be the local package index, however,  query result is always real package name which is required in order to run remote package action)
                set pkg_str $local_pkg_action_information($PKG_INFORMATION_NAME)
            }
            
            set pkg_display_name $local_pkg_action_information($PKG_INFORMATION_DISPLAY_NAME)
            set pkg_key          $local_pkg_action_information($PKG_INFORMATION_KEY)
            
        } elseif {$action != $ping(ACTION_IMPORT)} {
            set_member_status $members_str    ""
            writeLog $ping(LOG_INFO) "[fn]" "Package index doesn't exist on the local member ('$original_pkg_str')\n"
            return -1
        }
    } 

    
    if {$action == $ping(ACTION_DOWNLOAD_AND_INSTALL) || $action == $ping(ACTION_DOWNLOAD)} {
        set pkg_action_verification_member_str $ping(local_member_id)
        set pkg_action_verification_action     $ping(ACTION_DOWNLOAD)
    } else {
        set pkg_action_verification_member_str $members_str
        set pkg_action_verification_action     $action
    }
    
    
    # Run package verification before action (And suggest to import pkg on members that the pkg wasn't imported on in case action='install')
    if {[should_verify_pkg_action $pkg_action_verification_action]     && 
        [verify_pkg_action        $pkg_action_verification_action $pkg_str $pkg_action_verification_member_str [array get local_pkg_action_information] ] } {
            putsd "[fn](): Package verification has failed"
            # Package verification has failed #
            
            # set pkg_str with the original pkg_str in case package wasn't found in GAIA DB
            if {$pkg_str == $ping(NA_STR)} {
                set pkg_str $original_pkg_str
            }
            #if members_cant_do_action is not "" , print the msg
            if { [info exists members_cant_do_action] } {
                writeLog $ping(LOG_INFO) "[fn]" "$action $pkg_str cannot be performed on members :$members_cant_do_action please use member_ids option with the correct members list \n"
            }
            writeLog $ping(LOG_INFO) "[fn]" "Action was aborted ($action, '$pkg_str'). \n"
            return -1
    }
    putsd "[fn](): SUCCESS: After Package verification(maybe it was skipped)"
    
    # Sleep
    after $ping(DELAYED_DISPLAY)
    
    if { $should_audit_action  &&
         [audit_action $pkg_str $action $members_str  $skip_installation_on_local_member] } {
        return -1
    }

    
    # Clear CPUSE progress value in GAIA DB
    if {[clear_progress $members_str]} {
        # Failed to clear DB values
        return -1
    }
    
    # Clear package progress(Just in case the result from last action for this pkg wasn't cleared)
    if {$pkg_str != ""} {
        clear_db_package_progress $pkg_str $members_str ;# No need to check return value because key might not exist in DB
    }
    
    putsd "[fn](): After clear_progress"
    writeLog $ping(LOG_INFO) "[fn]" "Action start: ACTION=$action , resolved_pkg_str=$pkg_str , members_str=$members_str" 0
    
    putsd "[fn](): switch action=$action"
    switch -- $action \
    $ping(ACTION_IMPORT) {
        return [pkg_import $action $pkg_str $pkg_display_name $members_str]
    }\
    $ping(ACTION_DOWNLOAD) {
        return [pkg_download $action $pkg_str $pkg_display_name $pkg_key $members_str]
    }\
    $ping(ACTION_DOWNLOAD_AND_INSTALL) {
        return [pkg_download_and_install $action $pkg_str $pkg_display_name $pkg_key $members_str  $skip_installation_on_local_member]
    }\
    default {
        return [invoke_action $action $members_str $pkg_str]
    }
    
}
  
  
proc get_pkg_value {pkg_str key value} {
    global ping
    upvar $value loc_value
    set loc_value ""
    
    set cmd "$ping(DBGET_BINARY) $ping(BIND_PING):$ping(BIND_PACKAGES):$pkg_str:$key"
    
    if {[gclish_exec $cmd ex_output] } {
        writeLog $ping(LOG_DEBUG) "[fn]" "Failed to get package value from DB. ($ping(BIND_PING):$ping(BIND_PACKAGES):$key)"
        return -1
    }
    set loc_value $ex_output
    
    
    return 0
}
  
proc get_pkg_reboot_flag {pkg_str reboot_flag} {
    global ping
    upvar $reboot_flag loc_reboot_flag
    set loc_reboot_flag ""
    
    if {[get_pkg_value $pkg_str $ping(BIND_REBOOT) loc_reboot_flag]} {
        writeLog $ping(LOG_DEBUG) "[fn]" "Failed to get package reboot flag value from DB."
        return -1
    }
    
    return 0
}

proc set_supress_reboot {members_str val} {
    global ping
    global F_BIND_SUPRESS_REBOOT

    writeLog $ping(LOG_DEBUG) "[fn]" "Clearing $F_BIND_SUPRESS_REBOOT"
            
    set cmd "$ping(DBSET_BINARY) $F_BIND_SUPRESS_REBOOT $val"

    set res 0
    if {[gclish_exec_remote_members $members_str $cmd member_result_array]} {
        writeLog $ping(LOG_DEBUG) "[fn]" "Failed to set DB value(gclish_exec failed)."
        set res -1
    } else {
        # Verify that execution output is empty(means success) per member_id
        foreach member_id [array names member_result_array] {
            if {$member_result_array($member_id) != ""} {
                writeLog $ping(LOG_DEBUG) "[fn]" "Failed to set DB value:\n[array get member_result_array]"
                # dbset failed on the current member_id
                set res -1
                break
            }
        }     
    }
    if {$res} {
        writeLog $ping(LOG_DEBUG) "[fn]" "Failed to set DB value."
        return -1
    }   
    return 0
}

proc are_all_members_in_admin_down {members_str} {
	global ping

	# Get admin_down state of all members that are going to be upgraded
	set active_member_list ""
	set cmd "cphaprob list | grep admin_down"
	if {[gclish_exec_remote_members $members_str $cmd member_result_array]} {
		writeLog $ping(LOG_DEBUG) "[fn]" "Failed to check admin_down pnote"
		return 0
	} else {
		foreach member_id [array names member_result_array] {
			if {$member_result_array($member_id) == ""} {
				# Member is not in admin_down state
				writeLog $ping(LOG_DEBUG) "[fn]" "Member $member_id is not in admin_down state"
				append active_member_list "${member_id},"
			}
		}
		if {$active_member_list != ""} {
			# Not all members are in admin_down state - prevent upgrade
			if {[catch {exec gresolve -d $active_member_list -p ign_state} parsed_active_member_list]} {
				writeLog $ping(LOG_DEBUG) "[fn]" "Failed to parse list of members not in admin_down state"
			} else {
				set active_members_msg "Attention! The following members are not in admin-down state: $parsed_active_member_list. Please set the members to admin-down state to continue.\nTo set them in admin-down state, run from expert mode: g_clusterXL_admin -b $parsed_active_member_list down\n"
				puts_ -nonewline $active_members_msg
			}
			return 0
		}
	}
	return 1
}

proc reboot_handler {action pkg_str members_str} {
    global ping SKIP_AUDIT
    global F_BIND_SUPRESS_REBOOT
    
    # Check whether package requires reboot
    if {[get_pkg_reboot_flag $pkg_str reboot_flag]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to get package value from DB."
        return -1
    }
    
    if {$reboot_flag == 1} {
        puts "The machines($members_str) will automatically reboot after $action."
        puts "Do you want to continue? (\[y\]es / \[n\]o)"
        
        get_user_input user_input

        # Handle user input
        if {[string equal -nocase $user_input "y"]  ||
            [string equal -nocase $user_input "yes"]   } {

            if {$action == $ping(ACTION_UPGRADE)} {
                if {![are_all_members_in_admin_down $members_str]} {
                        return -1
                }
            }

            if {$action == $ping(ACTION_UNINSTALL) || $action == $ping(ACTION_DOWNLOAD_AND_INSTALL) || $action == $ping(ACTION_INSTALL) || $action == $ping(ACTION_REINSTALL) || $action == $ping(ACTION_UPGRADE)} {
                if {[catch {exec dbget smo:image_clone:auto} val]} {
                    set val "N/A"
                }
                if {$val == "on"} {
                    set image_clone_msg "Attention! Image auto-clone is enabled. Please turn it off to continue.\nTo disable image auto-clone run: set cluster configuration image auto-clone state off\n"
                } elseif {$val == "N/A"} {
                    set image_clone_msg "Attention! In case image auto-clone is enabled, this operation may be overridden by auto-cloning from SMO during next boot.\nTo disable image auto-clone run: set cluster configuration auto-clone state off\n"
                } else {
                    return 0
                }
                puts_ -nonewline $image_clone_msg
                if { [confirm_msg] != 1 || $val == "on" } {
                    return -1
                }
            }
            return 0
        } elseif {$user_input == "s"} {
            ;# curretly reboot supressing is a limitation
            return -1 
            
            if {[set_supress_reboot $members_str "1"]} {
                writeLog $ping(LOG_ERROR) "[fn]" "Failed to set DB value(supress_reboot)."
                return -1
            }
        } elseif  {$user_input == "n"} {
            return -1
        } else {
            return -1
        }
    } else {
        ;# curretly reboot supressing is a limitation
        # Clear supress_reboot
        # if {[set_supress_reboot $members_str ""]} {
            # writeLog $ping(LOG_ERROR) "[fn]" "Failed to set DB value(supress_reboot)."
            # return -1
        # }
    }

    
    return 0
}

# Description: 
#   Ask user for action confirmation and audit in logs
#
# Return: 
#    0 - action has been confirmed
#   -1 - otherwise
#
proc audit_action {pkg_str action members_str skip_installation_on_local_member} {
    global ping
    global SKIP_AUDIT
    global REBOOT_ACTIONS

    if {$SKIP_AUDIT} {
        return 0 ;# Skip auditing
    }

    
    if {[isTestingMode]} {
        return 0 ;# Skip auditing
    }
    
    set LEVEL_WARNING  "warning"
    set LEVEL_CRITICAL "critical"
    set WARN_DELETE    "\n*** In case the package is installed, you might not be able to uninstall and Import the package ever again ! ***"
    set REASON         "Maintenance"

    # Action list to audit  {action,level,action_msg,warn_msg,default_reason}
    set AUDIT_ACTION_LIST [list \
        $ping(ACTION_AGENT_DISABLE)        $LEVEL_WARNING  "\nDAService disable"                            ""            $REASON  \
        $ping(ACTION_AGENT_ENABLE)         $LEVEL_WARNING  "\nDAService enable "                            ""            $REASON  \
        $ping(ACTION_AGENT_START)          $LEVEL_WARNING  "\nDAService start"                              ""            $REASON  \
        $ping(ACTION_AGENT_STOP)           $LEVEL_WARNING  "\nDAService stop"                               ""            $REASON  \
        $ping(ACTION_DELETE)               $LEVEL_CRITICAL "\nPackage delete ($pkg_str)"                    $WARN_DELETE  $REASON  \
        $ping(ACTION_DOWNLOAD_AND_INSTALL) $LEVEL_WARNING  "\nPackage download and Installation ($pkg_str)" ""            $REASON  \
        $ping(ACTION_INSTALL)              $LEVEL_WARNING  "\nPackage Installation ($pkg_str)"              ""            $REASON  \
        $ping(ACTION_REINSTALL)            $LEVEL_WARNING  "\nPackage Reinstallation ($pkg_str)"            ""            $REASON  \
        $ping(ACTION_UNINSTALL)            $LEVEL_WARNING  "\nPackage Uninstallation ($pkg_str)"            ""            $REASON  \
        $ping(ACTION_UPGRADE)              $LEVEL_WARNING  "\nUpgrade ($pkg_str)"                           ""            $REASON  \
    ]
  
    # Check if <action> appear in <AUDIT_ACTION_LIST>
    foreach {cur_action cur_level cur_action_msg cur_warn_msg cur_reason} $AUDIT_ACTION_LIST {
        if { $action == $cur_action } {

            if {$skip_installation_on_local_member && $action == $ping(ACTION_DOWNLOAD_AND_INSTALL)} {
                set cur_action_msg  "\nPackage download and Installation ($pkg_str) \[Package won't be installed on the local member\]"
            }
            

            # Handle reboot flag if nesscary
            if {[lsearch $REBOOT_ACTIONS $action] >= 0} {
                if {[reboot_handler $action $pkg_str $members_str]} {
                    return -1
                }
            }
            
            # confirm_and_audit #
            clear_stdin
            if { [confirm_and_audit $cur_level $cur_action_msg $members_str $cur_warn_msg $cur_reason] != 1 } {
                writeLog $ping(LOG_INFO) "[fn]" "Action was cancelled"
                # Action wasn't confirmed confirmed
                return -1
            }
            
            if {![isDebugMode] && ![isTestingMode]} {
                # Remove audit output from the screen
                screen_reset_with_blank_str    
                print_members_status
            }
            
            # Action was confirmed
            return 0
        }
    }
    
    # Action doesn't require auditing
    return 0
}
 
# Description: 
#   Get package information using CDK QUERY on the local member 
#   (e.g. whether or not DAService allow the <action> on <pkg_name>)
#
# Return: 
#    0  - Query succeeded
#   -1  - Query has failed
#
#        <pkg_action_info_arr> - Array containing package information
#
proc get_local_pkg_action_information {action pkg_name pkg_action_info_arr run_on_all_members} {
    global ping 
    global ACTION_VERIFICATION_EXCLUTION
    
    global    PKG_INFORMATION_NAME    PKG_INFORMATION_DISPLAY_NAME       PKG_INFORMATION_STATE  \
              PKG_INFORMATION_KEY      PKG_INFORMATION_IS_VALID_ACTION     PKG_INFORMATION_QUERY_OUTPUT
    
    upvar $pkg_action_info_arr loc_pkg_action_info_arr
    
    set ERR_MSG "Failed to retrieve package information.\n"
    
    # Initialize  pkg_action_information array
    if {[info exist loc_pkg_action_info_arr]}     {unset loc_pkg_action_info_arr}
    array set loc_pkg_action_info_arr  {}
    
    writeLog $ping(LOG_DEBUG) "[fn]" "ACTION=$action , pkg_name=$pkg_name"

    # Execute CDK query
    if {[get_package_information $ping(local_member_id) $action $pkg_name members_output_array $run_on_all_members]} { 
        writeLog $ping(LOG_ERROR) "[fn]" $ERR_MSG
        return -1
    }

    # Parse execution output per Package information category into Package Information array
    set execution_output  $members_output_array($ping(local_member_id))
    set loc_pkg_action_info_arr($PKG_INFORMATION_NAME)              [extract_v_from_DAClient_output $execution_output $ping(BIND_VALIDATE_PACKAGE_NAME)            ]
    putsd "loc_pkg_action_info_arr($PKG_INFORMATION_NAME)=$loc_pkg_action_info_arr($PKG_INFORMATION_NAME)"
    set loc_pkg_action_info_arr($PKG_INFORMATION_DISPLAY_NAME)      [extract_v_from_DAClient_output $execution_output $ping(BIND_VALIDATE_PACKAGE_DISPLAY_NAME)    ]
    putsd "loc_pkg_action_info_arr($PKG_INFORMATION_DISPLAY_NAME)=$loc_pkg_action_info_arr($PKG_INFORMATION_DISPLAY_NAME)"
    set loc_pkg_action_info_arr($PKG_INFORMATION_STATE)              [extract_v_from_DAClient_output $execution_output $ping(BIND_VALIDATE_PACKAGE_STATE)        ]
    putsd "loc_pkg_action_info_arr($PKG_INFORMATION_STATE)=$loc_pkg_action_info_arr($PKG_INFORMATION_STATE)"
    set loc_pkg_action_info_arr($PKG_INFORMATION_KEY)              [extract_v_from_DAClient_output $execution_output $ping(BIND_VALIDATE_PACKAGE_KEY)            ]
    set loc_pkg_action_info_arr($PKG_INFORMATION_IS_VALID_ACTION) [extract_v_from_DAClient_output $execution_output $ping(BIND_VALIDATE_PACKAGE_ACTION_VALIDATION)    ]
    putsd "loc_pkg_action_info_arr($PKG_INFORMATION_IS_VALID_ACTION)=$loc_pkg_action_info_arr($PKG_INFORMATION_IS_VALID_ACTION)"

    set loc_pkg_action_info_arr($PKG_INFORMATION_QUERY_OUTPUT)        $execution_output
    writeLog $ping(LOG_DEBUG) "[fn]" "loc_pkg_action_info_arr=[array get loc_pkg_action_info_arr]"
     
    # Verify Package_Information_array content
    if { $loc_pkg_action_info_arr($PKG_INFORMATION_NAME) == "" } {
        # Query has failed !  in case query succeeded the package name should be "NA" or a real package name
        writeLog $ping(LOG_ERROR) "[fn]" $ERR_MSG
        return -1
    }
    return 0 
}

proc pkg_download_and_install {action pkg_str pkg_display_name pkg_key members_str  skip_installation_on_local_member} { 
    global ping

    # Change bind string to download only
    set old_value $ping(BIND_DOWNLOAD_AND_INSTALL)
    set new_value $ping(BIND_DOWNLOAD)
    
    
    # Download package on local member and import package on the rest of the members
    if {[pkg_download $ping(ACTION_DOWNLOAD) $pkg_str $pkg_display_name $pkg_key $members_str]} {
        # package download has failed
        return -1
    }
    
    # Change ping package name
    set pkg_file_name [file tail $pkg_str]
    
    set dont_audit_action 0
    
    if {$skip_installation_on_local_member} {
        # Only download should be performed on local member because it's a must
        # Installation shouldn't be performed on local member because the user input didn't include local_member_id
        set members_str [get_members_str_without_local_id $members_str]     ;# Remove local_member_id from members_str
    }

    
    # Install package
    set res [perform_installer_action  $ping(ACTION_INSTALL) $pkg_file_name $pkg_display_name $members_str $dont_audit_action]
    return $res
}



proc is_pkg_state_pending_install {pkg_str} {
    global ping
    global PKG_INFORMATION_STATE
    
    array set local_pkg_action_information {}

    if { [get_local_pkg_action_information "all" $pkg_str local_pkg_action_information 0] } {
        writeLog $ping(LOG_DEBUG) "[fn]" "Failed to retrieve local package information"
        putsd "is_pkg_state_pending_install(): Failed to retrieve local package information"
        return 0
    }
    
    set pkg_state $local_pkg_action_information($PKG_INFORMATION_STATE)
    
    if {$pkg_state != $ping(PKG_STATE_PENDING_INSTALL_STATE) } {
        writeLog $ping(LOG_DEBUG) "[fn]" "pkg_state=$pkg_state"
        putsd "is_pkg_state_pending_install(): pkg_state=$pkg_state"
        return 0
    } 
    putsd "is_pkg_state_pending_install(): return 1"
    return 1
}



proc pkg_download {action pkg_str pkg_display_name pkg_key members_str} { 
    global ping

    global PKG_INFORMATION_NAME   PKG_INFORMATION_DISPLAY_NAME   PKG_INFORMATION_STATE  \
           PKG_INFORMATION_KEY  PKG_INFORMATION_IS_VALID_ACTION PKG_INFORMATION_QUERY_OUTPUT

    
    set members_str_without_local_id  [get_members_str_without_local_id $members_str]
    
    set_member_status $members_str_without_local_id "Waiting for package"
     
    # Download package to local member only and run download verification #
    if {[invoke_action $action $ping(local_member_id) $pkg_str]  || [is_pkg_state_pending_install $pkg_str] != 1  } {
        putsd "[fn](): Download package to local member only"
        set_member_status $members_str_without_local_id "" ;#Clear member status
        print_members_status
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to download package"
        return -1
    }
     
    # Package was downloaded to local member #
    
    # Import package to other members #
    if { $members_str_without_local_id != ""} {
        # Extract package file name
        set pkg_file_name    [file tail $pkg_str]

        # Change package path to <package full path from local repository> for [transfer_package]
        set pkg_path "$ping(REPO_PATH)/$pkg_key/$pkg_file_name"
     
        # Verify that the downloaded package exist in the local member repository
        if {![file exists $pkg_path]} {
            writeLog $ping(LOG_ERROR) "[fn]" "File doesn't exist($pkg_path)"
            return -1
        }
 
        # Import package on other members
        set res [perform_installer_action  $ping(ACTION_IMPORT) $pkg_path $pkg_display_name $members_str_without_local_id]
        return $res
    }
    return 0
}


# Remove the local_member_id from member_ids string(e.g. "2_01,1_02") and return it
proc get_members_str_without_local_id {members_str} {
    global ping
    
    set res [remove_member_id $members_str $ping(local_member_id)]
    return $res
}

proc remove_member_id {members_str   member_id_to_remove} {
    set member_ids_list [split $members_str ","]
    set members_str_without_member_id    ""

    # Add only members which are not member_id_to_remove
    foreach member_id $member_ids_list {
        if {$member_id != $member_id_to_remove } {
            if {$members_str_without_member_id != ""} { append members_str_without_member_id   ","}
            append members_str_without_member_id        $member_id
        }
    }

    return $members_str_without_member_id 
}




# Clear previous installer values from GAIA DB according to <action> before starting installer action
proc clear_db_value_by_action {action members_str} {
    global ping
    
    set key ""
    
    switch -- $action \
        $ping(ACTION_IMPORT_CLOUD) - \
        $ping(ACTION_IMPORT) {
            set key "$ping(BIND_PING):$ping(BIND_CLISH_IMPORT_ID_STR)"
        }\
        $ping(ACTION_INSTALL) {
            set key "$ping(BIND_PING):$ping(BIND_CLISH_INSTALL_ID_STR)"
        }\
        $ping(ACTION_REINSTALL) {
            set key "$ping(BIND_PING):$ping(BIND_CLISH_INSTALL_ID_STR)"
        }\
        default {
            # No need to clear GAIA DB values
            return 0
        }
    
    if { [clear_db_values $members_str $key] } {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to clear Gaia DB values\n"
        return -1
    }
    
    return 0
}


# Description: Get package/action information using CDK binding (validate_package() function in DAcdk.c) on remote  members
#              (e.g. whether or not DAService allow the <action> on <pkg_name>)
#
# Return: 
#         0 - Query succeeded
#        -1 - Query has failed
#
#        <query_result_array> - Array containing package information per member_id
#
proc get_package_information {members_str    action_name   pkg_string   query_result_array run_on_all_members} {
    global ping
    
    upvar $query_result_array loc_query_result_array
    
    if {[info exist loc_query_result_array]} {unset loc_query_result_array}
    array set loc_query_result_array {}
    
    # Extract package file_name from path (In case it's a file_path)
    set pkg_string_to_query [file tail $pkg_string]
    # Build Gaia DB key
    set key "$ping(BIND_PING):$ping(BIND_VALIDATE_PACKAGE):$pkg_string_to_query:$action_name"

    writeLog $ping(LOG_DEBUG) "[fn]" "members_str='$members_str' "
    
    # Execute package query on cluster members
    # if {[dbget_exec $members_str   $key   loc_query_result_array]} {
        # writeLog    $ping(LOG_ERROR) "[fn]" "Failed to retrieve pkg information."
        # return -1
    # }

    
    set cmd "$ping(WRAPPER_PATH) $ping(DACLIENT_WRAPPER_FLAG) "  
    if {$run_on_all_members == 1} {
        append cmd "package_validation_all_members"
    } else {
        append cmd "package_validation"
    }
    append cmd " $pkg_string $action_name"
    putsd "[fn](): about to execute command"
    if {[gclish_exec_remote_members  $members_str $cmd loc_query_result_array 20]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to get Gaia DB value." 0
        return -1
    }
    
    
    # parray loc_query_result_array
    # putsd "exit [fn]" ; exit

    
    writeLog $ping(LOG_DEBUG) "[fn]" "members_str='$members_str'  , Result: \n[array get loc_query_result_array]"
    return 0    
}


proc should_verify_pkg_action {action} {
    putsd "[fn]()"

    global ping
    global ACTION_VERIFICATION_EXCLUTION
    
    set SKIP_VERIFICATION         0
    set VERIFY_PKG_ACTION     1
    
    foreach action_exclusion $ACTION_VERIFICATION_EXCLUTION {
        if {$action == $action_exclusion} { 
            return $SKIP_VERIFICATION
        }
    }
    
    return $VERIFY_PKG_ACTION
}

 

proc verify_pkg_action_get_pkg_information {package_name    action   members_str    local_pkg_action_information_arr_data    member_db_array} {
    putsd "[fn]()"
    global ping
    global PKG_INFORMATION_QUERY_OUTPUT
    
    upvar $member_db_array loc_member_db_array
    array set local_pkg_action_information_arr $local_pkg_action_information_arr_data
    
    array set loc_member_db_array  {}

    if {     [info exist local_pkg_action_information_arr($PKG_INFORMATION_QUERY_OUTPUT)]
            &&
            ($action == $ping(ACTION_DOWNLOAD)    ||
            $action == $ping(ACTION_DOWNLOAD_AND_INSTALL))} {
        # No need to execute query
        # set query result only with local member output
        putsd "[fn](): No need to execute query"
        set loc_member_db_array($ping(local_member_id)) $local_pkg_action_information_arr($PKG_INFORMATION_QUERY_OUTPUT)
    } else {
        # Execute query
        if {[get_package_information $members_str $action $package_name loc_member_db_array 0]} { 
            writeLog $ping(LOG_ERROR) "[fn]" "Failed to query package information from member's DB."
            return -1
        }
    }
    
    return 0
}
 
proc get_DA_build {member_id} {

    global ping
    set cmd "$ping(DBGET_BINARY) $ping(BIND_PING):da_build"
    if {[gclish_exec_remote_members  $member_id $cmd build 20]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to get Gaia DB value." 0
        return -1
    }
    return $build($member_id)

}

proc get_required_DA_build {} {
    global REQUIRED_DA_BUILD_FILE ping
    if [catch {open $REQUIRED_DA_BUILD_FILE} fd] {
        writeLog $ping(LOG_ERROR) "[fn]" "conf file $REQUIRED_DA_BUILD_FILE does not exist." 0
        return 0		
    }
    set data [read -nonewline $fd]
    close $fd
    if ![regexp -line -- {^(\d+)} $data match required_build] {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to get required DA build from file." 0
        return 0
    }
    return $required_build
}

proc verify_pkg_action_verify_query {action   member_db_array_data    members_str_without_pkg} {
    putsd "[fn]()"
    global ping members_cant_do_action
    global PKG_FAILED_RETRIEVING_INFO    PKG_DOESNT_EXIST    PKG_INVALID_FOR_ACTION    PKG_KEY_MISMATCH   PKG_OLD_DA_VERSION 
    global MAX_DA_BUILD_NUMBER
    
    upvar $members_str_without_pkg loc_members_str_without_pkg 
    
    array set member_db_array $member_db_array_data
    
    set loc_members_str_without_pkg ""
    set error_mask        0
  
    # read DA build from file
    set required_build [get_required_DA_build]  

    # Initialize the pkg_key which will be compared against all cluster members pkg_key
    set main_pkg_key     ""
    
    # Verify that the package is identical and action is valid on all members #
    foreach member_id [array names member_db_array] {
        # Extract package information from querying the current member
        set pkg_name              [extract_v_from_DAClient_output $member_db_array($member_id) $ping(BIND_VALIDATE_PACKAGE_NAME)   ]
        set pkg_state             [extract_v_from_DAClient_output $member_db_array($member_id) $ping(BIND_VALIDATE_PACKAGE_STATE)  ]
        set pkg_key               [extract_v_from_DAClient_output $member_db_array($member_id) $ping(BIND_VALIDATE_PACKAGE_KEY)    ]
        set pkg_action_validation [extract_v_from_DAClient_output $member_db_array($member_id) $ping(BIND_VALIDATE_PACKAGE_ACTION_VALIDATION)]
        
        # Initialize main_pkg_key with the first valid pkg_key found in case it's empty (e.g. when installing a package only on the remote machines and not on the local machine)
        if {$main_pkg_key == "" &&  $pkg_key != "" && $pkg_key != $ping(NA_STR)} {
            set main_pkg_key      $pkg_key
            set ping(package_key) $pkg_key
        }
 
        set status ""
        set current_build [get_DA_build $member_id]       
        if {$action == "upgrade" && ( [expr {$current_build < $required_build} ] ) } {
            set status "Can't complete the action. Please update Deployment Agent according to the instructions in the Upgrade chapter in the Scalable Platform Administration Guide."
            set error_mask [expr {$error_mask | $PKG_OLD_DA_VERSION}]
        } elseif { $pkg_name == "" || $pkg_state == "" || $pkg_key == "" || $pkg_action_validation == "" } {
            set status "Failed to retrieve package information"
            set error_mask [expr {$error_mask | $PKG_FAILED_RETRIEVING_INFO}]
        } elseif {  $pkg_name == $ping(NA_STR) ||
                   ($pkg_action_validation == $ping(INVALID_PKG_ACTION) && $pkg_state == $ping(PKG_STATE_PENDING_DOWNLOAD_STATE) && $action == $ping(ACTION_INSTALL)) } {
            # Package doesn't exit  OR we tried to install a package which is pending for download(which means it doesn't exist)
            set status "Package doesn't exist"
            append members_cant_do_action "$member_id,"
            if {$loc_members_str_without_pkg != ""} { append loc_members_str_without_pkg "," }
            append loc_members_str_without_pkg "$member_id"
            set error_mask [expr {$error_mask | $PKG_DOESNT_EXIST}]
        } elseif {$pkg_action_validation == $ping(INVALID_PKG_ACTION)} {
            # Package is state is invalid for the action

            # Get invalid action reason
            set status [get_validation_error_message $action $pkg_state]
            
            set error_mask [expr {$error_mask | $PKG_INVALID_FOR_ACTION}]
            append members_cant_do_action "$member_id,"
        } elseif {$pkg_key != $main_pkg_key} {
            set status "Package key mismatch between members"
            set error_mask [expr {$error_mask | $PKG_KEY_MISMATCH}]
            append members_cant_do_action "$member_id,"
        } else {
            set status [get_verified_action_status $action]
            
            if {$member_id == $ping(local_member_id)} {
                set pkg_action_is_valid_on_local_member 1
            }
            
        }
        
        # Update member's status
        set_member_status $member_id $status
        putsd "[fn](): member=$member_id status= $status"
    } ;# For loop END
    
    return $error_mask
}
 
# Verify package state before action
proc verify_pkg_action {action package_name members_str local_pkg_action_information} {
    global ping
    
    global PKG_FAILED_RETRIEVING_INFO    PKG_DOESNT_EXIST    PKG_INVALID_FOR_ACTION    PKG_KEY_MISMATCH    
    
    global  PKG_INFORMATION_NAME PKG_INFORMATION_DISPLAY_NAME    PKG_INFORMATION_STATE  \
            PKG_INFORMATION_KEY  PKG_INFORMATION_IS_VALID_ACTION PKG_INFORMATION_QUERY_OUTPUT
    
    # Set argument as array
    array set local_pkg_action_information_arr $local_pkg_action_information
    
    # Variables initialization
    set pkg_action_is_valid_on_local_member     0
    
    if {[isDebugMode]} {puts "[fn](): #$action#,#$package_name#,#$members_str#"}
    
    if { $package_name == $ping(NA_STR)} {
        putsd "[fn](): package_name=$ping(NA_STR)"
        # Package index doesn't exist or invalid for current action
        
        # Clear all member's status
        set_all_members_status ""
        print_members_status
        writeLog $ping(LOG_ERROR) "[fn]" "Package doesn't exist.\n"
        
        return -1
    }

    
    # puts "members_str=$members_str"
     
    # Execute CDK Query for package information by <Package Name>
    verify_pkg_action_get_pkg_information    $package_name $action $members_str $local_pkg_action_information   member_db_array
    putsd "[fn](): after verify_pkg_action_get_pkg_information()"

    set error_mask [ verify_pkg_action_verify_query  $action [array get member_db_array]    members_str_without_pkg ]
    putsd "[fn](): after verify_pkg_action_verify_query()"
    
    # Display verification result #
    print_members_status
    
    if { [expr ($error_mask & $PKG_KEY_MISMATCH) ] != 0 } {
        puts "\nThe package is not identical on all members.\n"
    }
 
    
    # Try to import package on the required members before aborting installation
    if {[should_import_pkg_on_remote_members  $package_name  $action  $error_mask  [array get member_db_array]  local_pkg_action_information_arr] } {
        return [import_pkg_on_required_members $members_str_without_pkg  [array get local_pkg_action_information_arr] ]
    }        
    
    set verification_result    [expr (($error_mask==0)?0:-1)] 
    putsd "[fn](): verification_result=$verification_result"
    
    
    return  $verification_result
}
   
   
   
proc should_import_pkg_on_remote_members {package_name   action error_mask  member_db_array_data   local_pkg_action_information_array} {
    putsd "[fn]()"
    global ping
    global PKG_DOESNT_EXIST
    global PKG_INFORMATION_NAME
    
    upvar     $local_pkg_action_information_array loc_local_pkg_action_information_array     
    array set member_db_array $member_db_array_data
    
    set FAILED_TO_RETRIEVE_LOCAL_PKG_INFO  -1
    set DONT_IMPORT_PKG_ON_REMOTE_MEMEBERS  0
    set IMPORT_PKG_ON_REMOTE_MEMEBERS       1
    
    
    if { $error_mask == $PKG_DOESNT_EXIST && 
         $action == $ping(ACTION_INSTALL)   } {
            
        set pkg_should_be_install_on_local_member [info exist member_db_array($ping(local_member_id))]
        # putsd "should_import_pkg_on_remote_members(): pkg_should_be_install_on_local_member=$pkg_should_be_install_on_local_member"
        # putsd "should_import_pkg_on_remote_members(): extract val: [extract_v_from_DAClient_output $member_db_array($ping(local_member_id))    $ping(BIND_VALIDATE_PACKAGE_ACTION_VALIDATION)]"
        # putsd "should_import_pkg_on_remote_members(): local_member_id: $member_db_array($ping(local_member_id))"
        # putsd "should_import_pkg_on_remote_members(): ping(VALID_PKG_ACTION)=$ping(VALID_PKG_ACTION)"
        
        if { $pkg_should_be_install_on_local_member } {
            set pkg_action_is_valid_for_local_member [expr [extract_v_from_DAClient_output $member_db_array($ping(local_member_id))    $ping(BIND_VALIDATE_PACKAGE_ACTION_VALIDATION)] == $ping(VALID_PKG_ACTION)]
            
            if {$pkg_action_is_valid_for_local_member} {
                return $IMPORT_PKG_ON_REMOTE_MEMEBERS
            } 
        } else {
        
            set is_pkg_info_of_local_member_exist  [info exist loc_local_pkg_action_information_array($PKG_INFORMATION_NAME)]

            if {!$is_pkg_info_of_local_member_exist} {
                if { [get_local_pkg_action_information    $action $package_name loc_local_pkg_action_information_array 0] } {
                    return $FAILED_TO_RETRIEVE_LOCAL_PKG_INFO
                }
            }
             
            # Check if package exist on local member
            set pkg_name $loc_local_pkg_action_information_array($PKG_INFORMATION_NAME)
            if { ![string equal $pkg_name "" ] && 
                 ![string equal $pkg_name $ping(NA_STR)]  } {
                return $IMPORT_PKG_ON_REMOTE_MEMEBERS
            }
        }
    }

    return $DONT_IMPORT_PKG_ON_REMOTE_MEMEBERS
}



proc get_verified_action_status {action} {
    global ping
    
    if {$action == $ping(ACTION_VERIFY)} {
        return "Package is ready for verification"
    } elseif {$action == $ping(ACTION_DOWNLOAD_AND_INSTALL)} {
        return "Package is ready for download and install"
    } elseif {$action == $ping(ACTION_INSTALL)} {
        return "Package is ready for installation"
    } elseif {$action == $ping(ACTION_REINSTALL)} {
        return "Package is ready for reinstallation"
    } elseif {$action == $ping(ACTION_UNINSTALL)} {
        return "Package is ready for uninstallation"
    } else   {
        return "Package is ready for $action"
    }

}



# Import package to remote members in case package exist only on local member
proc import_pkg_on_required_members {members_str local_pkg_action_information} {
    global ping
    global SKIP_AUDIT
    global PKG_INFORMATION_NAME   PKG_INFORMATION_DISPLAY_NAME   PKG_INFORMATION_STATE  \
           PKG_INFORMATION_KEY  PKG_INFORMATION_IS_VALID_ACTION PKG_INFORMATION_QUERY_OUTPUT
    
    # Set as an array
    array set local_pkg_action_information_array $local_pkg_action_information
    
    # Extract local package information
    set pkg_file_name    [file tail $local_pkg_action_information_array($PKG_INFORMATION_NAME)]
    set pkg_key          $local_pkg_action_information_array($PKG_INFORMATION_KEY)
    set pkg_display_name $local_pkg_action_information_array($PKG_INFORMATION_DISPLAY_NAME)
    
    # Suggest the user to import package
    
    # exec tput cup [$console_lines-5] 0 >/dev/tty
    # screen_end  5
    puts "\nPackage doesn't exist on several members,"
    puts "\nWould you like to import and install the package on the remote members? ($pkg_file_name)"
    puts "(Y - yes, any other key - no)"
    
    if {![isTestingMode] && $SKIP_AUDIT==0} {
        get_user_input user_input
    } else {
        set user_input "y"
    }

    # Handle user input
    if { $user_input != "y" } {
        # User has declined
        return -1
    } 

    # set package path
    set pkg_path "$ping(REPO_PATH)/$pkg_key/$pkg_file_name"
 
    # Verify that package exist in the local member repository
    putsd "[fn](): Verifying that package exist in the local member repository"
    if {![file exists $pkg_path]} {
        putsd "[fn](): file doesn't exist"
        writeLog $ping(LOG_ERROR) "[fn]" "File doesn't exist($pkg_path)"
        return -1
    }

    # Perform package import
    set res [perform_installer_action $ping(ACTION_IMPORT) \
             $pkg_path $pkg_display_name $members_str]
    return $res
}

# Execute 'DAClient' on cluster members according to <member_ids_str>
#
# Return the execution output per cluster member into  <output_array>
#     
proc DAClient_exec { member_ids_str action pkg_str pkg_display_name output_array} {
    global ping
    upvar $output_array     loc_output_array

    pvars member_ids_str action pkg_str
    
    # For "installer agent install <pkg>"
    if {[string compare $action $ping(ACTION_INSTALL_DA)] == 0} {
        # File does not exist
        if { [file exists $pkg_str] == 0 } {
            puts "The file you provided does not exist. Please provide an absolute path to the file.\n"
            return -1
        }
        # Copy pkg to other members
        if {[transfer_package $pkg_str $member_ids_str]} {
            return -1 ; # Couldn't copy package to other members
        } 
        
        # Copy was successful
        # Deployment agent was uploaded to IMPORT_PATH so update pkg_str 
        set pkg_basename [file tail $pkg_str]
        set pkg_str $ping(IMPORT_PATH)/$pkg_basename
    }
    
    # Convert <$ping(ACTION_CHECK_FOR_UPDATES)> into the appropriate DAClient argument for "CHECK_FOR_UPDATES"
    if {$action == $ping(ACTION_CHECK_FOR_UPDATES)} {
        set action "update"
    }
      
    if {$action == $ping(ACTION_UNINSTALL) && $ping(is_uninstall_completely) == 1} {
        # Convert the action "uninstall" to "uninstall_completelly"
        set action "uninstall_completely"
    }
    
    pvars action ping(is_uninstall_completely)
    
    
    if {$pkg_display_name != ""} {
        set cmd "$ping(WRAPPER_PATH) $ping(DACLIENT_WRAPPER_FLAG) $action $pkg_str '$pkg_display_name'"
    } else {
        set cmd "$ping(WRAPPER_PATH) $ping(DACLIENT_WRAPPER_FLAG) $action $pkg_str"
    }
    
    
    if {[gclish_exec_remote_members $member_ids_str  $cmd  loc_output_array]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to execute action." 0
        return -1
    }
    return 0
}

# Execute 'dbget' on cluster members according to <member_ids_str> in order to retrieve Gaia DB values / invoke CDK action
#
# Return the execution output per cluster member into  <output_array>
#
proc dbget_exec { member_ids_str key output_array} {
    global ping
    upvar $output_array     loc_output_array
    
    set cmd "$ping(DBGET_BINARY) \'$key\' "
    
    if {[gclish_exec_remote_members $member_ids_str $cmd loc_output_array]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to get Gaia DB value." 0
        return -1
    }
    
    return 0
}

# Execute 'dbset' on cluster members according to <member_ids_str> in order to set/clear Gaia DB values
#
# Return the execution output per cluster member into  <output_array>
#
proc dbset_exec { member_ids_str key value output_array} {
    global ping
    upvar $output_array     loc_output_array

    set cmd "$ping(DBSET_BINARY) $key $value"
    
    if {[gclish_exec_remote_members $member_ids_str $cmd loc_output_array]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to set Gaia DB value"
        return -1
    }
    
    return 0
}


proc pkg_import {action pkg_str pkg_display_name members_str} { 
    global ping
    
    # Transfer package file to members according to $members_str
    if {[transfer_package $pkg_str $members_str] } {
        # Failed to transfer package to some of the members
        return -1
    }

    set filename [file tail $pkg_str]
    # package import command: DAClient import <file name without path>
    # and the package should be located at 
    set res [invoke_action_ex $action $members_str $filename $pkg_display_name]
    return $res
}


# Description: Returns whther or not we display DA progress for <action>
#              (e.g. no need to display 'the progress' of disabling DA service)
#
# Input:
#       DAClient action
#
# Return: 
#       1 - We should display the action progress
#       0 - Otherwise
#
proc should_display_progress {members_str action} {
    global ping members_status SKIP_PROGRESS_DISPLAY
    
    if {$members_str == $ping(local_member_id) && 
        $members_status($ping(local_member_id)) == $ping(MSG_DA_ISNT_RUNNING)} {
        # In case we are trying to perform some action, only on the local member
        # but DAClient returned a message which says that DAService is down, then
        # don't wait for progress
        return 0
    }
    
    if { [lsearch $SKIP_PROGRESS_DISPLAY $action] != -1 } {
        # Item found, skip progress display
        return 0
    }
    return 1
}

proc invoke_action_ex {action members_str pkg_str pkg_display_name} {
    global ping
    pvars action members_str pkg_str
    

    if {[clear_db_value_by_action $action $members_str]} {
        # Failed to clear action values from Gaia DB
        return -1
    }
    putsd "[fn](): After clear_db_value_by_action"    
    
    # Initialize execution output array
    array set member_result_array {}
    
    if {[DAClient_exec $members_str $action $pkg_str $pkg_display_name member_result_array]} {
        # Failed to execute dbget on members
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to invoke action."
        return -1
    }
    putsd "[fn](): After dbget_exec"    
    
    # Set and Display dbget output #
    foreach member_id [array names member_result_array] {
        set output [reformat_DAClient_exec_output $action $member_result_array($member_id)]
        # Update member's status
        set_member_status $member_id $output
    }
    putsd "[fn](): After Set and Display dbget output"    
    
    
    print_members_status
    
    after $ping(DELAYED_DISPLAY)
    
    set res 0
    
    if {[should_display_progress $members_str $action]} {
        set res [wait_for_result $action $pkg_str $members_str]
    }

    # Success/failure message for "installer agent install"
    if { $action == $ping(ACTION_INSTALL_DA) && $res == 0 } {
        puts "Successfully installed new Deployment Agent. Leaving CLISH."
    }

    return $res
}

# Invoke DAClient action
# Note: pkg_str may contain only pkg_name or also pkg custom display name, e.g. "pkg.tgz my custom package"
proc invoke_action {action members_str pkg_str} {
    set pkg_display_name ""
    return [invoke_action_ex $action $members_str $pkg_str $pkg_display_name]
}

# Reformat CDK invocation output after execution 'dbset <key>' command
proc reformat_cdk_invocation_output {execution_output} {
    set c_output $execution_output

    # Remove "" character [CLISH_TAG_VALUE_SEPERATOR]
    regsub -all "\x80" $c_output " " c_output
    
    # Remove <Delete> character  [CLISH_LINE_SEPERATOR]
    regsub -all "\x7F" $c_output ""  c_output
    
    # Return result
    return $c_output
} 

proc reformat_DAClient_exec_output {action exec_output} {
    global ping
    global DISPLAY_REAL_DACLIENT_OUTPUT
    
    if {$DISPLAY_REAL_DACLIENT_OUTPUT} {
        return $exec_output
    }
    
    # According to basic_common.h
    set sig_h "Signalled to service, type="
    if {[string match -nocase "*${sig_h}21*" $exec_output]} {
        return "Disable singal was sent."
    } elseif {[string match -nocase "*DAService is now disabled*" $exec_output]} {
        return "DAService is disabled."
    } elseif {[string match -nocase "*DAService is disabled. Use \"DAClient enable\"*" $exec_output]} {
        return "DAService is disabled. Use \"installer agent enable\" command"
    } elseif {[string match -nocase "*${sig_h}1" $exec_output]} {
        return "Stop singal was sent."
    } elseif {[string match -nocase "*${sig_h}5,*" $exec_output]} {
        return "Verifying package..."
    } elseif {[string match -nocase "*${sig_h}6,*" $exec_output]} {
        return "Importing package..."
    } elseif {[string match -nocase "*${sig_h}8,*" $exec_output]} {
        return "Deleting package..."
    } elseif {[string match -nocase "*${sig_h}9,*" $exec_output]} {
        return $ping(MSG_SEARCH_IN_CLOUD)
    } elseif {[string match -nocase "*${sig_h}13,*" $exec_output]} {
        return "Downloading package..."
    } elseif {[string match -nocase "*${sig_h}14,*" $exec_output]} {
        return "Installing package..."
    } elseif {[string match -nocase "*${sig_h}15,*" $exec_output]} {
        return "Uninstalling package..."
    } elseif {[string match -nocase "*${sig_h}16,*" $exec_output]} {
        return "Upgrading..."
    }
	
    
    if { [string match -nocase "*Process DASERVICE is alive, process won't be started again*" $exec_output] == 1 } {
        return "DAService is already up."
    } elseif { [string match -nocase "*DASERVICE started successfully*" $exec_output] == 1 } {
        return "DAService started successfully."
    }
    
    # note, blocking this will allow us to return actual DA output.
    #if { [string match -nocase "*Signalled to service*" $exec_output] == 0 &&
    #     [string match -nocase "*Enabling DAService*" $exec_output] == 0 } {
    #    putsd "reformat_DAClient_exec_output(): Failed to signal DAClient"
    #    set e_msg "Action failed ($action)"
    #    writeLog $ping(LOG_ERROR) "[fn]" $e_msg
    #    return $e_msg
    #}
    
    # Return result
    return $exec_output
    set res [action_to_status $action]
        
    return $res
}

# TODO: Elaborate regarding transfer failed per member ID
proc transfer_package {pkg_str members_str} {
    putsd "transfer_package(): pkg_str=$pkg_str"
    global ping
    
    # Extract the filename from the file path
    set filename [file tail $pkg_str]

    if {$filename == $pkg_str} {
        # Handle the case where DAclish_import.c->get_import_bind()
        # Already copied the package into <ping(IMPORT_PATH)> (e.g. import local / import ftp)
        set pkg_src_path    "${ping(IMPORT_PATH)}/$pkg_str"
    } else {
        # Handle other cases, such as import from local repository
        # To other members, when trying to install package to several members,
        # but other members don't have the package
        set pkg_src_path    "$pkg_str"
    }
    
    
    # Note to self: after my new changes to 'DAclish.c -> action_import', src path is always
    # ${ping(IMPORT_PATH)}/$filename
    # Except!!! when try to install package on all members, when pkg is missing on some of them
    
    
    # set pkg_src_path $pkg_str
    
    # # Set package path according to import "local"/"ftp" GCLISH arguments
    # if {[string match -nocase "*ftp*"  $ping(gclish_cmd)]} {
        # set filename            [file tail $pkg_str]
        # set pkg_src_path    "${ping(IMPORT_PATH)}/$filename"
    # }
    
    if {![file exists $pkg_src_path]} {
        writeLog $ping(LOG_ERROR) "[fn]" "File doesn't exist($pkg_src_path)"
        return -1
    }

    set_member_status $members_str "Verifying disk space..."
    print_members_status
    after $ping(DELAYED_DISPLAY)
    
    # Verify enough disk space is available
    if {[verify_remote_disk_space $members_str [file size $pkg_src_path]]} {
        return -1 ;# function call prints proper error message
    }
    
    set dst_file_path "${ping(IMPORT_PATH)}/[file tail $pkg_src_path]"
    
    set_member_status $members_str "Transferring package..."
    print_members_status
    
    # Transfer package to other members, verify transfer and display the result
    if {[cp2blades_exec $members_str $pkg_src_path $dst_file_path] || 
        [cp2blades_exec_verify_and_display_result $members_str $dst_file_path]} {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to transfer package to several members, Import was aborted.\n" 
        return -1
    }
    
    # Sleep
    after $ping(DELAYED_DISPLAY) ; # So that the user can review the result before display is changed
    return 0
}

 
# Verify that the file exist on remote members and display verification result per member
proc cp2blades_exec_verify_and_display_result {member_str dst_file_path} {
    global ping
    
    set VERIFICATION_FAILED_OUTPUT "No such file or directory"
    set verification_cmd           "ls -l $dst_file_path"

    # Execute file verification #
    if {[gclish_exec_remote_members $member_str $verification_cmd output_array]} {
        # Failed to verify file transfer
        writeLog $ping(LOG_ERROR) "[fn]" "Transfer verification has failed." 1
        writeLog $ping(LOG_DEBUG) "[fn]" "Verification output: \n[array get output_array]"
        return -1
    }
    
    # Verify execution output(that file exist) and display result #
    set result    0
    set member_id_lst [split $member_str ","]
    
    foreach member_id $member_id_lst {
        if { [regexp -nocase -- $VERIFICATION_FAILED_OUTPUT  $output_array($member_id)]   || 
            ![regexp -nocase -- $dst_file_path  $output_array($member_id)] }  {
                set_member_status $member_id "Package transfer has failed" 
                set result -1
        } else {
                set_member_status $member_id "Package was transferred successfully" 
        }
    }

    #Display result
    print_members_status
    
    if {$result == -1} {
        # Transfer verification has failed, Log verification output
        writeLog $ping(LOG_DEBUG) "[fn]" "Verification result: \n[array get output_array]"
        return -1
    }
    
    return 0
}



# Go to the beginning of the screen
proc screen_start {} {
    global ping

    screen_set_cursor 0
    
    # if {![isDebugMode] && ![isTestingMode]} {
        # puts \x1b\[H  ; 
        # exec tput cup 0 0 >/dev/tty
        
    # }
}

proc screen_end {{lines_before_end 0}} {
    global ping
    global console_lines
    
    if {![isDebugMode] && ![isTestingMode]} {
        # puts \x1b\[H  ;  Go to First line
        # exec tput cup [expr $console_lines-$lines_before_end] 0 >/dev/tty
    }
}

  
# Go to the beginning of the screen and clear it
proc screen_reset {} {
    if {![isDebugMode] && ![isTestingMode]} {
        screen_start  ; #Go to the beginning of the screen
    }
}

proc screen_reset_with_blank_str {} {
    if {![isDebugMode] && ![isTestingMode]} {
        screen_start  ; #Go to the beginning of the screen
        screen_print_blank_str
        screen_start  ; #Go to the beginning of the screen
    }
}


proc screen_print_blank_str {} {
    global console_columns
    global console_lines

    set blank_str ""
    
    if {![isDebugMode] && ![isTestingMode]} {
        # Build <blank string>
        for {set line_idx 0} {$line_idx <= $console_lines} {incr line_idx} {
            if {$blank_str != ""} {append blank_str "\n"}
            append blank_str [format "%-*s" $console_columns ""]
        }
        # Print <blank string>
        puts $blank_str
    }
}





# New black screen in the terminal
proc screen_clear {} {
    if {![isDebugMode] && ![isTestingMode] } {
        puts \x1b\[2J ; 
    }
}

proc print_time_and_date_header {} {
    global console_columns
    
    set header ""
    
    append header [format "%-*s" $console_columns "[epoch2human [clock seconds]]" ]
    append header [get_empty_line_str]
    append header [format "%-*s" $console_columns "Update Service Engine"]
    append header [get_empty_line_str]
    
    puts $header
}

# Return consecutive spaces in the length of the terminal width
# (In order to erase the existing characters on the SSH window)
proc get_empty_line_str {} {
    global console_columns
    
    return [format "%-*s" $console_columns ""]
}

proc screen_update_console_size {} {
    global console_lines   console_columns
    
    set prev_console_lines          $console_lines
    set prev_console_columns    $console_columns

    # Get current terminal size
    if {![catch {exec stty -a} err] && [regexp {rows (\d+); columns (\d+)} $err match rows cols]} {
        set console_lines [expr $rows - 1]
        set console_columns $cols      
    }
    
    # Reset screen completely in case terminal window size has changed
    if {$prev_console_lines         != $console_lines     || 
        $prev_console_columns != $console_columns } {
        screen_reset_with_blank_str
    }
    
}
 
# Print current members status
proc print_members_status {} {
    global ping  
    global member_ids   members_status  
    global PRINT_EXCLUTION_LIST
    global header_lines table_lines
    global TABLE_COLUMN_WIDTH_MEMBER_ID TABLE_COLUMN_WIDTH_MEMBER_STATUS
    
    # Initialize the number of table lines
    set table_lines 0
    
    set width_list [list  $TABLE_COLUMN_WIDTH_MEMBER_ID   $TABLE_COLUMN_WIDTH_MEMBER_STATUS]  ;#  Total width = 77
    
    screen_update_console_size
      
    # Abort print_members_status in case members status didn't change and currently in testing mode
    if { $ping(members_status_changed) == 0 && [isTestingMode]} {
        # screen_start
        # print_time_and_date_header
        # screen_end
        return
    }
  
    # Go to the beginning of the screen
    screen_start
    # screen_reset_with_blank_str
    print_time_and_date_header
     
    # Print headers
    set table_lines [expr $table_lines + 3]
    puts_table      "header_line" $width_list
    puts_table      "body"        $width_list [list "Member ID" "Status"]
    puts_table      "line"        $width_list
    

    # Print member's status
    foreach {  member_id }  $member_ids {
        
        if {$member_id==$ping(local_member_id)} {
            # Local member
            set member_id_str "$member_id (local)"
        } else {
            # Remote member
            set member_id_str "$member_id"
        }
  
        set status $members_status($member_id)
        
        puts_table_body_    $width_list  [list  $member_id_str  $status]  $PRINT_EXCLUTION_LIST
        # puts_table_body_    $width_list $row  [list]
        
        # Print line separator between rows
        # puts_table             "line" $width_list
        # set table_lines        [expr $table_lines + 1]
    }
    
    puts_table             "line" $width_list
    set table_lines        [expr $table_lines + 1]
    
    # puts ""
    
    # puts $table_lines
    
    screen_clear_after_table
    
    # Go to last line
    # screen_end

    # flush stdout
    
    
    # Initialize  'members_status_changed' (Since last printing) to 0
    set ping(members_status_changed) 0
}

proc screen_clear_after_table {} {
    global header_lines  table_lines
    global console_columns
    global console_lines
    
    set blank_str ""
    
    if {![isDebugMode] && ![isTestingMode]} {
        set table_and_header_size [expr $header_lines + $table_lines]
        
        # Print blank string
        for {set line_idx 0} {$line_idx < [expr $console_lines - $table_and_header_size] } {incr line_idx} {
            append blank_str [format "%-*s" $console_columns ""]
        }
        
        puts $blank_str

        screen_set_cursor [expr $table_and_header_size + 1]
    }
}

proc screen_set_cursor {line_index} {
    if {![isDebugMode] && ![isTestingMode]} {
        exec tput cup $line_index 0 >/dev/tty
    }
}


# Will remove long white spaces...
proc seperate_cell_to_lines_by_width {width cell} {
    upvar $cell loc_cell
    
    set new_cell    ""
    
    set cur_line_length 0
    
    foreach word $loc_cell {

        if {$new_cell != ""} {
            set word  " $word"
        }
    
        set word_length [string length $word]
        
        if {$cur_line_length == 0 || [expr $cur_line_length + $word_length] <= $width} {
            append new_cell $word
            set cur_line_length [expr $cur_line_length + $word_length]
        } else {
            set word [string trim $word]
            append new_cell "\n$word"
            set cur_line_length $word_length
        }
    }
    
    set loc_cell $new_cell
    
    return 0
}

proc remove_excluded_lines {str exclusion_list} {
    set lines [split $str "\n"]
    set output ""
    
    foreach current_line $lines {
        set include_line 1

        foreach exclusion_item $exclusion_list {
            if {[string match $exclusion_item  $current_line]} {
                # Don't print current line
#                puts "MATCH ON REMOVE=$current_line"  
                set include_line 0
                break
            }
        }
        
        if {$include_line} {
#            puts "NO MATCH ON REMOVE=$current_line++"
            append output "$current_line\n"
        }
    }
    
    # Remove last "\n"
    set output [string range $output 0 end-1]

    #    puts "new output:"
#    puts "$output"
#    puts "----------------------------!!!"
    
    return $output
}
 
# puts_table "body" extension. (Support multiple lines in one cell)
proc puts_table_body_ { width_list content exclusion_list} { ;# (content is a table's row)
    global table_lines
    
    set table [list]
     
    # Split each cell to lines and add that list to "table"

    set content_length [llength $content]
    
    # Split each cell in the row into list of lines, and append it to 'table'
    for {set column_index 0} {$column_index<[llength $width_list]} {incr column_index} {
        set cur_cell             [lindex $content $column_index        ]
        set cur_cell_width     [lindex $width_list $column_index    ]
        
        # Remove "\n\\\n" (from /config/active output)
        regsub -all {\n\\\n} $cur_cell "\n" cur_cell

        # Remove excluded lines
        set cur_cell [remove_excluded_lines $cur_cell $exclusion_list]
        
        if {[string length $cur_cell] > $cur_cell_width} {
            # Split cell to lines by it's width to avoid breaking words 
            seperate_cell_to_lines_by_width $cur_cell_width cur_cell
        }
        
        # Split to lines
        set cell_lines [split $cur_cell "\n"]
        lappend table $cell_lines
    }
    
    # Print row after row using "puts_table"
    set row_index 0
    
    while 1 {
        set skipped_line 0
        set found_value  0 ; # <Found value in the current row>
        set row         [list] 
        
        # Build a row from the columns of the generated 'table' for puts_table()
        foreach {cell} $table {
            if { [llength $cell] > $row_index } {
                # The cell in the current column exists
                
                set current_line [lindex $cell $row_index]
                # Add current line to the row
                lappend row $current_line
                set found_value 1
                
            } else {
                # The cell in the current column doens't exist
                lappend row "" 
            }
        }
        
        
        # Print row in case found value
        if { $found_value } {
            # Print row
            puts_table "body" $width_list $row
            set table_lines        [expr $table_lines + 1]
            
        } elseif {!$skipped_line} {
            # Finish printing in case no line was skipped
            break
        }
        
        incr row_index
    }
}


proc strip_HTML_tags {str} {
    # remove '\n\\\n'
    regsub -all {\n\\\\\n} $str "\n" str
    # remove '\'
    regsub -all {\\} $str "" str
    # remove '<*>'
    regsub -all {<[^>]+>} $str "" str
    # replace '::' with ':'
    regsub -all {::} $str ":" str
    
    set str [string trim $str]
    
    return $str 
}
 
 
# Description: Returns the member_ids which are current "Installing new agent version" into self_update_in_progress_member_ids array
#
# Input:
#        members_Gaia_DB_content_array per member_id  (e.g. the output of the following command: # dbget -rv installer   per member_id)
#
# Return: 
#        <self_update_in_progress_member_ids>    - Members which are currently "Installing new agent version"
#
proc get_member_ids_in_self_update_state {members_db_array_data    self_update_in_progress_member_ids} {
    global ping
    global KEY_SELF_UPDATE_IN_PROGRESS   STATE_SELF_UPDATE_IN_PROGRESS
    
    upvar $self_update_in_progress_member_ids     loc_self_update_in_progress_member_ids
    # Initialize array
    if {[info exist loc_self_update_in_progress_member_ids]}     {unset loc_self_update_in_progress_member_ids}
    array set loc_self_update_in_progress_member_ids     {}
    
    # Convert array_data into array
    array set  members_db_array                         $members_db_array_data
            
    # Check if any member is currently "Installing new agent version" according to it's DB_data and add it to the self_update_in_progress_member_ids array
    foreach member_id [array names members_db_array] {
        # Extract key's value from DB data
        set self_update_value [extract_v_from_k_v $members_db_array($member_id) $KEY_SELF_UPDATE_IN_PROGRESS]
        # Verify key's value
        if { $self_update_value == $STATE_SELF_UPDATE_IN_PROGRESS } {
              # Add member_id to self_update_in_progress_member_ids array (Installing new agent version).
              set loc_self_update_in_progress_member_ids($member_id) 1
        }
    }    
}


 
proc wait_for_result_process_db_query {action   members_str   members_db_content_array_data   progress_package_name pkg_str} {
    global ping  
    global members_status
    global DISPLAY_PROGRESS_EXCULSION_LIST
    global KEY_CLISH_PROGRESS_PKG    KEY_CLISH_PROGRESS_RESULT
    global ALL_MEMBERS_ACTIONS_WERE_COMPLETED
    global KEY_AGENT_INSTALL_MSG
    
    
    upvar $members_str           loc_members_str
    upvar $progress_package_name loc_progress_package_name
    
    set MSG_AGENT_ERROR             "Error: Deployment Agent"
    set MSG_CONCURRENT_ACTION_ERROR "Error: another installation is in progress"
    set MSG_GENERAL_ERROR           "Error"
    
    array set members_db_content_array $members_db_content_array_data
    
    # Get members which are currently "Installing new agent version" #
    get_member_ids_in_self_update_state $members_db_content_array_data  self_update_in_progress_member_ids
    if {[string compare $action $ping(ACTION_INSTALL_DA)] != 0 } {
        foreach member_id [array names self_update_in_progress_member_ids] {
            set_member_status $member_id "Operation was aborted.\nDeployment agent is installing new agent version."
        }
    }
    # print_members_status
     
     
    set found_member_with_no_action_result 0
    set loc_progress_package_name          ""  ; # The installer service sets the package name which bound to the current action in GAIA DB
    array set member_ids_action_completed  {}  ; # Member ids which completed with action_result

    # Check if all members action was completed and has "action_result" value in GAIA DB #
    # Check if any member progress value exist
     
    foreach member_id [array names members_db_content_array] {

        # Get the package name which is bound to the current installer action
        if {$loc_progress_package_name == ""} {
            set loc_progress_package_name [extract_v_from_k_v $members_db_content_array($member_id) $KEY_CLISH_PROGRESS_PKG]
        }
        
        # Get action_result value
        if {$action == $ping(ACTION_INSTALL_DA)} {
            # In the special case of DA installation, use $KEY_AGENT_INSTALL_MSG instead of $KEY_CLISH_PROGRESS_RESULT
            set package_result [extract_v_from_k_v $members_db_content_array($member_id) $KEY_AGENT_INSTALL_MSG]
        } else {
            set package_result [extract_v_from_k_v $members_db_content_array($member_id) $KEY_CLISH_PROGRESS_RESULT]
        }

        set is_member_in_self_update_state [info exists self_update_in_progress_member_ids($member_id)]
        
        
        if { !$is_member_in_self_update_state || $action == $ping(ACTION_INSTALL_DA) } {
            
            if {[string first $ping(MSG_AGENT_IS_UP_TO_DATE) $members_status($member_id)] != -1   ||   
                [string first $MSG_CONCURRENT_ACTION_ERROR   $members_status($member_id)] != -1   ||   
                [string first $MSG_GENERAL_ERROR             $members_status($member_id)] != -1   ||   
                [string first $MSG_AGENT_ERROR               $members_status($member_id)] != -1 } {
                
                # CPUSE agent action was completed (Possibly due to an error)
                set member_ids_action_completed($member_id) 1
                continue
            }
        
            # Check if current member action was completed (beside exception members, which are currently updating their agent)
            if { $package_result == "" } {
               # Member action wasn't completed
               set found_member_with_no_action_result 1
            } elseif { [regexp -nocase -- ".*new package.*" $package_result] && $action != "check-for-updates"} {
                # Member action wasn't completed (when action is not check for updates) , but check for updates was used during the action
                set found_member_with_no_action_result 1
                set loc_progress_package_name $pkg_str 		
                set package_result "" 
            } else { 
               # member action was completed, set member_status with <action result value>
               set_member_status $member_id [strip_HTML_tags $package_result]
               # Add member_id to completed_action array
               set member_ids_action_completed($member_id) 1
               # Remove finished member_id from member_str so it wouldn't be queried again
               set loc_members_str [remove_member_id $loc_members_str $member_id]
               # Log action_result
               writeLog  $ping(LOG_INFO) "[fn]" "$member_id: action completed, $members_status($member_id)" 0
            }
        }
    }
    
    
    if { $found_member_with_no_action_result == 0 } {
        # All member's actions were completed with "action_result" / with DBGET execution Error
        # print_members_status
        return $ALL_MEMBERS_ACTIONS_WERE_COMPLETED
    }
    
    # Skip progress display #
    if {[lsearch $DISPLAY_PROGRESS_EXCULSION_LIST $action] != -1} {
        return 0
    }
    
    
    if { $loc_progress_package_name != "" } {
        # Display action progress #
        
        # Outputs for debug mode
        if [isDebugMode] {
            puts "refresh_action_progress"
            puts "self_update_in_progress_member_ids=[array get self_update_in_progress_member_ids]"
            puts "member_ids_action_completed=[array get member_ids_action_completed]"
        }
        
        refresh_action_progress $action $loc_progress_package_name members_db_content_array self_update_in_progress_member_ids member_ids_action_completed
    }
      
    # print_members_status
    
    return 0
}

proc remove_inactive_member_ids_and_update_status {members_str} {
    global ping    members_status
    upvar $members_str loc_members_str
    
    # Save original members_str
    set member_str_all $loc_members_str
    
    set MSG_INACTIVE_MEMBER " (Down)"
    
    set res_remove_inactive_member_ids    [remove_inactive_member_ids loc_members_str]
    
    if {$res_remove_inactive_member_ids < 0 } {
        # Failed to remove_inactive_member_ids
        return $res_remove_inactive_member_ids
    }
    
    # Update inactive member's status
    set members_all_lst    [split $member_str_all  ","]
    set members_active_lst [split $loc_members_str ","]
    
    foreach member_id $members_all_lst {
        if {[lsearch $members_active_lst $member_id] == -1} {
            # Found inactive member
            set member_status $members_status($member_id)
            if {[string first $MSG_INACTIVE_MEMBER $member_status] == -1 } {
                # Update member status as inactive
                append member_status $MSG_INACTIVE_MEMBER
                set_member_status $member_id  $member_status
                print_members_status
                # Log inactive member
                writeLog  $ping(LOG_INFO) "[fn]" "$member_id is down" 0
            }
        }
    }
    
    return $res_remove_inactive_member_ids
}


proc remove_inactive_member_ids {members_str} {
    global ping
    global ALL_MEMBERS_ARE_ACTIVE   SEVERAL_MEMBERS_ARE_INACTIVE
    upvar $members_str loc_members_str
    
    set communicating_members_str ""
    
    set ret_val  $ALL_MEMBERS_ARE_ACTIVE
    
    # Get communicating member ids
    get_communicating_members_lst $loc_members_str communicating_members
    
    # Convert to formal blade ids
    set communicating_members [convert_to_formal_member_ids $communicating_members]
    
    set members_all_lst [split $loc_members_str ","]
    
    # Append only communicating member_ids from <loc_members_str> to <communicating_members_str>
    foreach member_id $members_all_lst {
        if {[lsearch $communicating_members $member_id] != -1} {
            # Current member_id is active
            if {$communicating_members_str != ""} {  append communicating_members_str "," }
            append communicating_members_str $member_id
            
        } else {
            # Found inactive member
            set ret_val  $SEVERAL_MEMBERS_ARE_INACTIVE
        }
    }
    
    # Update loc_members_str with loc_members_str without inactive members
    set loc_members_str $communicating_members_str
    
    return $ret_val
}

# Get cluster members installer values from DB into members_db_array
proc wait_for_result_pull_db {action members_str members_db_array} {
    global ping
    global NO_ACTIVE_MEMBERS   ALL_MEMBERS_ARE_ACTIVE

    upvar $members_db_array loc_members_db_array
    
    set DB_QUERY_CMD "$ping(DBGET_BINARY) -rv $ping(BIND_PING)"
    set MAX_QUERY_ATTEMPTS       8
    set TIME_BETWEEN_ATTEMPTS  500
    set FAILED_TO_PULL_DB_DATA  -1
    
    set members_str_original  $members_str
    
    for {set current_attempt 1} {$current_attempt <= $MAX_QUERY_ATTEMPTS} {incr current_attempt} {

        if {$current_attempt != 1} {
            # Sleep between attempts
            after $TIME_BETWEEN_ATTEMPTS
        }
    
        set members_str $members_str_original
        set remove_inactive_res [remove_inactive_member_ids_and_update_status  members_str]
        if {$remove_inactive_res < 0} {
            writeLog $ping(LOG_DEBUG) "[fn]" \
                     "Failed to remove inactive members from members_str(attempt $current_attempt)"
            # Try to query DB again
            continue
        }
        
        if {$members_str == ""} {
            # There are no active members to pull result from
            return $NO_ACTIVE_MEMBERS
        }
        
        ## Pull GAIA DB data from members ##
        if {[gclish_exec_remote_members $members_str $DB_QUERY_CMD loc_members_db_array] } {
            writeLog $ping(LOG_DEBUG)    "[fn]" "Failed to read DB values(attempt $current_attempt) : [array get members_db_content_array]"
            # Try to query DB again
            continue
        }
        
        return $remove_inactive_res
    } 
    
    writeLog $ping(LOG_ERROR) "[fn]" "Failed to pull DB data."
    return $FAILED_TO_PULL_DB_DATA
 }
 
# Convert member str(e.g. "1_1,1_2") into list of member ids(e.g. "1_1 1_2")
proc member_str_to_lst {member_str} {
    global MEMBER_STR_SEPERATOR
    return [split $member_str $MEMBER_STR_SEPERATOR]
}

# Verify that there were no errors in the current status(action status) of all members
# return 1 on error otherwise 0
proc validate_members_status {member_ids} {
    global members_status
    set err_pattern_lst [list "*failed*" "*error*"]
    
    foreach member_id $member_ids {    
        set m_status $members_status($member_id)
        foreach pattern $err_pattern_lst {
            if {[string match -nocase $pattern $m_status]} {
                return 1
            }
        }
    }
    return 0
}


# Display action progress
proc wait_for_result {action pkg_str members_str} {
    global ping
    global NO_ACTIVE_MEMBERS ALL_MEMBERS_ARE_ACTIVE
    global ALL_MEMBERS_ACTIONS_WERE_COMPLETED
    
    set ret_value 0
    set progress_package_name ""
    
    #after $ping(DELAYED_DISPLAY) ; # So that the user can read dbget result before display is refreshed
    
    set last_status_change_time [clock clicks -milliseconds]
    
    set members_str_original $members_str
    
    # Refresh and Display progress
    while 1 {
        # Delay between refresh
        after $ping(PROGRESS_REFRESH_TIME)
        
        set process_db_query_result 0
        set pull_db_result [wait_for_result_pull_db $action $members_str members_db_content_array]
        if {$pull_db_result < 0} {
            return -1 ;# Failed to pull DB
        } elseif {$pull_db_result != $NO_ACTIVE_MEMBERS &&
                  $members_str != ""} {
            # Success pulling DB from at least some members
            set process_db_query_result [wait_for_result_process_db_query $action members_str [array get members_db_content_array] progress_package_name $pkg_str]
        }
        
        # Time_out handler
        if {[wait_for_result_time_out_handler last_status_change_time]} {
            writeLog $ping(LOG_INFO) "[fn]" "Quitting due to time-out.\n"
            break
        }
     
        # Note: don't call <print_members_status> before  <wait_for_result_time_out_handler> so it wouldn't initialize  <ping(members_status_changed)> and return wrong result
        print_members_status
         
        if { $process_db_query_result == $ALL_MEMBERS_ACTIONS_WERE_COMPLETED  &&  
             $pull_db_result == $ALL_MEMBERS_ARE_ACTIVE } {
            break
        } elseif {$process_db_query_result < 0} {
            # Error has occurred
            set ret_value $process_db_query_result
            break
        }
    } 
    # End of while loop
    
    # Restore original members_str before clear GAIA DB values
    set members_str $members_str_original
    set member_ids [member_str_to_lst $members_str]
    if {[validate_members_status $member_ids]} {
        # Failed to validate member's status - probably action failed at least on one member
        writeLog $ping(LOG_ERROR) "[fn]" "Action failed ($action, $pkg_str). \n"
        set ret_value -1
    }
    
    # Remove remove_inactive_member_ids before from members_str clearing DB values
    set res [remove_inactive_member_ids members_str]
    
    return $ret_value
}

proc wait_for_result_time_out_handler {last_status_change_time} {
    global ping
    upvar $last_status_change_time loc_last_status_change_time

    # DEFINES
    set TIME_OUT_DIDNT_OCCUR 0
    set TIME_OUT_OCCURED     1
    
    set time_elapsed_since_status_has_changed [expr [clock clicks -milliseconds] - $loc_last_status_change_time]
    
    if { $ping(members_status_changed) == 0 } {
        if {[expr $time_elapsed_since_status_has_changed / 1000] >= $ping(WAIT_FOR_RESULT_TIME_OUT_SECONDS) } {
            return $TIME_OUT_OCCURED
        }
    } else {
        # Initialize time loc_last_status_change_time because status change was detected (action is still in progress)
        set loc_last_status_change_time [clock clicks -milliseconds]
    }
    
    return $TIME_OUT_DIDNT_OCCUR
}
 
proc wait_for_result_clear_installer_db_values {members_str progress_package_name members_db_content_array_data self_update_in_progress_member_ids} {
    global ping
    
    set result 0
    
    if { $progress_package_name != "" } {
        # Package name is not empty -> progress values should be cleared
        set member_ids_str_to_clear [get_member_ids_str_to_clear $members_db_content_array_data $self_update_in_progress_member_ids]

        # Clear CPUSE package progress from GAIA DB
        if {[clear_db_package_progress $progress_package_name $member_ids_str_to_clear ]} {
            writeLog $ping(LOG_ERROR) "[fn]" "Failed to clear package_progress value from Gaia DB."
            set result -1
        }
    }
  
    
    # Clear CPUSE general progress from GAIA DB
    if {[clear_progress $members_str]} {
        # Failed to clear DB values
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to clear package_progress value from Gaia DB."
        set result -1
    }
    
    return $result
}


proc refresh_action_progress {action pkg_name member_db_array  member_ids_exceptions  member_ids_completed_action } {
    global ping
    upvar $member_db_array loc_member_db_array
    upvar $member_ids_exceptions loc_member_ids_exceptions
    upvar $member_ids_completed_action loc_member_ids_completed_action
    
    set KEY_PROGRESS            "$ping(BIND_PING):$ping(BIND_PACKAGES):$pkg_name:$ping(BIND_PROGRESS)"
    set KEY_STATUS                    "$ping(BIND_PING):$ping(BIND_PACKAGES):$pkg_name:$ping(BIND_STATUS)"
    set KEY_REMAINING_TIME    "$ping(BIND_PING):$ping(BIND_PACKAGES):$pkg_name:$ping(BIND_PROGRESS):$ping(BIND_REMAINING_TIME)"

    array set installation_progress {}
      
    if {[isDebugMode]} { puts "Display progress (refresh_action_progress() )" }
    
    # Display progress
    foreach member_id [array names loc_member_db_array] {
        # Skip members that already finished
        if {[info exists loc_member_ids_completed_action($member_id)]} { 
            continue
        }
        
        if {[isDebugMode]} { puts "refresh_action_progress(): Update progress of $member_id" }
        
        # Build progress value

        # Extract percentage value
        set progress [extract_v_from_k_v $loc_member_db_array($member_id) $KEY_PROGRESS]

        # Extract status text
        set status [extract_v_from_k_v $loc_member_db_array($member_id) $KEY_STATUS]

        # Clear HTML tags from status
        set status [strip_HTML_tags $status]

        # Initialize remaining_time_text
        set remaining_time_text ""
        
        # Remaining time value in DB exist 
        if {[string first $ping(ACTION_DOWNLOAD) $action]!=-1} {
            # Extract remaining time
            set remaining_time_value [extract_v_from_k_v $loc_member_db_array($member_id) $KEY_REMAINING_TIME]
            
            if {$remaining_time_value!=""} {
                set remaining_time_text ". Remaining time: $remaining_time_value"
            }
        }
         
        set no_member_exception ![info exists loc_member_ids_exceptions($member_id)]
        
        # Update member status
        if {$status!="" && [expr $no_member_exception]} {
            
            if {[isDebugMode]} { puts "refresh_action_progress(): Setting member's progress value of $member_id" }
            
            # Set progress value
            if {$progress != ""} {append progress "%"}
            set_member_status $member_id "[format "%-*s " "28" $status] $progress$remaining_time_text"
        }
    }

}

# Description: Return a member_ids_str which the member's installer progress values should be cleared from Gaia DB after installer action was completed
#
# Input:
#        members_Gaia_DB_content_array per member_id  (e.g. the output of the following command: # dbget -rv installer   per member_id)
#        member_ids_exceptions - List of member_ids which shouldn't be added to member_ids_str_to_clear
#
# Return: 
#        member_str(e.g. "1_01,1_02") of cluster members which the member's installer progress values should be cleared from Gaia DB according to the conditions
#        in wait_for_result() function located in DAclish_utils.c
#
proc get_member_ids_str_to_clear {members_db_array_data member_ids_exceptions} {
    global ping 
    global KEY_CLISH_PROGRESS_RESULT
    
    array set member_db_array $members_db_array_data

    set member_ids_to_clear ""
    
    foreach member_id [array names member_db_array] {
        # Extract action_result value
        set action_result [extract_v_from_k_v $member_db_array($member_id) "$KEY_CLISH_PROGRESS_RESULT"]
        
        # These conditions are according to wait_for_result() function located in DAclish_utils.c
        if { $action_result != ""                                 &&
             [lsearch $member_ids_exceptions $member_id] == -1    && 
             ![regexp -nocase -- "Import Failed"  $action_result] &&
             ![regexp -nocase -- "Legacy package" $action_result]    } {
            if {$member_ids_to_clear != ""} {append member_ids_to_clear ","}  ;# append ","
            append member_ids_to_clear $member_id
        }
    }

    return $member_ids_to_clear
}
   

# Clear CPUSE package progress (status and percentage) from GAIA DB
proc clear_db_package_progress {pkg_name member_ids_str_to_clear} {
    global ping 
    global KEY_CLISH_PROGRESS_RESULT
    
    set KEYS_LIST [list "$ping(BIND_PING):$ping(BIND_PACKAGES):$pkg_name:$ping(BIND_STATUS)"  \
                        "$ping(BIND_PING):$ping(BIND_PACKAGES):$pkg_name:$ping(BIND_PROGRESS)"  ]

    if {$pkg_name == ""} {
        writeLog $ping(LOG_ERROR) "[fn]" "Package name is empty"
        return -1
    }
    
    if { $member_ids_str_to_clear == "" } {
        return 0
    }

    return [clear_db_values $member_ids_str_to_clear $KEYS_LIST]
}


# Description: Extract key's value from Gaia DB content (e.g. extract a value of a key, from the following input format: "KEY   VALUE\nKEY   VALUE...")
#
# Return: 
#        <Key's value> - In case parsing succeeded
#        ""            - In case the key wasn't found
#
proc extract_v_from_k_v {key_value_data  key {should_strip_HTML_tags 1}} {
    
    # Split data into lines
    set lines_list [split $key_value_data "\n"]
    
    # Initialize return value
    set value_of_key ""
    
    # Flag to deal with values with multiple lines
    set key_found 0
    
    # Search for the key and build it's value (Necessary in case DB value contains multiple lines)
    foreach line $lines_list {
        
        if {$key_found == 0} {
            set first_space_index [string first " " $line]
            set current_key [string range $line 0 [expr $first_space_index-1]] 

            if {$key == $current_key} {
                # Key was found
                set key_found 1
                append value_of_key [string range $line [expr $first_space_index+1] end]
            }
        } elseif {$key_found == 1} {
            # Append entire line to value
            append value_of_key $line
        }
        
        # Check whether the key's value has additional lines
        if {$key_found == 1} {
            set lastChar [string index $line end]
            
            # In case there is extra line in GAIA DB value, the last char in the line is '\'
            if { $lastChar=="\\"} {
                # Key's value has additional lines
                append value_of_key "\n"
            } else {
                # Done building key's value
                break
            }
        }
        
    }
    
    if {$should_strip_HTML_tags} {
        # Remove HTML tags from the extracted value
        set value_of_key [strip_HTML_tags $value_of_key]
    }
    
    return $value_of_key
}


# Description: Extract key's value from DAClient output
#
# Return: 
#        <Key's value>    - In case parsing succeeded
#        ""                - In case the key wasn't found
#
proc extract_v_from_DAClient_output {exec_output key {should_strip_HTML_tags 1}} {
    # Split data into lines
    set lines_list [split $exec_output "\n"]
    
    # Initialize return value
    set value_of_key ""
    
    set key_found 0
    
    # Search for the key and build it's value (Necessary in case DB value contains multiple lines)
    set key "${key}: " ; # Append ": " to key according to the DAClient output format
    set value_start_pos [string length $key] ; # The value begins after the key (e.g. "key_1: value_1")
    
    foreach line $lines_list {
        if {[string first $key $line] == 0} {
            # Key was found at the begining of the line
            # Extract the value
            set value_of_key [string range $line $value_start_pos end]            
            set key_found 1
            break
        }
    }
    
    if {$key_found && $should_strip_HTML_tags} {
        # Remove HTML tags from the extracted value
        set value_of_key [strip_HTML_tags $value_of_key]
    }
    
    return $value_of_key
}


# Build command string for clearing values from GAIA DB 
#
# Input:     List of GAIA DB keys             (e.g. : "installer:da_build installer:package_running"                                )
# Output:     Command string to execute    (e.g. : "dbset installer:da_build ; dbset installer:package_running ; ...."    )
#
proc build_cmd_for_clearing_db_values {db_key_lst} {
    global ping

    # Build command string #
    set cmd ""

    foreach key $db_key_lst {
        if {$cmd != ""} { append cmd " ; "}
        append cmd "$ping(DBSET_BINARY) $key"
    }
    
    return $cmd
}


# Clear values from GAIA DB
proc clear_db_values {members_str db_key_lst} {
    global ping
    set MAX_ATTEMPTS             3
    set ATTEMPT_TIME_DELAY    500

    set cmd [build_cmd_for_clearing_db_values  $db_key_lst]
    
    # Try to clear DB values #
    for {set current_attempt 1} {$current_attempt <= $MAX_ATTEMPTS} {incr current_attempt} {
        while 1 {
            # Execute command
            if {[gclish_exec_remote_members $members_str $cmd member_result_array]} {
                break  ;# Failed to clear DB value, Try again
            }
                
            # Verify that execution output is empty(means success) per member_id #
            set all_members_execution_output_is_empty 1
            foreach member_id [array names member_result_array] {
                if {$member_result_array($member_id) != ""} {
                    set all_members_execution_output_is_empty 0
                    break  ;# Failed to clear DB value, Try again
                }
            }
            
            if {$all_members_execution_output_is_empty != 1} {
                break ;# Failed to clear DB value, Try again
            }
            
            # Command was executed successfully on all members
            return 0
        }
        
        # Failed to clear DB value, Try again
        writeLog $ping(LOG_DEBUG) "[fn]" "Attempt $current_attempt, Failed to clear DB values"
        after $ATTEMPT_TIME_DELAY ; # Wait before next attempt
    }
    
    # Failed to clear values <MAX_ATTEMPTS> times #
    print_members_status
    writeLog $ping(LOG_ERROR) "[fn]" "Failed to clear DB values."
    writeLog $ping(LOG_DEBUG) "[fn]" "$cmd"
    
    return -1
}

# Clear Installer progress values from GAIA DB
proc clear_progress {members_str} {
    global ping
    
    if {[isDebugMode]} { puts "clear_progress {}"}
    
    # Build the list of keys to clear their values in GAIA DB   [list "action_result" "action_type" "action_pkg"]
    set KEYS [list  $ping(BIND_CLISH_PROGRESS_RESULT) $ping(BIND_CLISH_PROGRESS_TYPE) $ping(BIND_CLISH_PROGRESS_PKG)]

    set keys_list [list]
    
    # Append <BIND_PING> to each of of the keys
    
    foreach key $KEYS { 
        lappend keys_list "$ping(BIND_PING):$key"
    }
   
    # Clear DB values
    return [clear_db_values $members_str $keys_list]
}
 
proc set_member_status {members_str status} {
    global ping members_status
    
    set members_list [split $members_str ","]
    
#    if {[isDebugMode]} {    
#        puts "set_member_status(), members_str='$members_str'" 
#        puts "set_member_status(), members_list='$members_list'"
#    }
       
    
    foreach member_id $members_list {
         
#        if {[isDebugMode]} {    
#            puts "set_member_status(), current member_id='$member_id'" 
#            puts "set_member_status(), members_status=[array get members_status    ]"
#        }
        
        set status [convert_member_status                     $status]
        
        if {[fit_status_to_cell $status new_status]==0} {
            set status $new_status
        } else {
            # "Failed" to fit_status_to_cell
            if {[isDebugMode]} {
                puts "'Failed' to fit_status_to_cell"
            }
        }
        
        if {$members_status($member_id)!=$status} {
            set members_status($member_id)   $status
            set ping(members_status_changed)   1
        }
    }
}



proc set_all_members_status {  status } {
    global member_ids  members_status
    
    # Set each member's status
    foreach { member_id } $member_ids {
        set_member_status $member_id  $status
    }

    print_members_status
}

# Convert '01_2'  to '1_02'  , <other> to <other>
proc convert_to_formal_member_id { member_id } {
    set formal_member_id $member_id
    
    # Split string by "_"
    set splitted_member_id [split $member_id "_" ]
    
    if {[llength  $splitted_member_id] == 2} {
        # Found "_" in  $member_id
        
        set chassis_id     [lindex $splitted_member_id 0]   ;# '1'
        set member_id     [lindex $splitted_member_id 1]   ;# '02'  or '2'
        
        if {[string is integer -strict $chassis_id] && [string is integer -strict $member_id] }  {
            # Both chassis_id and member_id are integers
            
            # Convert '01' to '1'
            set chassis_id [expr $chassis_id]
            
            # Build formal member_id
            set         formal_member_id  "$chassis_id"
            append formal_member_id  "_"
            
            if {[string length $member_id] < 2} {
                # Blade number is less than 2 characters , Add '0' prefix for blade number and build final formal_member_id
                 append formal_member_id "0"
            }
            
            append formal_member_id $member_id
        }
    }
    
    return $formal_member_id
}

# Convert member_id_list into 'formal' member_id_list
#
# Example: "1_1 1_2" to "1_01 1_02"
#
proc convert_to_formal_member_ids {member_id_lst} {
    set formal_member_id_lst [list]
    
    # Convert to formal id and add to list
    foreach member_id $member_id_lst {
        lappend formal_member_id_lst [convert_to_formal_member_id $member_id]
    }
    return $formal_member_id_lst
}

# Create  <member_ids list> and member's status array from $members_str
proc init_members_ids_lst_and_status {members_str} {
    global member_ids members_status  
    
    # Convert members string to list
    set members_str_lst [split $members_str  ","]
    
    # Initialize member_ids list
    set member_ids [list]
    
    # Initialize member_status array
    if {[info exist members_status]} {unset members_status}
    array set members_status {}
    
    foreach { member_id } $members_str_lst {
        # Initialize member's status
        set members_status($member_id)   ""
        # Append member_id to member_ids list
        lappend member_ids $member_id
    }

    # Sort member IDs list
    set member_ids [lsort -dictionary  $member_ids]
    
    return 0
}

# Initialize member id's string 
# (e.g. Convert 'all' to cluster active member id's string,
#  Convert blade range to expanded blade string)
# In case action is 'download..', the function
# make sure member_str contains local_member_id 
proc init_member_ids_str_from_user_input {action member_str} {
    global ping   
    global skip_installation_on_local_member
    
    upvar $member_str loc_member_str
    
    if [isDebugMode] { puts  "loc_member_str=$loc_member_str"}
     
    # Convert to lower case
    set loc_member_str [string tolower $loc_member_str]
    
    # Replace 'local' in loc_member_str with the <local_blade_id>
    regsub -all $ping(LOCAL_MEMBER_STR) $loc_member_str $ping(local_member_id) loc_member_str
    
    set all_members_str [list_to_csv $ping(all_cluster_members)]
    
    # Replace 'all' in loc_member_str with <all cluster members ids>
    regsub -all $ping(ALL_MEMBERS_STR) $loc_member_str $all_members_str loc_member_str

    # Get blade_ids from BLADE RANGE
    set error_list                 ""
	set loc_member_str	[sg_range2ids $loc_member_str error_list]

    if {$error_list != "" || $loc_member_str == ""}  {
        # Failed to parse member ids range
        if {$error_list != ""} {
            set error_str ""
            foreach error $error_list {
                append error_str \n$error
            }
            writeLog $ping(LOG_ERROR) "[fn]" "Invalid member ids: $error_str\n"
        } elseif {$loc_member_str == ""} {
            writeLog $ping(LOG_ERROR) "[fn]" "Invalid member_ids (None)."
        }
        return -1
    }

    # Convert to formal member id  #
    
    # Member_ids string to list
    set member_ids_lst [split $loc_member_str  ","]

    # Initialize members string
    set loc_member_str ""
    
    # Convert members ids to their formal form (Array is used in order to avoid duplications)
    array set formal_members_ids {}
        
    foreach { member_id } $member_ids_lst {
        set formal_members_ids([convert_to_formal_member_id $member_id]) 1
    }
    
    # In case action is download/download-and-install then local_member_id must be included in member_str
    # package_download is being performed on local member only and then transferred to the rest of the security group
    if {     ($action == $ping(ACTION_DOWNLOAD) || $action == $ping(ACTION_DOWNLOAD_AND_INSTALL)) 
            && 
            [info exist formal_members_ids($ping(local_member_id))] == 0 } {
            
        if {[get_user_approval_to_download_pkg_on_local_member] == 1} {
            set formal_members_ids($ping(local_member_id)) 1
            set skip_installation_on_local_member    1
        } else {
            return -1 ;# User has declined
        }
    }
    
    # Sort member_ids
    set formal_member_ids_lst [lsort -dictionary  [array names formal_members_ids] ]

    # Append each formal member_id to <loc_member_str>
    foreach { formal_member_id } $formal_member_ids_lst {
        if {$loc_member_str != "" } {append loc_member_str ","}
        append loc_member_str     $formal_member_id
    }
    return 0
}


proc get_user_approval_to_download_pkg_on_local_member {} {
    global ping
    
    puts "Package must be downloaded on the local member in order to continue,"
    puts "\nWould you like to continue and download the package on the local member($ping(local_member_id))?"
    puts "(Y - yes, any other key - no)"
    
    if {![isTestingMode]} {
        get_user_input user_input
    } else {
        set user_input "y"
    }

    # Handle user input
    if { $user_input != "y" } {
        # User has declined
        return -1
    } 
    
    return 1
}


# Verify member_str according to the real cluster members_ids and suggest the user to replace it with the active_member_ids in case of error
proc verify_and_fix_member_ids_str {gclish_cmd members_str} {
    global ping
    upvar $members_str loc_members_str

    set member_ids_lst [split $loc_members_str ","]
    
    ## Build invalid/inactive member_id's string ##
    set inactive_member_ids ""
    set invalid_member_ids  ""

    get_communicating_members_lst $loc_members_str communicating_members
    
    # Check if the members_ids which was entered by the user are valid
    foreach { member_id } $member_ids_lst {
        
        if { [lsearch $ping(all_cluster_members) $member_id]         == -1 } {
            # Invalid member_id
            if {$invalid_member_ids != ""}     {append invalid_member_ids ","}
            append invalid_member_ids "$member_id"
        } elseif { [lsearch $communicating_members $member_id]    == -1 } {
            # member_id is valid but inactive
            if {$inactive_member_ids != ""}     {append inactive_member_ids ","}
            append inactive_member_ids "$member_id" 
        }
    }
    
    set were_invalid_found  [expr [string length $invalid_member_ids]  > 0]
    set were_inactive_found [expr [string length $inactive_member_ids] > 0]
    
    # Display Invalid/Inactive members #
    if { $were_invalid_found || $were_inactive_found } {
        if {$were_invalid_found} {
            # Build and display output message # 
            set output_msg "\nFailed to initialize members. ("
            if {$were_invalid_found } {append output_msg    "Invalid member ids: $invalid_member_ids"}
            if {$were_invalid_found && $were_inactive_found } {append output_msg ", "}
            if {$were_inactive_found} {append output_msg    "Inactive member ids: $inactive_member_ids"}
            append output_msg "\)"
        } else {
            set output_msg "Several cluster members are not communicating($inactive_member_ids)\n"
        }
        
        puts $output_msg
             
        if {[remove_inactive_member_ids loc_members_str] == -1} {
            writeLog $ping(LOG_ERROR) "[fn]" "Failed to get active member ids"
            return -1
        }
        
        if {$loc_members_str == ""} {
            writeLog $ping(LOG_ERROR) "[fn]" "All selected members are not communicating($inactive_member_ids)"
            return -1
        }

        # Suggest to perform action on active members only in case 
        # NOT 'show installer...'/'delete installer...' commands
        # Otherwise, perform it only the active members, without getting the user approval
        if { [string match -nocase "*show*installer*"   $gclish_cmd] == 0 && 
             [string match -nocase "*delete*installer*" $gclish_cmd] == 0    } {
             
            if {[suggest_to_perform_on_active_members $loc_members_str] == 1} {
                return 0
            } else {
                return -1 ;# User has declined
            }
        }
    }
    return 0
}


proc list_to_csv {lst} {
    return [join $lst ","]
}

proc suggest_to_perform_on_active_members {active_members} {
    global ping
    
    puts "Would you like to perform the action on the active members($active_members)"
    puts "(Y - yes, any other key - no)"
    
    if {![isTestingMode]} {
        get_user_input user_input
    } else {
        set user_input "y"
    }

    # Handle user input
    if { $user_input=="y" } {
        return 1
    } 

    return 0
}


# Description: Get the local cluster member id
#
# Return: 
#         0  - Command execution succeeded
#        -1  - Command execution failed
#
#        <local_member_id> - Local cluster member id
#
proc get_local_member_id {local_member_id} {
    global ping
    upvar $local_member_id loc_local_member_id

    # Build the command
    set cmd "$ping(CMD_PREFIX) ; $ping(CMD_GET_LOCAL_MEMBER_ID)"    
    
    # Execute command
    return [gclish_exec $cmd loc_local_member_id]
}

   
# Parse DAclish_utils.c arguments (CLI arguments) and set the values of:  
# gclish_cmd
# e.g. args = "-cmd installer import local 61k_package.tgz member_ids local"
proc parse_cli_args {args gclish_cmd pkg_str} {
    global ping 

    upvar $gclish_cmd  loc_gclish_cmd
    upvar $pkg_str     loc_pkg_str

    set loc_gclish_cmd  ""
    set loc_pkg_str     ""
    
    set PARSING_ERR_MSG "Failed to parse DAclish_utils.c arguments, CLI switch is missing"
    
    # Iterate through the command line arguments and build the variables values
    set var_to_append ""
    foreach cur_arg $args {
        switch -- $cur_arg \
            $ping(DACLISH_UTILS_CMD_SWITCH)      {set var_to_append "loc_gclish_cmd"}   \
            $ping(DACLISH_UTILS_PKG_NAME_SWITCH) {set var_to_append "loc_pkg_str"}      \
            default {
                if {$var_to_append == ""} {
                    writeLog $ping(LOG_ERROR) "[fn]" $PARSING_ERR_MSG
                    writeLog $ping(LOG_DEBUG) "[fn]" "args=$args"
                    return -1
                }
                append $var_to_append "$cur_arg "
            }
        
    }
      
    # Trim values
    set loc_gclish_cmd  [string trim $loc_gclish_cmd]
    set loc_pkg_str     [string trim $loc_pkg_str]

    pvars loc_gclish_cmd loc_pkg_str
    
    # Values verification
    if {$loc_gclish_cmd == "" } {
        writeLog $ping(LOG_ERROR) "[fn]" $PARSING_ERR_MSG
        writeLog $ping(LOG_DEBUG) "[fn]" "args=$args"
        return -1
    }
    
    # Note: Don't convert gclish command to lower case in order to avoid changing the package file name to lower case.
    return 0
}

proc switch_to_test_mode {} {
    global ping 
    global TEST_MODE

    # Disable delayed display
    set ping(DELAYED_DISPLAY) 0
    set TEST_MODE 1
    
    writeLog $ping(LOG_INFO) "[fn]" "Switching to test mode..."
}

# Handle "show installer *" commands
proc perform_show {gclish_cmd} {
    global member_ids ping
    global PKG_INFORMATION_NAME
    
	set local_only 0 ;# Show action only on local member
	set subject [lindex $gclish_cmd 2]
    set pkg_display_name ""
    set pkg_str          ""
    
    pvars gclish_cmd subject
    
    # User global member_ids because this is a sorted list
    set members_str [join $member_ids ","]
    
    # Prepare exec command according to gclish_cmd
    if {$subject == "packages"} {
        set type [lindex $gclish_cmd 3]
        if {$type == "all"} {
            set bind_type "show_status"
        } elseif {$type == "available-for-download"} {
            set local_only 1
            set bind_type   "show_ap"
        } elseif {$type == "downloaded"} {
            set local_only 1
            set bind_type "show_alp"
        } elseif {$type == "imported"} {
            set bind_type "show_import"
        } elseif {$type == "installed"} {
            set bind_type "show_ip"
        } elseif {$type == "recommended"} {
            set local_only 1
            set bind_type "show_rp"
        } else {
            set bind_type "show_status"
        }
        
        set action "info $bind_type"
    } elseif {$subject == "package"} {
        set pkg_str [lindex $gclish_cmd 3]
        set field   [lindex $gclish_cmd 4]
        
        if {$field == ""} {
            set field "all"
        }
        
        if {[string is integer -strict $pkg_str]} {
            # Convert pkg_number to pkg_name according to local member in order
            # to make sure all members will show the same package
            # in case the pkg has different index on different members
            if { [get_local_pkg_action_information "" $pkg_str local_pkg_action_information 0]} {
                writeLog $ping(LOG_DEBUG) "[fn]" "Failed to retrieve local package information"
                putsd "[fn](): Failed to retrieve local package information"
                return -1
            }
            set pkg_str_new $local_pkg_action_information($PKG_INFORMATION_NAME)
            putsd "[fn]: package string conversion: $pkg_str -> $pkg_str_new"
            set pkg_str $pkg_str_new
            if {$pkg_str == $ping(NA_STR)} {
                puts "Package index doesn't exist on local member. \n"
                return -1
            }
        }
        
        set action "package_info $pkg_str $field"
    } elseif {$subject == "policy"} {
        set type [lindex $gclish_cmd 3]
        if {$type == "all"} {
            set bind_type "all"
        } elseif {$type == "check-for-updates-period"} {
            set bind_type   "show_check_for_updates_period"
        } elseif {$type == "downloads"} {
            set bind_type "show_downloads_policy"
        } elseif {$type == "periodically-self-update"} {
            set bind_type "show_periodic_update"
        } elseif {$type == "self-test"} {
            set tst_type [lindex $gclish_cmd 4]
            if {$tst_type == ""} {
                set bind_type "self-test-all"
            } else {
                set bind_type "self-test-$tst_type"
            }
        } elseif {$type == "send-cpuse-data"} {
            set bind_type "show_send_data"
        } else {
            set bind_type "all"
        }
        
        set action "da_policy_info $bind_type"
    } elseif {$subject == "status"} {
        set bind_type [lindex $gclish_cmd 3]
        if {$bind_type == ""} {
            set bind_type "all"
        }
        
        set action "da_info $bind_type"
    } elseif {$subject == "mail-notifications"} {
        set email_address [lindex $gclish_cmd 3]
        set action "show_mails $email_address"
		# puts $action
    } else {
        # Action was not handled
        writeLog $ping(LOG_ERROR) "[fn]" "Action is disabled"
        return 0
    }

	if {$local_only == 1} {
		set members_str $ping(local_member_id)
	}
    # Invoke action
    return [DAClient_exec_simple $members_str $action 1]
}

# Handle "show installer *" commands
proc perform_delete {gclish_cmd} {
    global member_ids ping
    global PKG_INFORMATION_NAME
    
	set local_only 0 ;# Perform action only on local member
	set subject [lindex $gclish_cmd 2]
    set pkg_display_name ""
    set pkg_str          ""
    
    pvars gclish_cmd subject
    
    # User global member_ids because this is a sorted list
    set members_str [join $member_ids ","]
    
    # Prepare exec command according to gclish_cmd
    if {$subject == "mail-notifications"} {
        set email  [lindex $gclish_cmd 3]
		set action "delete_mail_notifications $email"
    } else {
        # Action was not handled
        writeLog $ping(LOG_ERROR) "[fn]" "Action is disabled"
        return 0
    }

	if {$local_only == 1} {
		set members_str $ping(local_member_id)
	}
	
    # Invoke action
    return [DAClient_exec_simple $members_str $action 0]
}

proc reformat_show_packages_output {member_id mem_output} {
	global ping

    # Remove specific headers only from non-local members
    if {$member_id != $ping(local_member_id)} {
        # Remove connection error header
        set rp_ce_h    "\\*\\* \\*+ \\*\\*\n"
        append rp_ce_h "\\*\\* +Connection error. Packages list might be incomplete +\\*\\*\n"
        append rp_ce_h "\\*\\* \\*+ \\*\\*\n"
        regsub -all -nocase $rp_ce_h $mem_output "" mem_output
        
        # Remove 'Checking for new available packages...' header
        set rp_chp_h    "\\*\\* \\*+ \\*\\*\n"
        append rp_chp_h "\\*\\* +Checking for new available packages is in progress +\\*\\*\n"
        append rp_chp_h "\\*\\* \\*+ \\*\\*\n"
        regsub -all -nocase $rp_chp_h $mem_output "" mem_output
    }
    
    # Remove Hotfixes header
    set rp_hf_h    "\\*\\* \\*+ \\*\\*\n"
    append rp_hf_h "\\*\\* +Hotfixes +\\*\\*\n"
    append rp_hf_h "\\*\\* \\*+ \\*\\*\n"
    regsub -all -nocase $rp_hf_h $mem_output "" mem_output

    # Remove 'Num' column #
    # Remove 'Num' header (i.e. "Num Display name Status" to "Display name Status")
    regsub -all -line -nocase "^Num " $mem_output "" mem_output
    # Remove pacakges indexes (i.e. "1  sim_HOTFIX_HOP_HF_001_GA_FULL.tgz Available for Install" to "sim_HOTFIX_HOP_HF_001_GA_FULL.tgz Available for Install")
    regsub -all -line -nocase "^(\\d+) +" $mem_output "" mem_output
    
    return $mem_output
}

proc DAClient_exec_simple {members_str action is_show_cmd} {
	global ping
	set pkg_display_name ""
	set pkg_str          ""
	
    if {[DAClient_exec $members_str $action $pkg_str $pkg_display_name member_result_array]} {
        # Failed to execute dbget on members
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to invoke action."
        return -1
    } else {
        # Print output per member
        set sorted_member_ids [lsort -dictionary [array names member_result_array]]
        foreach member_id $sorted_member_ids {
            # Print member_id header
            puts "${member_id}:" 
             
            set mem_output $member_result_array($member_id)       
            
            if {$is_show_cmd} {
                set mem_output [reformat_show_packages_output $member_id $mem_output]
            }
            
            # Print member's output
            puts $mem_output
        }
    }
    
    return 0
}

proc main {cli_args} {
    global ping 
    pvars cli_args
    
    if {[string match "*${ping(DACLIENT_WRAPPER_FLAG)}*" $cli_args]} {
        # DAClient wrapper(Execute DAClient binary and clean/reformat the output)
        regsub -all -- "-DAClient" $cli_args "" cli_args
        return [DAClient_exec_and_clean $cli_args]
    }
    
    if {[file exists $ping(TEST_MODE_FILE)]} {
        switch_to_test_mode
    } 
    
    set pkg_str ""
    if {[parse_cli_args $cli_args ping(gclish_cmd) pkg_str] } { 
        return -1
    }
    
    pvars ping(gclish_cmd)
	
    if  {[extract_action $ping(gclish_cmd) installer_action]} {
        return -1
    }
	
    pvars installer_action

    if {[should_run_only_locally $installer_action]} {
        set members_str $ping(LOCAL_MEMBER_STR)
    } else {
        if {[extract_members_ids $ping(gclish_cmd) members_str]} {
            return -1
        }
    }

    pvars pkg_str
    # exit
    if {$pkg_str == ""} {
        if  {[extract_pkg_str $ping(gclish_cmd) $installer_action pkg_str]} {
            return -1
        }
    }
    pvars pkg_str

    if {[get_cluster_member_ids ping(all_cluster_members) ping(local_member_id)]} { 
        return -1 
    }
     
    if {[init_member_ids_str_from_user_input $installer_action members_str]} { 
        return -1 ; # Invalid member ids were given by the user
    } 
	pvars members_str
	
    if {[verify_and_fix_member_ids_str $ping(gclish_cmd) members_str]} { 
        return -1 
    }
	pvars members_str

    init_members_ids_lst_and_status $members_str
	
	if {[regexp -nocase -- ".*show *installer.*" $ping(gclish_cmd)]} {
		return [perform_show $ping(gclish_cmd)]
	} elseif {[regexp -nocase -- ".*delete *installer.*" $ping(gclish_cmd)]} {
		return [perform_delete $ping(gclish_cmd)]
	}
     
    screen_clear
     
    print_members_status

	pvars installer_action pkg_str members_str

    set pkg_display_name $pkg_str
    set res [perform_installer_action \
             $installer_action $pkg_str \
             $pkg_display_name $members_str]
    
    log_last_action_status
    
    return $res
}

proc log_last_action_status {} {
    global ping
    global member_ids members_status

    writeLog  $ping(LOG_INFO) "[fn]" "Last status:" 0
    
    foreach member_id $member_ids {
        writeLog  $ping(LOG_INFO) "[fn]" "$member_id: $members_status($member_id)" 0
    }
}

# Description: Get communicating member ids 
#
# Example:
# Input:   members_str: "1_01,1_02,1_03
# Output:  communicating_members: "1_01 1_03"
proc get_communicating_members_lst {members_str communicating_members} {
    upvar $communicating_members loc_communicating_members
    set loc_communicating_members [list]
    set echo_str "alive"
    set cmd "echo $echo_str"
    
    gclish_exec_remote_members $members_str $cmd output_array 5 1
    
    foreach i [array names output_array] {
        if {[string match "*${echo_str}*" $output_array($i)]} {
            lappend loc_communicating_members $i
        }
    }
    
    return 0
}


# Description: Get cluster member ids
#
# Return: 
#        0        - Success
#       -1        - Otherwise
#
proc get_cluster_member_ids {all_members local_member} {
    global ping

    upvar $all_members      loc_all_members
    upvar $local_member     loc_local_member
    set loc_all_members    ""
    set loc_local_member   ""
    
    puts "\nQuerying cluster members...\n"
    
    # Get All cluster members and Active cluster members IDs 
    # (using 'cphaprob')
    if { [get_blades_ids loc_all_members 1] ||
         [regexp -nocase -- ".*module not started.*" $loc_all_members] ||
         $ping(all_cluster_members) == ""         } {
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to get cluster member ids"
        return -1
    }
    
    set loc_all_members    [convert_to_formal_member_ids $loc_all_members]

    # Initialize local member id
    set loc_local_member ""
    if { [get_local_member_id loc_local_member] || 
         $loc_local_member == ""                    } {
        # Failed to get local member id
        writeLog  $ping(LOG_ERROR) "[fn]" "Failed to get local member id"
        return -1
    } else {
        set loc_local_member [convert_to_formal_member_id $loc_local_member]
    }
    return 0
}


# Description: Extract the package string from GCLISH command OR from the binding string according to the action    
#              (e.g. extract 'pkg_1.tgz' from 'installer import local pkg_1.tgz'  into <pkg_str>
#                    extract '2' from 'installer install 2')
#
# Return: 
#        0        - In case parsing succeeded
#       -1        - In case parsing failed
#
proc extract_pkg_str {gclish_cmd installer_action pkg_str} {
    global ping 
    upvar $pkg_str loc_pkg_str
    set loc_pkg_str ""
    
    switch -- $installer_action             \
        $ping(ACTION_IMPORT_CLOUD)      -   \
        $ping(ACTION_INSTALL_DA)        -   \
        $ping(ACTION_IMPORT) {
            # Get package file name from the bind string (Might have been downloaded from an FTP server by DAClish_import.c)
            # set pkg_file_name        [lindex $bind_arg_list end]
            # set loc_pkg_str             $pkg_file_name
            
            # GCLISH command: # installer import local <pkg_str>
            set loc_pkg_str [lindex $gclish_cmd 3]
        }\
        $ping(ACTION_CHECK_FOR_UPDATES) -   \
        $ping(ACTION_AGENT_START)       -   \
        $ping(ACTION_AGENT_STOP)        -   \
        $ping(ACTION_AGENT_DISABLE)     -   \
        $ping(ACTION_AGENT_ENABLE)      -   \
        $ping(ACTION_UPDATE_DA)    {
            set loc_pkg_str ""
        }\
        default {
            if {[llength $gclish_cmd] < 3} {
                writeLog $ping(LOG_ERROR) "[fn]" "Failed to parse package string."
                writeLog $ping(LOG_DEBUG) "[fn]" "gclish_cmd=$gclish_cmd"
                return -1
            }

            set loc_pkg_str [lindex $gclish_cmd 2]
        }
    return 0
}

# Description: 
#       Returns whether or not action should be performed
#       only on the local member.
#       (e.g. checking updates from the cloud)
# Return: 
#       1 - Action should be performed only locally
#       0 - Otherwise
#
proc should_run_only_locally {installer_action} {
    global ACTIONS_FOR_LOCAL_MEMBER_ONLY
    
    if {[lsearch $ACTIONS_FOR_LOCAL_MEMBER_ONLY $installer_action] != -1 } {
        return 1 ;# Action was found
    }
    return 0
}

# Description: 
#       Extract member_ids string(e.g. "2_01,1_02,local") from GCLISH command 
#       (Note: Not all GCLISH commands require member_str)
# Return: 
#       Always return 0.
#        <members_str> will contain the member_ids string from the GCLISH command
#       (e.g. "2_01,1_02,local") 
#
proc extract_members_ids {gclish_cmd members_str} {
    global ping 
    
    upvar $members_str loc_members_str
    set loc_members_str "" ; # Initialize members_str
    
    # Extract member_ids string from GCLISH arguments #
    set members_ids_index [lsearch -nocase $gclish_cmd $ping(GCLISH_MEMBER_IDS)]
    set members_ids_str_index [expr $members_ids_index + 1]

    if {$members_ids_index == -1 || 
        $members_ids_str_index >= [llength $gclish_cmd]} { 
        # member_ids_str wasn't found in gclish_cmd, set default member_str
        set loc_members_str $ping(ALL_MEMBERS_STR)
    } else {
        set loc_members_str [lindex $gclish_cmd $members_ids_str_index]
    }
    
    return 0
}
 
 
# Description: Extract the action from GCLISH command into <action> 
#              (e.g. extract 'import' from 'installer import local pkg.tgz')  
#
# Return: 
#       0   - In case parsing succeeded
#      -1   - In case parsing failed
#
proc extract_action {gclish_cmd action} {
    global ping 
    
    upvar $action loc_action
    set loc_action ""
    
    # While loop is used for error handling in order to avoid code duplication
    while 1 {
        if {[llength $gclish_cmd] < 2} {
            break ;# Error, gclish_cmd should have at least 2 arguments (installer <action> ...)
        }
        
        # Set loc_action to be the 2nd/3rd argument of the GCLISH command  ##   (e.g. in 'installer import local pkg.tgz' , loc_action = 'import')
        set gclish_arg_2 [lindex $gclish_cmd 1]
        set gclish_arg_2 [string tolower $gclish_arg_2]
 
        switch -- $gclish_arg_2 \
            $ping(GCLISH_ARG_AGENT) {
                    if {[llength $gclish_cmd] < 3} {
                        break ;# Error, gclish_cmd should have at least 3 arguments (installer agent <action> ...)
                    }
                    set loc_action [lindex $gclish_cmd 2]
                    set loc_action [string tolower $loc_action]
                    
                    if {[string match -nocase $ping(GCLISH_ARG_AGENT_UPDATE) $loc_action]} {
                        # Convert action_name so it would conform with validate_package function in DAcdk.c (This function verifies pkg_name,pkg_state,pkg_action tuple)
                        set loc_action $ping(ACTION_UPDATE_DA)
                    } elseif {[string match -nocase $ping(GCLISH_ARG_AGENT_INSTALL) $loc_action]} {
                        set loc_action $ping(ACTION_INSTALL_DA)
                    }
            }\
            $ping(GCLISH_ARG_DOWNLOAD_AND_INSTALL) {
                        # Convert action_name so it would conform with validate_package function in DAcdk.c (This function verifies pkg_name,pkg_state,pkg_action tuple)
                        set loc_action $ping(ACTION_DOWNLOAD_AND_INSTALL)
            }\
            $ping(GCLISH_ARG_IMPORT) {
                    if {[regexp -nocase -- $ping(GCLISH_IMPORT_CLOUD) $gclish_cmd]} {
                        # Convert "import" into "import_cloud" in order to distinguish between "import local/ftp" and "import cloud"
                        set loc_action $ping(ACTION_IMPORT_CLOUD)
                    } else {
                        set loc_action $gclish_arg_2
                    }
            }\
            $ping(ACTION_UNINSTALL) {
                    # If the uninstallation should be completely we want to change the command "uninstall" to "uninstall_completely" later.
                    if {[regexp -nocase -- $ping(GCLISH_COMPLETELY) $gclish_cmd]} {
                        set ping(is_uninstall_completely) 1
                    }
                    set loc_action $gclish_arg_2
            }\
            default {
                    set loc_action $gclish_arg_2
            }
            pvars loc_action $gclish_arg_2 ping(is_uninstall_completely) gclish_cmd
        # Parsed successfully
        return 0
    }
 
    # Error occured #
    writeLog $ping(LOG_ERROR) "[fn]" "Failed to parse installer action."
    writeLog $ping(LOG_DEBUG) "[fn]" "gclish_cmd=$gclish_cmd"
    return -1
}



proc verify_remote_disk_space {members_str required_free_space_bytes} {
    putsd "verify_remote_disk_space()"
    global ping
    
    set CMD "df -B1 -P ${ping(IMPORT_PATH)}/"
    
    # Execute command, read GAIA DB
    if {[gclish_exec_remote_members   $members_str $CMD member_output_array] } {
        # Failed to execute command
        writeLog $ping(LOG_ERROR) "[fn]" "Disk space verification has failed."
        writeLog $ping(LOG_ERROR) "[fn]" "Failed to execute disk space verification: \n[array get member_output_array]"    0
        return -1
    }
    
    # Clear member's status
    set_member_status $members_str ""
    print_members_status
    
    # Verify that Enough disk space is available on all members 

    set result 0
    set not_enough_space 0
    
    foreach member_id [array names member_output_array] {
        
        # Extract disk space from execution output
        set free_space [df_extract_disk_space $member_output_array($member_id)]
        putsd "[fn](): free_space=$free_space"
        # puts [ printLine [info frame] ]
        
        if {$free_space == -1} {
            set_member_status $member_id "Disk space verification has failed"
            writeLog $ping(LOG_ERROR)    "[fn]" "Failed to parse result:\n[array get member_output_array]"  0
            set result -1
        }
        
        # Compare partition free space against package size
        if {[expr $free_space>$required_free_space_bytes]} {
            set_member_status $member_id "Disk space is available"
        } else {
            # Not enough disk space
            set_member_status $member_id "Not enough disk space"
            set not_enough_space 1
            set result -1
        }
    }
    
    print_members_status
    
    if {$not_enough_space} {
        writeLog $ping(LOG_ERROR) "[fn]" "Not enough disk space in $ping(IMPORT_PATH) partition"
    }
    
    if {$result < 0} {
        writeLog $ping(LOG_ERROR) "[fn]" "Disk space verification has failed."
    }
    
    return $result
}

# Description: Extract the disk space from Linux "df" command output
#
# Input:
#        exec_output: First line is df header, second line is the Filesystem which the disk space is extracted from (Command example: # df -B1 -P /var/log)
#                                                                                                                                                                -P        use the POSIX output format
#                                                                                                                                                                -B1    block size will be 1 byte
# Return: 
#        disk_space_in_bytes    - In case parsing succeeded
#        -1                                - In case parsing failed
#
proc df_extract_disk_space {exec_output} {
    putsd "[fn]()"
    
    putsd "[fn](): exec_output=$exec_output"
    # Split execution output into lines
    set lines    [split $exec_output "\n"]
    
    # Verify header + row existence
    if {[llength $lines]<2} {
        putsd "[fn](): insufficient lines"
        return -1
    }
    
    # Get the second line
    set disk_space_line     [lindex $lines 1]
    
    # Replace multiple spaces with single space 
    regsub -all {\s+} $disk_space_line " " disk_space_line
     
    # Split into words
    set disk_space_words    [split $disk_space_line " "]
    
    # (df columns are : Filesystem , Size , Used ,Available , ...)
    if {[llength $disk_space_words]<4} {
        putsd "[fn](): df columns"
        return -1
    }
     
    # Get the 4th word into free_space (Available)
    set free_space     [lindex $disk_space_words 3]    
    putsd "[fn](): free_space=$free_space"
    
    
    # Verify that free_space is a number
    if {![string is double -strict $free_space]} {
        putsd "[fn](): Verify that free_space is a number"
        return -1
    }
     
    return $free_space
}

proc clean_exec_output {execution_output} {
	set c_output $execution_output ;# Cleaned output
 
	# Replace "" character [CLISH_TAG_VALUE_SEPERATOR] with " "
	regsub -all "\x80" $c_output " "  c_output
	
	# Remove <Delete> character [CLISH_LINE_SEPERATOR]
	regsub -all "\x7F" $c_output "\n" c_output

	# Remove all lines after we see 3 consecutive newlines because of duplicate output.
	regsub -all "\n\n\n.*" $c_output "\n" c_output
	
	# Return result
	return $c_output
}

# Execute DAClient with <cli_args> as its parameters and clean/reformat it's output
# Each member execute this function locally and return cleaned output.
# (In case we don't use this 'wrapper' then g_all return error because it is unable to handle these special characters)
proc DAClient_exec_and_clean {cli_args} {
    if {[catch {exec DAClient {*}$cli_args} output]} {
        return 1 ;# exec failed
    }
    puts [clean_exec_output $output]
    return 0
}
   
# Check if Script was executed directly from CLI
if { [info exists argv0] && [info script] eq $argv0 } {
    # Call function according to CLI args and Exit
    return [main $argv]
}
