#!/usr/bin/cgisu -r

source showResult.tcl
source json.tcl
source subs.tcl
source debug.tcl
source validate.tcl

namespace eval ::su_policy {
	global dFile
	set dFile "/tmp/installer_policy.post.debug"
	global err_list [list]
	global set_list [list]
	global result 0
	global myDb
	global method
	
	set bind_ping "installer"
	set trigger_file "/opt/CPda/bin/DAClient"

# getfails - run on error to report it to the browser so hopefully the
# user will find out about it.
# Syntax:
#	getfails errorcode errormsg additional_msg_text
proc getfails {ecode emsg addl_msg} {
    set hdr "Status: ${ecode} ${emsg}\r\n"
    append hdr "Content-Type: text/plain\r\n"
    append hdr "\r\n"
    puts -nonewline stdout $hdr

    set body "The file could not be retrieved.\n"

    if {$addl_msg != ""} {
	append body "Here's what we know about\n"
	append body "what went wrong:\n"
	append body "$addl_msg"
    }
    if {![string equal [get_Method] "HEAD"]} {
	# For GET we also print this "document body" to tell the user
	# more about the error (if the browser shows it to them); but
	# in a HEAD request we're not supposed to present a body.
	puts stdout $body
    }

    # The following puts the message in the error log.
    puts stderr "backup_restore_getter fails, status ${ecode} ${emsg}"
    if {$addl_msg != ""} {
	puts stderr "More detailed message: ${addl_msg}"
    }
    global env
    if {[info exists env(QUERY_STRING)]} {
	puts stderr "QUERY_STRING was: ${env(QUERY_STRING)}"
    }
    if {[info exists env(REQUEST_METHOD)]} {
	puts stderr "REQUEST_METHOD was: ${env(REQUEST_METHOD)}"
    }
    exit 0
}

# execute the given command and catch any errors
# args can be empty, however if more than one argument is passed -
# they will be merged into one argument! (this will probably not work)
# returns a list of 4 parameters: the command return code, the output,
# and in case of error - the error code and the stderr
proc safe_exec { cmd { args "" } } {
	set errorcode "no error code"
	set errorinfo "no error info"
	set rc [ catch { exec $cmd $args } output ]
	if { $rc != 0 } {
		set errorcode $::errorCode
		set errorinfo $::errorInfo
	}
	return [ list $rc $output $errorcode $errorinfo ]
}
	
# initialize the database handle
if {[catch {libdb init -local} myDb]} {
	#failed to initialize!!!
	lappend err_list "Unable to connect to database ${myDb}"
	exit
}
ipso -nohtml

# get what method is being used to invoke this script.
set method [get_Method]
set mime_type "application/octet-stream"

# **** Debug section ****
#	printDebug

##################### POST #############################
	if {$method == "POST"} {
		# set MIME type text inside POST function and not outside because of save file feature that uses MIME stream type
		MIME text
		set save 0

		global Q_Names
		set d_stype ""
		set d_weekday ""
		set d_monthday ""
		set d_month ""
		set d_year ""
		set d_minutes ""
		set d_hours ""
	
		set current_check_for_updates_period [libdb get $myDb "${bind_ping}:check_for_updates_period"] 
		set current_check_for_updates_status [libdb get $myDb "${bind_ping}:check_for_updates_status"]
		set update_period_changed "false"
		set disable_update "false"
		
		foreach var $Q_Names {
			set val [getVal $var]
			
			if { $var == "legacy_display" } {
				libdb set $myDb "${bind_ping}:legacy_display" $val
			}
			
			if { $var == "force_maj" } {
				libdb set $myDb "${bind_ping}:force_major_upgrade" $val
			}

			if { $var == "auto_download" } {
				if { $val == "false" } {
					libdb set $myDb "${bind_ping}:self_update_policy_no_permission" "1"
				} else { 
					libdb set $myDb "${bind_ping}:self_update_policy_no_permission" ""
				}
			}

			
			if { $var == "sanity_proc" } {
				libdb set $myDb "${bind_ping}:sanity_test_processes" $val
			}
			
			if { $var == "sanity_install_policy" } {
				libdb set $myDb "${bind_ping}:sanity_install_policy" $val
			}
			
			if { $var == "sanity_network_access" } {
				libdb set $myDb "${bind_ping}:sanity_network_access" $val
			}
			
			if { $var == "check_for_updates_period" } {
				if { $val != $current_check_for_updates_period } {
					set update_period_changed "true"
					if { $val == "0" } {
						set disable_update "true"
					}
					libdb set $myDb "${bind_ping}:check_for_updates_period" $val
				}
			}
				
			if { $var == "check_for_updates_status" } {
				if { $val != $current_check_for_updates_status } {
					libdb set $myDb "${bind_ping}:check_for_updates_status" $val
				}
			}
			
			if { $var == "d_radio" } {
				if { $val != "schedule" } {
					exec $trigger_file set_download_configuration $val
				}
				libdb set $myDb "${bind_ping}:downloadmode" $val
				libdb set $myDb "${bind_ping}:d_stype" $val
		}
			# download scheduling
			if { $var == "d_stype" } {
				set d_stype $val
			}
			if { $var == "d_weekdays" } {
				set d_weekday $val
			}
			if { $var == "d_monthdays" } {
				set d_monthday $val
			}
			if { $var == "d_months" } {
				set d_month $val
			}
			if { $var == "d_year" } {
				set d_year $val
			}
			if { $var == "d_minutes" } {
				set d_minutes $val
			}
			if { $var == "d_hours" } {
				set d_hours $val
			}
		}
		
		if { $d_stype == "daily" } {
			exec $trigger_file set_download_configuration schedule daily $d_hours $d_minutes
			libdb set $myDb "${bind_ping}:d_minutes" $d_minutes
			libdb set $myDb "${bind_ping}:d_hours" $d_hours
		}
		if { $d_stype == "weekly" } {
			exec $trigger_file set_download_configuration schedule weekly $d_weekdays $d_hours $d_minutes
			libdb set $myDb "${bind_ping}:d_minutes" $d_minutes
			libdb set $myDb "${bind_ping}:d_hours" $d_hours
			libdb set $myDb "${bind_ping}:d_weekday" $d_weekdays
		}
		if { $d_stype == "monthly" } {
			exec $trigger_file set_download_configuration schedule monthly $d_monthdays $d_hours $d_minutes
			libdb set $myDb "${bind_ping}:d_minutes" $d_minutes
			libdb set $myDb "${bind_ping}:d_hours" $d_hours
			libdb set $myDb "${bind_ping}:d_monthday" $d_monthdays
		}		
		if { $d_stype == "singular" } {
			exec $trigger_file set_download_configuration schedule singular $d_year $d_months $d_monthdays $d_hours $d_minutes
			libdb set $myDb "${bind_ping}:d_minutes" $d_minutes
			libdb set $myDb "${bind_ping}:d_hours" $d_hours
			libdb set $myDb "${bind_ping}:d_year" $d_year
			libdb set $myDb "${bind_ping}:d_month" $d_months
			libdb set $myDb "${bind_ping}:d_monthday" $d_monthdays
		}
		
		set policy_results [ safe_exec $trigger_file new_policy ]
		set response "policy_results: $policy_results"

		if { $update_period_changed == "true" && $disable_update == "false" } {
			set update_results [ safe_exec $trigger_file update ]
			append response "\nupdate_results: $update_results"
		}

		libdb set $myDb ":save" ""
		
		HTML $response

	} else {
##################### GET #############################
		set pkg [getVal package]
		set type [getVal type]
		if { $pkg != "" } {
			if { $pkg == "FullEventLog" } {
				set path "/opt/CPInstLog/DA_UI.log"
			} else {
				if { $type != "" } {
					set path [libdb get $myDb "installer:packages:${pkg}:${type}_file"]
				} else {
					getfails 400 "Missing package type (install/uninstall)"
				}
			}

			#set result 0

			if {![file exists $path]} {
				getfails 404 "Not found" \
				"The file ${path} doesn't exist"
			}

			if {![file isfile $path]} {
				getfails 400 "Bad request" \
				"The file ${path} is not a regular file"
			}

			if {[file size $path] < 1} {
				getfails 500 "Internal Error" \
				"The file ${path} is empty"
			}

			# Now we can open the file and prepare to dump it.
			if {[catch {open $path r} fh]} {
				set em "Failed to open file ${pkg} for reading; the system claimed\n"
				append em "the error was: ${fh}"
				getfails 500 "Internal Error" $em
			}

			if {[catch {
				fconfigure $fh -encoding binary -translation binary
				fconfigure stdout -encoding binary -translation binary
			} res]} {
				set em "some mysterious error happened while we were preparing to"
				append em " treat the file as binary: $res"
				getfails 500 "Internal Error" $em
			}

			# The content type tells the web browser what to do with the file.
			# The 'content-disposition' header is nonstandard in HTTP but widely used.
			# It's defined as a MIME header in RFC2183, and in RFC2616 (HTTP/1.1) its
			# use is discussed.  The idea is that we want to tell the user agent what
			# the file is named, since what it can get from the URL is really the name
			# of this Tcl script.  Some user agents might not recognize this header field
			# in which case they'll hopefully just ignore it (and leave it to the user
			# to choose a useful name for the file).
			set basename [file tail $path]
			set hdr "Content-type: ${mime_type}\r\n"
			append hdr "Content-Disposition: attachment;"
			append hdr " filename=\"$basename\"\r\n"
			append hdr "\r\n"
			puts -nonewline stdout $hdr

			if {[string equal [get_Method] "HEAD"]} {
				# I don't know if this should be called but it seems to be.  HEAD requests
				# have no use for the file body.  So we're done at this point.
				exit 0
			}

			# Now we read the file and dump it.
			if {[catch {
				while {![eof $fh]} {
				set buf [read $fh 8192]
				puts -nonewline stdout $buf
				}
				close $fh
			} res]} {
				# There's no good way to tell the user what went wrong at this point;
				# we just write this for debugging purposes (it'll go in httpd_error_log).
				puts stderr "Something went wrong in reading/writing the file: $res"
			}
		} else {
			# set MIME type text inside the else part of GET function and not outside because of "save file" feature that uses MIME type stream
			MIME text
			set downloadmode [libdb get $myDb "${bind_ping}:downloadmode"]
		
			set check_for_updates_period [libdb get $myDb "${bind_ping}:check_for_updates_period"] 
			set check_for_updates_status [libdb get $myDb "${bind_ping}:check_for_updates_status"] 
			
			# Setting default value for check_for_updates_period (default value is 6 hours)
			if { $check_for_updates_period == "" } {
				set check_for_updates_period 21600
			}
			
			set legacy_display [libdb get $myDb "${bind_ping}:legacy_display"]
			if { $legacy_display == "" } {
				# Setting default value for the legacy_display binding
				libdb set $myDb "${bind_ping}:legacy_display" "true"
				set legacy_display 1
			}
			set auto_download [libdb get $myDb "${bind_ping}:self_update_policy_no_permission"]
			if { $auto_download == "" } {
				set auto_download true
		} else {
			set auto_download false
		}


		set force_maj [libdb get $myDb "${bind_ping}:force_major_upgrade"]
		if { $force_maj == "" } {
			# Setting default value for the force_major_upgrade binding
			libdb set $myDb "${bind_ping}:force_major_upgrade" "false"
		}

		set sanity_auto_rollback [libdb get $myDb "${bind_ping}:sanity_auto_rollback"]
		if { $sanity_auto_rollback == "" } {
			# Setting default value for the sanity_test_processes binding
			libdb set $myDb "${bind_ping}:sanity_auto_rollback" "false"
			set sanity_auto_rollback 0
		}
		
		set sanity_proc [libdb get $myDb "${bind_ping}:sanity_test_processes"]
		if { $sanity_proc == "" } {
			# Setting default value for the sanity_test_processes binding
			libdb set $myDb "${bind_ping}:sanity_test_processes" "true"
			set sanity_proc 1
		}

		set sanity_install_policy [libdb get $myDb "${bind_ping}:sanity_install_policy"]
		set sanity_network_access [libdb get $myDb "${bind_ping}:sanity_network_access"]

		dict set SUPolicyDict success true
		set dataDict [dict create \
			d_radio $downloadmode \
			check_for_updates_period $check_for_updates_period \
			check_for_updates_status $check_for_updates_status \
			force_maj $force_maj \
			legacy_display $legacy_display\
			auto_download $auto_download\
			sanity_auto_rollback $sanity_auto_rollback\
			sanity_proc $sanity_proc\
			sanity_install_policy $sanity_install_policy\
			sanity_network_access $sanity_network_access]
			
		if {$downloadmode == "schedule"} {
			set hideBrand [libdb get $myDb "${bind_ping}:hide_brand"]
			set d_stype [libdb get $myDb "${bind_ping}:d_stype"]
			dict set dataDict hideBrand $hideBrand
			dict set dataDict d_stype $d_stype
			dict set dataDict d_minutes [libdb get $myDb "${bind_ping}:d_minutes"]
			dict set dataDict d_hours [libdb get $myDb "${bind_ping}:d_hours"]
			if {$d_stype == "weekly"} {
				dict set dataDict d_weekdays [libdb get $myDb "${bind_ping}:d_weekday"]
			}
			if {$d_stype == "monthly"} {
				dict set dataDict d_monthdays [libdb get $myDb "${bind_ping}:d_monthday"]
			}
			if {$d_stype == "singular"} {
				dict set dataDict d_monthdays [libdb get $myDb "${bind_ping}:d_monthday"]
				dict set dataDict d_year [libdb get $myDb "${bind_ping}:d_year"]
				dict set dataDict d_months [libdb get $myDb "${bind_ping}:d_month"]
			}
		}	
			dict set SUPolicyDict data $dataDict
			HTML [::dict2json::compile_json {dict data {dict} } $SUPolicyDict]
		}
	}

	libdb close $myDb
};#namespace

exit 0
