#!/bin/bash

BACKUP_DIR=/opt/CPda/backup
CPPROD_UTIL=${CPDIR}/bin/cpprod_util
CKP_REGEDIT=${CPDIR}/bin/ckp_regedit
DEBUG_LOG=/opt/CPInstLog/mds_restored.log

function initIfFailes(){
	[[ $2 == "Failed to find the value" ]] && eval "$1="
}

function deleteKeyFromReg(){
	echo "------deleting $1 $2 $3 if exists" >> $DEBUG_LOG
        # The ckp_regedit command returns its rc as output (e.g. rc=-11) so we have to eval the output
	eval $(${CKP_REGEDIT} -d $1 $2 $3 2>&1)
	if [[ ${rc} -eq 0 ]]; then
		FOUND_INCONS=1
		echo "------the key was deleted successfully" >> $DEBUG_LOG
	elif [[ ${rc} -eq -10 ]]; then
		echo "------the key does not exist" >> $DEBUG_LOG
	elif [[ ${rc} -eq -11 ]]; then
		echo "------the path for the key doesn't exist at all" >> $DEBUG_LOG
	else
		echo "------failed to delete the key. RC=${rc}" >> $DEBUG_LOG
	fi
}

function defineRegistryValues() {
    PROD=$1
    HF_ID=$2
    CUR_VER=$(${CPPROD_UTIL} CPPROD_GetCurrentVersion "${PROD}")
    	# trim trailing space
	CUR_VER=$(echo ${CUR_VER} | xargs)
	echo "----reading rollback packages from registry path: ${PROD}//${CUR_VER}//${HF_ID}" >> $DEBUG_LOG
    PING_PACK=$(${CPPROD_UTIL} CPPROD_GetValue "${PROD}"//${CUR_VER}//${HF_ID} ping_package 0);
	if [[ ${PING_PACK} != "1" ]]; then
		echo "-----This is not a CPUSE package" >> $DEBUG_LOG
		return 1
	fi
	
    RB=$(${CPPROD_UTIL} CPPROD_GetValue "${PROD}"//${CUR_VER}//${HF_ID} RollbackPackage 0)
    COMP_RB=$(${CPPROD_UTIL} CPPROD_GetValue "${PROD}"//${CUR_VER}//${HF_ID} RollbackPackageCompletely 0)
    PREV_RB=$(${CPPROD_UTIL} CPPROD_GetValue "${PROD}"//${CUR_VER}//${HF_ID}//PreviousInstalled RollbackPackage 0)
	
	initIfFailes RB "${RB}"
	initIfFailes COMP_RB "${COMP_RB}"
	initIfFailes PREV_RB "${PREV_RB}"
	
	echo "------RB=${RB}" >> $DEBUG_LOG
	echo "------COMP_RB=${COMP_RB}" >> $DEBUG_LOG
	echo "------PREV_RB=${PREV_RB}" >> $DEBUG_LOG
	return 0
}


function updateRegistry() {

	echo "----updating the registry" >> $DEBUG_LOG

	echo "------setting value of \"${PROD}\"//${CUR_VER}//${HF_ID} RollbackPackage=${RB}" >> $DEBUG_LOG
	${CPPROD_UTIL} CPPROD_SetValue "${PROD}"//${CUR_VER}//${HF_ID} RollbackPackage 1 ${RB} 0
	if [[ "X${PREV_RB}" != "X" && -f ${PREV_RB} && "X${COMP_RB}" != "X" && -f ${COMP_RB} ]]; then 
		echo "------setting value of \"${PROD}\"//${CUR_VER}//${HF_ID} RollbackPackageCompletely=${COMP_RB}" >> $DEBUG_LOG
		${CPPROD_UTIL} CPPROD_SetValue "${PROD}"//${CUR_VER}//${HF_ID} RollbackPackageCompletely 1 ${COMP_RB} 0
		echo "------setting value of \"${PROD}\"//${CUR_VER}//${HF_ID}//PreviousInstalled RollbackPackage=${COMP_RB}" >> $DEBUG_LOG
		${CPPROD_UTIL} CPPROD_SetValue "${PROD}"//${CUR_VER}//${HF_ID}//PreviousInstalled RollbackPackage 1 ${PREV_RB} 0
	else
		echo "------deleting obsolete keys" >> $DEBUG_LOG
		deleteKeyFromReg //SOFTWARE//CheckPoint//"${PROD}"//${CUR_VER}//${HF_ID} RollbackPackageCompletely
		deleteKeyFromReg //SOFTWARE//CheckPoint//"${PROD}"//${CUR_VER}//${HF_ID}//PreviousInstalled
	fi
}

function handleOneProd() {
    local PROD=$1
	echo "handling product: $PROD" >> $DEBUG_LOG
    for HF in $(${CPPROD_UTIL} CPPROD_GetHotFixes "${PROD}") ; do
		if [[ ${HF} == "Can" ]]; then
			echo "--No hotfixes installed for this product" >> $DEBUG_LOG
			return
		fi
		echo "--handling hotfix: $HF" >> $DEBUG_LOG
		if defineRegistryValues "${PROD}" ${HF}; then 
			if [[ "X${COMP_RB}" == "X" && "X${PREV_RB}" == "X" || -f ${COMP_RB} ]] && [[ "X${PREV_RB}" == "X" || -f ${PREV_RB} ]]  && [[ -f ${RB} ]]; then
				echo "------all files exist :-)" >> $DEBUG_LOG
			else
				updateRegistry
			fi
		fi
    done
}

#backup registry
cp -a $CPDIR/registry/HKLM_registry.data $CPDIR/registry/HKLM_registry.data.rollback_fix
pIFS=$IFS
IFS=$'\n'

echo "********* Starting MDS restored script ******" > $DEBUG_LOG
for PROD_NAME in $(${CPPROD_UTIL} CPPROD_GetInstalledProducts); do
	IFS=$pIFS
	handleOneProd "${PROD_NAME}"
done;

if [[ ! ${FOUND_INCONS} ]]; then
	#remove backup if nothing was changed
	rm $CPDIR/registry/HKLM_registry.data.rollback_fix
	echo "******  No inconsitency was found  ******" >> $DEBUG_LOG
fi

