#!/bin/bash -x

######################################################################
#
# This script is called after the installation of the files was
# completed.
#
######################################################################

BASE_CPDIR=$1;
CHROOT_TARGET_CPDIR=$2;
BASE_FWDIR=$3;
MNT_DIR=$6;
TARGET_CPDIR=$MNT_DIR/$2;
VAR_TARGET_CPDIR=$MNT_DIR/var/$2
TARGET_FWDIR=$MNT_DIR/$4;
VAR_TARGET_FWDIR=$MNT_DIR/var/$4;
VER=$5;
VSID=$7

XINSTALL=/opt/CPsuite-$VER/CPinstall/XInstall
TARGET_FW_BOOT_DIR=$MNT_DIR/var/opt/fw.boot
SVN_UP_50_SO_File=$VAR_TARGET_CPDIR/conf/upgrade_50.conf
FW1_UP_50_SO_File=$TARGET_FWDIR/scripts/upgrade_50.conf
FW_BOOT_DIR_COREXL=/var/opt/fw.boot
BASE_BOOT_CONF="$BASE_FWDIR/boot/boot.conf"    
DEBUG=/opt/CPInstLog/install_fw1_$VER.elg
BASE_CPPROD=$BASE_CPDIR/bin/cpprod_util
TARGET_CPPROD=$TARGET_CPDIR/bin/cpprod_util

#path to current appliance_config.xml
CUR_APPLIANCE_CONFIG=/etc/appliance_config.xml
#flag indicating whether upgrade will activate hyper-threading. 0=There is hyper-threading. 1=No hyper-threading. 2=hyper-threading not defined
NOHT_FLAG="2"

#special error number for CPUSE
ERROR_XINSTALL=17

$BASE_CPPROD CPPROD_GetValue "PROVIDER-1" MDSDIR 1
if [[ $? == 0 ]]; then
	IS_MDS=1
else
	IS_MDS=0
fi

IS_GW=$($BASE_CPPROD FwIsFirewallModule 2>> $DEBUG)
IS_MGMT=$($BASE_CPPROD FwIsFirewallMgmt 2>> $DEBUG)
IS_VSX=$($BASE_CPPROD FwIsVSX 2>> $DEBUG)
IS_VS0=0

AWK=/bin/awk
CAT=/bin/cat
CD=cd
CHOWN=/bin/chown
CHGRP=/bin/chgrp
CHMOD=/bin/chmod
CHROOT=/usr/sbin/chroot
CP=/bin/cp
ECHO=/bin/echo
EX=/bin/ex
EXPR=/usr/bin/expr
FIND=/usr/bin/find
GREP=/bin/grep
KILL=/bin/kill
LN=/bin/ln
LS=/bin/ls
MKDIR=/bin/mkdir
MV=/bin/mv
PS=/bin/ps
PWD=/bin/pwd
RM=/bin/rm
SED=/bin/sed
TAR=/bin/tar
TEE=/usr/bin/tee
TOUCH=/bin/touch
READLINK=/usr/bin/readlink
XARGS=/usr/bin/xargs
XMLLINT=/usr/bin/xmllint
RPM2CPIO=/usr/bin/rpm2cpio
CPIO=/bin/cpio

VSX_CHROOT_EXT=""
if [[ $IS_VSX -eq 1 ]]; then
	echo "Current VS id is $VSID" >> $DEBUG

	if [ $VSID -ne 0 ]; then
		VSX_CHROOT_EXT="source /etc/profile.d/vsenv.sh > /dev/null ; export UPGRADE_CONTEXT=1; vsenv ${VSID} ;"
		$VSX_CHROOT_EXT
		# Updating debug file
		DEBUG=/opt/CPInstLog/install_fw1_${VER}_VS${VSID}.elg
	else
		IS_VS0=1
		source /tmp/.CPprofile.sh > /dev/null
	fi
else
	. /opt/CPshared/5.0/tmp/.CPprofile.sh > /dev/null
fi

# command (single var)
function run_chroot() {
	$CHROOT $MNT_DIR /bin/bash -c ". /tmp/.CPprofile.sh ; ${VSX_CHROOT_EXT} $1"  1>>$DEBUG 2>&1
}

function run_chroot_stdout() {
	$CHROOT $MNT_DIR /bin/bash -c ". /tmp/.CPprofile.sh ; ${VSX_CHROOT_EXT} $1" |tee -a $DEBUG 2>>$DEBUG
}

function debug() {
	$ECHO "$1" >> $DEBUG
}

function isNotMainVS() {
	if [[ $IS_VSX -eq 1 && $IS_VS0 -ne 1 ]]; then
		debug "Operation is for VS0 only"
		return 0
	fi
	return 1
}

function handleNewVsSymlinkPre() {
	debug "Start handleNewVsSymlinkPre"
	
	if [ ! -d "$2/$1" ]; then
		debug "Directory $2/$1 doesn't exist, skip handling files for this path"
		debug "End handleNewVsSymlinkPre"
		return
	fi
	
	ARCHIVE_NAME=vs_${VSID}_symlink_${1}.tar
	SYMLINKS_LIST=vs_${VSID}_${1}_links_for_tar

	$RM $SYMLINKS_LIST
	$RM $ARCHIVE_NAME

	$FIND $2/$1/ -type l >> $SYMLINKS_LIST 

	debug "Creating TAR file"
	$TAR -Pcvf $ARCHIVE_NAME -T $SYMLINKS_LIST 1>>$DEBUG  2>&1  

	debug "*************************************************************"
	debug "The following links are backed up from $2/$1"
	$CAT $SYMLINKS_LIST >> $DEBUG 
	debug "*************************************************************"

	$CAT $SYMLINKS_LIST | $XARGS -I{} $RM "{}" 1>>$DEBUG  2>&1  

	$RM $SYMLINKS_LIST 1>>$DEBUG  2>&1  
	debug "List of files after tar shrink and clean in $2/$1"
	ls -lh $2/$1/ 1>>$DEBUG  2>&1  

	debug "End handleNewVsSymlinkPre"
}

function handleNewVsSymlinkPost() {
	debug "Start handleNewVsSymlinkPost"

	ARCHIVE_NAME=vs_${VSID}_symlink_${1}.tar
	
	if [ ! -e "$ARCHIVE_NAME" ]; then
		debug "${ARCHIVE_NAME} file doesn't exist, skip handling files for this path"
		debug "End handleNewVsSymlinkPost"
		return;
	fi

	debug "list of files that will be removed"
	$TAR -tvf $ARCHIVE_NAME >> $DEBUG

	debug "Remove all file/dir that should be re-written by symlinks"
	$TAR -tf $ARCHIVE_NAME | $XARGS -I{} $RM -rf "{}" 1>>$DEBUG  2>&1  

	debug "Deploy TAR file"
	$TAR -Pxvf $ARCHIVE_NAME 1>>$DEBUG  2>&1  
	$RM $ARCHIVE_NAME 1>>$DEBUG  2>&1  

	debug "End handleNewVsSymlinkPost"
}

function handleOldVsSymlinkPre() {
	debug "Start handleOldVsSymlinkPre"
	
	if [ ! -d "$2/$1" ]; then
		debug "Directory $2/$1 doesn't exist, skip handling files for this path"
		debug "End handleOldVsSymlinkPre"
		return
	fi
	
	BROKEN_ARCHIVE_NAME=vs_${VSID}_broken_symlink_${1}.tar
	BROKEN_SYMLINKS_LIST=vs_${VSID}_${1}_broken_links_for_tar

	$RM $BROKEN_SYMLINKS_LIST
	$RM $BROKEN_ARCHIVE_NAME

	$FIND $2/$1/ -type l -xtype l >> $BROKEN_SYMLINKS_LIST 

	debug "Creating TAR file"
	$TAR -Pcvf $BROKEN_ARCHIVE_NAME -T $BROKEN_SYMLINKS_LIST 1>>$DEBUG  2>&1  

	debug "*************************************************************"
	debug "The following links are backed up from $2/$1"
	$CAT $BROKEN_SYMLINKS_LIST >> $DEBUG 
	debug "*************************************************************"

	$CAT $BROKEN_SYMLINKS_LIST | $XARGS -I{} $RM "{}" 1>>$DEBUG  2>&1  

	$RM $BROKEN_SYMLINKS_LIST 1>>$DEBUG  2>&1  
	debug "List of files broken links $2/$1"
	ls -lh $2/$1/ 1>>$DEBUG  2>&1  

	debug "End handleOldVsSymlinkPre"
}

function handleOldVsSymlinkPost() {
	debug "Start handleOldVsSymlinkPost"

	BROKEN_ARCHIVE_NAME=vs_${VSID}_broken_symlink_${1}.tar
	
	if [ ! -e "$BROKEN_ARCHIVE_NAME" ]; then
		debug "${BROKEN_ARCHIVE_NAME} file doesn't exist, skip handling files for this path"
		debug "End handleOldVsSymlinkPost"
		return;
	fi

	debug "list of broken links that will return to their place"
	$TAR -tvf $BROKEN_ARCHIVE_NAME >> $DEBUG

	debug "Deploy TAR file"
	$TAR -Pxvf $BROKEN_ARCHIVE_NAME 1>>$DEBUG  2>&1  
	$RM $BROKEN_ARCHIVE_NAME 1>>$DEBUG  2>&1  

	debug "End handleOldVsSymlinkPost"
}

# handleVsSymlinks <pre/post> BASE_DIR TARGET_DIR
function handleVsSymlinks() {
	BASE_DIR=$2
	TARGET_DIR=$3
	directories=( "state" "conf" "database" )
	if [[ $IS_VSX -eq 0 || $IS_VS0 -eq 1 ]]; then
		return
	fi
	debug "handleNewVsSymlinks was called with $1 parameter in $2 path"

	for i in "${directories[@]}"
	do
		if [ "X$1" = "Xpre" ]; then
			handleOldVsSymlinkPre "$i" "$BASE_DIR"
			handleNewVsSymlinkPre "$i" "$TARGET_DIR"
		fi
		if [ "X$1" = "Xpost" ]; then
			handleNewVsSymlinkPost "$i" "$BASE_DIR"
			handleOldVsSymlinkPost "$i" "$TARGET_DIR"
		fi
	done
}

function upgradeFiles() {
	debug "UpgradeFiles $SVN_UP_50_SO_File and $FW1_UP_50_SO_File"
	export TDERROR_XInstall_LOGPATH=/opt/CPInstLog/XInstall.log
	export TDERROR_XInstall_ALL=5
	[[ $IS_VSX -eq 1 && $IS_VS0 -eq 0 ]] && handleVsSymlinks pre $BASE_CPDIR $VAR_TARGET_CPDIR
	$MNT_DIR/$XINSTALL UpgradeFiles $BASE_CPDIR $VAR_TARGET_CPDIR $SVN_UP_50_SO_File  1>>$DEBUG 2>&1
	if [[ $? -ne 0 ]]; then
		debug "ERROR while running: $MNT_DIR/$XINSTALL UpgradeFiles $BASE_CPDIR $VAR_TARGET_CPDIR $SVN_UP_50_SO_File"
		return $ERROR_XINSTALL
	fi
	[[ $IS_VSX -eq 1 && $IS_VS0 -eq 0 ]] && handleVsSymlinks post $BASE_CPDIR $VAR_TARGET_CPDIR

	[[ $IS_VSX -eq 1 && $IS_VS0 -eq 0 ]] && handleVsSymlinks pre $BASE_FWDIR $VAR_TARGET_FWDIR
	$MNT_DIR/$XINSTALL UpgradeFiles $BASE_FWDIR $VAR_TARGET_FWDIR $FW1_UP_50_SO_File 1>>$DEBUG 2>&1
	if [[ $? -ne 0 ]]; then
		debug "ERROR while running: $MNT_DIR/$XINSTALL UpgradeFiles $BASE_FWDIR $VAR_TARGET_FWDIR $FW1_UP_50_SO_File"
		return $ERROR_XINSTALL
	fi
	[[ $IS_VSX -eq 1 && $IS_VS0 -eq 0 ]] && handleVsSymlinks post $BASE_FWDIR $VAR_TARGET_FWDIR
}

function registerFw1Components() {
	debug "Register fw1 components from $TARGET_FWDIR/conf/components_reg.conf"
	run_chroot "$XINSTALL CPRegSvr fw1 \"\$FWDIR/conf/components_reg.conf\"" 1>>$DEBUG 2>&1
	return $?
}

function copyPolicyFiles() {
	debug "Copy policy files"
	isNotMainVS && return 0
	#lib
	$CP $BASE_FWDIR/lib/user.def $TARGET_FWDIR/lib 1>>$DEBUG 2>&1
	if [ -f $BASE_FWDIR/lib/user_early.def ]; then
		$CP $BASE_FWDIR/lib/user_early.def $TARGET_FWDIR/lib 1>>$DEBUG 2>&1
	fi

	if [ -f $TARGET_FWDIR/lib/defaultfilter.boot ]; then
		$MKDIR -p $VAR_TARGET_FWDIR/conf
		$CP $TARGET_FWDIR/lib/defaultfilter.boot $VAR_TARGET_FWDIR/conf/defaultfilter.pf
	fi
}

function copyAuthKeys() {
	$CP $BASE_CPDIR/database/*authkeys.C $TARGET_CPDIR/database 1>>$DEBUG  2>&1    
}


function copySshKeys() {
	debug "Copy ssh keys"
	$MKDIR -p $MNT_DIR/etc/ssh
	
	SSH_FILES="ssh_host_key ssh_host_key.pub ssh_host_rsa_key ssh_host_rsa_key.pub ssh_host_dsa_key ssh_host_dsa_key.pub"
	for file in $SSH_FILES
	do
		if [ -f /etc/ssh/$file ]; then
		$CP /etc/ssh/$file $MNT_DIR/etc/ssh/$file 1>>$DEBUG 2>&1
		fi
	done
}

function copyHaBootConf() {
	debug "Copy ha_boot.conf"

	isNotMainVS && return 0

	#ha
	VER_R77="R77"
	CLUSTER_ID_STRING="cluster_id"
	BASE_VER=`$READLINK -f ${BASE_CPPROD} | cut -d"-" -f2 | cut -d"/" -f1`

	$CP $BASE_FWDIR/boot/ha_boot.conf $TARGET_FW_BOOT_DIR 1>>$DEBUG 2>&1
	#cluster_id feature exist in ha_boot file only for R77.30
	if [ "$BASE_VER" == "$VER_R77" ]; then
		$SED -i '/'"$CLUSTER_ID_STRING"'/d' $TARGET_FW_BOOT_DIR/ha_boot.conf
	fi
	#MVC mode shouldn't be copied - MVC should act as designed by default in each version
	$SED -i '/^mvc \{1,2\}o[nf]f\{0,1\}/d' $TARGET_FW_BOOT_DIR/ha_boot.conf
	$SED -i '/^cphamvc \{1,2\}o[nf]f\{0,1\}/d' $TARGET_FW_BOOT_DIR/ha_boot.conf
	#CCP Encryption mode shouldn't be copied - should act as designed by default in each version
	$SED -i '/^ccp_encrypt \{1,2\}o[nf]f\{0,1\}/d' $TARGET_FW_BOOT_DIR/ha_boot.conf
}

function copyClusterxlMonitorScript() {
	debug "Copy clusterxl_monitor scripts"

	isNotMainVS && return 0
	
	if [ -f /etc/rc.d/rc.local ]; then
		debug "checking for custom scripts"
		$SED -n '/# start - run the/,/# end - run the/p' /etc/rc.d/rc.local >> $MNT_DIR/etc/rc.d/rc.local
	fi
}

function generateDefaultFilter() {
	debug "Generate default filter"
	isNotMainVS && return 0

	run_chroot "\$FWDIR/bin/fw defaultgen" 1>>$DEBUG 2>&1

	$RM -rf $VAR_TARGET_FWDIR/state/local 1>>$DEBUG 2>&1

	run_chroot "\$FWDIR/bin/comp_init_policy" 1>>$DEBUG 2>&1
	return 0
}

function setDefaultFilter() {
	debug "Set default filter"
	isNotMainVS && return 0

	if [ -f $VAR_TARGET_FWDIR/state/default.bin ]; then
		debug "Copying default.bin"
		$MV $VAR_TARGET_FWDIR/state/default.bin $TARGET_FW_BOOT_DIR/default.bin 1>>$DEBUG 2>&1
	fi

	if [ -f $VAR_TARGET_FWDIR/state/default.bin6 ]; then
		debug "Copying default.bin6"
		$MV $VAR_TARGET_FWDIR/state/default.bin6 $TARGET_FW_BOOT_DIR/default.bin6 1>>$DEBUG 2>&1
	fi

	if [ -f $TARGET_FW_BOOT_DIR/fwboot ] && [ -f $TARGET_FW_BOOT_DIR/default.bin ]; then
		debug "Running fwboot bootconf setdef"
		run_chroot "\$FW_BOOT_DIR/fwboot bootconf set_def \$FW_BOOT_DIR/default.bin" 1>>$DEBUG 2>&1
		return $?
	else
		debug "Error setting default filter"
		return -1
	fi
	return 0
}

function addAmonConfiguration() {
	debug "Add AMON configuration"
	isNotMainVS && return 0

	run_chroot "\$CPDIR/bin/amon_tool -f \$FWDIR/conf/amon_fw.conf -a" 1>>$DEBUG 2>&1
	return $?
}

function addProductLibraries() {
	debug "Add product libraries"
	run_chroot "\$CPDIR/bin/cptfmt_config add dll FireWall-1 Connections \$FWDIR/lib libfwtfmt.so"  1>>$DEBUG 2>&1
	run_chroot "\$CPDIR/bin/cptfmt_config add dll VPN-1 VPN \$FWDIR/lib libvpntfmt.so" 1>>$DEBUG 2>&1
}

function setBootSec() {
	debug "Run setbootsec"
	isNotMainVS && return 0
	IS_DAG=$($BASE_CPPROD FwIsDag)
	if [ -f $TARGET_FWDIR/scripts/setbootsec ] ; then
		run_chroot "aix=0 solaris2=0 is_dag=${IS_DAG} hpux=0 linux=1 \$FWDIR/scripts/setbootsec" 1>>$DEBUG 2>&1
		return $?
	fi
}

function fwauthdConfUpdate() {
	debug "Add stormd to fwauthd.conf"
	if [ -f $TARGET_FWDIR/conf/fwauthd.conf ]; then
		$SED -i "/\(sds\|dtls\|dtps\|stormd\)/d" $TARGET_FWDIR/conf/fwauthd.conf
	fi

	$ECHO "0        stormd          stormd          respawn         0" >> $TARGET_FWDIR/conf/fwauthd.conf
	$ECHO "0	sds		sdsd		respawn		0" >> $TARGET_FWDIR/conf/fwauthd.conf
	$ECHO "0	dtps		dtpsd		respawn		0" >> $TARGET_FWDIR/conf/fwauthd.conf
	$ECHO "0	dtls		dtlsd		respawn		0" >> $TARGET_FWDIR/conf/fwauthd.conf

	run_chroot "\$FWDIR/bin/fw authd_set -b -q fwauthd.conf" 1>>$DEBUG 2>&1
}

function upgradePersistency() {
	debug "Upgrade persistency"
	PrevVER=$($BASE_CPPROD CPPROD_GetCurrentVersion CPshared 2>>$DEBUG)
	PrevVER=${PrevVER:-0}
	PSP=$($BASE_CPPROD CPPROD_GetCurrentServicePack CPshared 2>>$DEBUG)
	PSP=${PSP:-0}
	PrevMSP=$($BASE_CPPROD CPPROD_GetCurrentMSP FW1 2>>$DEBUG)
	PrevMSP=${PrevMSP:-0}

	# TODO run chrooted?
	$MNT_DIR/$XINSTALL ExecuteFile "$TARGET_CPDIR/bin/upgrade_persistency $PrevVER $PSP $PrevMSP" 1>>$DEBUG 2>&1
	return $?
}

function copySimAffinityConfig() {
	debug "Copy SIM Affinity configuration"
	isNotMainVS && return 0
	if [ -f $PPK_BOOT_DIR/boot/modules/sim_aff.conf ] ; then
		$CP $PPK_BOOT_DIR/boot/modules/sim_aff.conf $MNT_DIR/$PPK_BOOT_DIR/boot/modules 1>>$DEBUG 2>&1
		[ $? -eq 0 ] && $ECHO "sim_aff.conf copied successfully..." >> $DEBUG
	fi
}

function copyFwKernConf() {
	debug "Copy fwkern.conf"
	isNotMainVS && return 0
	if [ -f /etc/fw.boot/modules/fwkern.conf ] ; then
		$CP /etc/fw.boot/modules/fwkern.conf  $TARGET_FW_BOOT_DIR/modules/fwkern.conf  1>>$DEBUG 2>&1
		[ $? -eq 0 ] && $ECHO "fwkern.conf copied successfully..." >> $DEBUG
	fi
}

function copySnmpdConf() {
        debug "Copy snmpd.conf"
        isNotMainVS && return 0
        if [ -f /var/lib/net-snmp/snmpd.conf ] ; then
                $CP /var/lib/net-snmp/snmpd.conf $MNT_DIR/var/lib/net-snmp/  1>>$DEBUG 2>&1
                [ $? -eq 0 ] && $ECHO "snmpd.conf copied successfully..." >> $DEBUG
        fi
}


function copySnmpmonitorConf() {
        debug "Copy snmpmonitor.conf"
        isNotMainVS && return 0
        if [ -f /etc/snmp/snmpmonitor.conf ] ; then
                $CP /etc/snmp/snmpmonitor.conf  $MNT_DIR/etc/snmp/  1>>$DEBUG 2>&1
                [ $? -eq 0 ] && $ECHO "snmpmonitor.conf copied successfully..." >> $DEBUG
        fi
}

function copyDynamicObjectsDB() {
		debug "Copy dynamic_objects.db"		
		if [ -f $FWDIR/database/dynamic_objects.db ]; then
			$CP $FWDIR/database/dynamic_objects.db $VAR_TARGET_FWDIR/database/ 1>>$DEBUG 2>&1
			[ $? -eq 0 ] && $ECHO "dynamic_objects.db copied successfully..." >> $DEBUG
		fi
}

function copyScratchpad () {
        debug "Copy scratchpad "
        isNotMainVS && return 0
        if [ -f /etc/scratchpad ] ; then
                $CP /etc/scratchpad $MNT_DIR/etc/  1>>$DEBUG 2>&1
                [ $? -eq 0 ] && $ECHO "scratchpad copied successfully..." >> $DEBUG
        fi
}

function copyAdjtime () {
        debug "Copy adjtime"
        isNotMainVS && return 0
        if [ -f /etc/adjtime  ] ; then
                $CP /etc/adjtime  $MNT_DIR/etc/adjtime   1>>$DEBUG 2>&1
                [ $? -eq 0 ] && $ECHO "adjtime copied successfully..." >> $DEBUG
        fi
}

function copyIoportConf() {
	debug "Copy /etc/ioport.conf"
	isNotMainVS && return 0
	if [ -f /etc/ioport.conf ] ; then
		$CP /etc/ioport.conf  $MNT_DIR/etc/ioport.conf  1>>$DEBUG 2>&1
		[ $? -eq 0 ] && $ECHO "ioport.conf copied successfully..." >> $DEBUG
	fi
}

function copySecureidFiles()
{
	if isNotMainVS
	then
		SecureIdFilesLocation=$FWDIR/conf
	else
		SecureIdFilesLocation=/var/ace
		if [ -d /var/ace ]
		then
		$MKDIR -p $MNT_DIR/var/ace
		else
			return 0
		fi
	fi

	for file in securid sdstatus.12 sdopts.rec sdconf.rec
	do	
		fullPath=$SecureIdFilesLocation/$file
		debug "Copy $fullPath"
		if [ -f $fullPath ]
		then
			$CP $fullPath $MNT_DIR/$fullPath 1>>$DEBUG 2>&1 && $ECHO "$fullPath copied successfully..." >> $DEBUG
	fi
	done
}

function copyModprobeConf() {
	debug "Copy /etc/modprobe.conf"
	isNotMainVS && return 0
	if [ -f /etc/modprobe.conf ] ; then
		$CP /etc/modprobe.conf  $MNT_DIR/etc/modprobe.conf  1>>$DEBUG 2>&1
		[ $? -eq 0 ] && $ECHO "modprobe.conf copied successfully..." >> $DEBUG
	fi
}

function sofaWareScript() {
	debug "Run SofaWare sw_post_install script"
	if [[ $IS_VSX -eq 1 ]]; then
		debug "Not running sw_post_install in VSX"
		return 0
	fi
	run_chroot "$XINSTALL ExecuteFile \"\$FWDIR/bin/sw_post_install\"" 1>>$DEBUG 2>&1
	return $?
}

function copyNCS() {
	debug "Copy directory $BASE_FWDIR/state/local/VSX to $VAR_TARGET_FWDIR/state/local"
	if [ -d $BASE_FWDIR/state/local/VSX ] && [ -d $VAR_TARGET_FWDIR/state/local ] ; then
		$CP -r $BASE_FWDIR/state/local/VSX $VAR_TARGET_FWDIR/state/local 1>>$DEBUG 2>&1
		if [ $? -eq 0 ] ; then
			$ECHO "Directory $BASE_FWDIR/state/local/VSX copied successfully..." >> $DEBUG
		else
			$ECHO "Failed to copy directory $BASE_FWDIR/state/local/VSX to $VAR_TARGET_FWDIR/state/local" >> $DEBUG
		fi
	else
		$ECHO "ERROR! At least one of the directories don't exist" >> $DEBUG
	fi
}

function generatePolicy() {
	copyPolicyFiles
	copyAuthKeys
	generateDefaultFilter
	setDefaultFilter
	copyNCS
}

function is_large_scale_supported() {
	test -f $FWDIR/boot/is_large_scale_supported
}

function is_large_scale_supported_chroot() {
	run_chroot "test -f \$FWDIR/boot/is_large_scale_supported"
	return $?
}

function executeCoreXL() {
	$FWDIR/boot/fwboot $1
	return_value=$?
	if  ! is_large_scale_supported  ; then
		echo $return_value
	fi
}

function executeChrootCoreXL() {
	if  is_large_scale_supported_chroot ; then
		run_chroot_stdout "\$FWDIR/boot/fwboot corexl $1" |tee -a $DEBUG 2>>$DEBUG 
	else
		run_chroot "\$FWDIR/boot/fwboot corexl $1" 1>>$DEBUG 2>&1
		echo $?
	fi
}

function enableCoreXL() {
	HT_PRE_INSTALL_STATUS=$($CAT /proc/smt_status)
	#In case HT is enabled then NOHT_POST_INSTALL_STATUS will be '0'. The reason is because it is the noht value in grub.cons
	NOHT_FLAG="2"
	setNohtFlag
	NOHT_POST_INSTALL_STATUS=$NOHT_FLAG
	if [ "X$NOHT_POST_INSTALL_STATUS" = "X0" ]; then
		$FWDIR/boot/fwboot cpuid -c
		NUM_OF_CORES_BEFORE_UPGRADE=$($ECHO $?)
		NEW_NUM_OF_CORES_AFTER_REBOOT=$(($NUM_OF_CORES_BEFORE_UPGRADE * 2))
		debug "Hyper threadind is enabled post upgrade"
		debug "NUM_OF_CORES_BEFORE_UPGRADE is $NUM_OF_CORES_BEFORE_UPGRADE"
		debug "NEW_NUM_OF_CORES_AFTER_REBOOT is $NEW_NUM_OF_CORES_AFTER_REBOOT"
		run_chroot "\$FWDIR/boot/fwboot bootconf set_core_override $NEW_NUM_OF_CORES_AFTER_REBOOT" 1>>$DEBUG 2>&1
	fi

	KERNEL_VERSION=`cat $MNT_DIR/kernel_version.txt`
	ORIG_STATIC_AFFINITY=$($BASE_CPPROD FwIsFwdStaticAffinity)
	debug "fwd static affinity: $ORIG_STATIC_AFFINITY"
	NEW_STATIC_AFFINITY=$ORIG_STATIC_AFFINITY
	debug "KERNEL_VERSION: $KERNEL_VERSION"
	if [ X$KERNEL_VERSION = "X4.18.0" ]; then
		NEW_STATIC_AFFINITY=0
	fi
	debug "NEW_STATIC_AFFINITY $NEW_STATIC_AFFINITY"
	run_chroot "\$CPDIR/bin/cpprod_util FwSetFwdStaticAffinity $NEW_STATIC_AFFINITY" 1>>$DEBUG 2>&1

	ORIG_KERN_INSTANCE_NUM=$($GREP KERN_INSTANCE_NUM $BASE_BOOT_CONF | $AWK '{print $2}')
	ORIG_DEFAULT_KERN_INSTANCE_NUM=$(executeCoreXL "corexl def_instance4_count")	
	ORIG_KERN6_INSTANCE_NUM=$($GREP KERN6_INSTANCE_NUM $BASE_BOOT_CONF | $AWK '{print $2}')
	ORIG_DEFAULT_KERN6_INSTANCE_NUM=$(executeCoreXL "corexl def_instance6_count")
	CPU_ALLOWED_BY_LIC_STR=$(fw ctl get int fwlic_num_of_allowed_cpus)
	debug "CPU_ALLOWED_BY_LIC_STR: $CPU_ALLOWED_BY_LIC_STR"
	read -r NAME COMP CPU_ALLOWED_BY_LIC <<< $CPU_ALLOWED_BY_LIC_STR
	debug "the num cpu allowed by license is $CPU_ALLOWED_BY_LIC"
	#This check is needed when we have an unlimited cpu license. 100000 or higher is used to signal an unlimited cpu license
	REGEX_DIGIT='^[0-9]+$'
	if ! [[ $CPU_ALLOWED_BY_LIC =~ $REGEX_DIGIT ]] ; then
		debug "bad CPU_ALLOWED_BY_LIC given!!"
		NEW_DEF_KERN_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance4_count")
		NEW_DEF_KERN6_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance6_count")
	else
		if [ $CPU_ALLOWED_BY_LIC -ge  100000 ]; then
			NEW_DEF_KERN_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance4_count")
			NEW_DEF_KERN6_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance6_count")
		else
			NEW_DEF_KERN_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance4_count_lic $CPU_ALLOWED_BY_LIC")
			NEW_DEF_KERN6_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance6_count_lic $CPU_ALLOWED_BY_LIC")
		fi
	fi
	debug "NEW_DEF_KERN_INSTANCE_NUM_DEF $NEW_DEF_KERN_INSTANCE_NUM_DEF"
	debug "NEW_DEF_KERN6_INSTANCE_NUM_DEF $NEW_DEF_KERN6_INSTANCE_NUM_DEF"
	
	debug "ORIG_KERN_INSTANCE_NUM is $ORIG_KERN_INSTANCE_NUM"
	debug "HT_PRE_INSTALL_STATUS is $HT_PRE_INSTALL_STATUS"
	debug "ORIG_DEFAULT_KERN_INSTANCE_NUM is $ORIG_DEFAULT_KERN_INSTANCE_NUM"
	

	
	if [ $ORIG_KERN_INSTANCE_NUM -eq $ORIG_DEFAULT_KERN_INSTANCE_NUM ]; then
		if [ "X$HT_PRE_INSTALL_STATUS" = "XSoft Disable" ] && [ "X$NOHT_POST_INSTALL_STATUS" = "X0" ] ; then
			NEW_KERN_INSTANCE_NUM=$(($NEW_DEF_KERN_INSTANCE_NUM_DEF * 2))
		else
			NEW_KERN_INSTANCE_NUM=$NEW_DEF_KERN_INSTANCE_NUM_DEF
		fi
	else
		if [ "X$HT_PRE_INSTALL_STATUS" = "XSoft Disable" ] && [ "X$NOHT_POST_INSTALL_STATUS" = "X0" ] ; then
			NEW_KERN_INSTANCE_NUM=$(($ORIG_KERN_INSTANCE_NUM * 2))
		else
			NEW_KERN_INSTANCE_NUM=$ORIG_KERN_INSTANCE_NUM
		fi
	fi
	debug "NEW_KERN_INSTANCE_NUM is $NEW_KERN_INSTANCE_NUM"
	
	if [[ -n $ORIG_KERN6_INSTANCE_NUM ]] ; then
		debug "ORIG_KERN6_INSTANCE_NUM is $ORIG_KERN6_INSTANCE_NUM"
		debug "ORIG_DEFAULT_KERN6_INSTANCE_NUM is $ORIG_DEFAULT_KERN6_INSTANCE_NUM"
		if [ $ORIG_KERN6_INSTANCE_NUM -eq $ORIG_DEFAULT_KERN6_INSTANCE_NUM ]; then
			if [ "X$HT_PRE_INSTALL_STATUS" = "XSoft Disable" ] && [ "X$NOHT_POST_INSTALL_STATUS" = "X0" ] ; then
				NEW_KERN6_INSTANCE_NUM=$(($NEW_DEF_KERN6_INSTANCE_NUM_DEF * 2))
			else
				NEW_KERN6_INSTANCE_NUM=$NEW_DEF_KERN6_INSTANCE_NUM_DEF
			fi
		else
			if [ "X$HT_PRE_INSTALL_STATUS" = "XSoft Disable" ] && [ "X$NOHT_POST_INSTALL_STATUS" = "X0" ] ; then
				NEW_KERN6_INSTANCE_NUM=$(($ORIG_KERN6_INSTANCE_NUM * 2))
			else
				NEW_KERN6_INSTANCE_NUM=$ORIG_KERN6_INSTANCE_NUM
			fi
		fi
	fi
	
	debug "NEW_KERN6_INSTANCE_NUM is $NEW_KERN6_INSTANCE_NUM"
	run_chroot "\$FWDIR/boot/fwboot corexl enable $NEW_KERN_INSTANCE_NUM" 1>>$DEBUG 2>&1
	if [ "X$NEW_KERN6_INSTANCE_NUM" != "X" ]; then
		ORIG_IPV6_STATE=$($FWDIR/boot/fwboot bootconf get_ipv6)
		debug "ORIG_IPV6_STATE is $ORIG_IPV6_STATE"
		run_chroot "\$FWDIR/boot/fwboot bootconf set_ipv6 $ORIG_IPV6_STATE" 1>>$DEBUG 2>&1
		run_chroot "\$FWDIR/boot/fwboot corexl enable -6 $NEW_KERN6_INSTANCE_NUM" 1>>$DEBUG 2>&1
	fi
	
	return 0
}

function enableCoreXLSP() {
	HT_PRE_INSTALL_STATUS=$($CAT /proc/smt_status)
	#In case HT is enabled then NOHT_POST_INSTALL_STATUS will be '0'. The reason is because it is the noht value in grub.cons
	NOHT_FLAG="2"
	setNohtFlag
	NOHT_POST_INSTALL_STATUS=$NOHT_FLAG
	if [ "X$NOHT_POST_INSTALL_STATUS" = "X0" ]; then
		$FWDIR/boot/fwboot cpuid -c
		NUM_OF_CORES_BEFORE_UPGRADE=$($ECHO $?)
		NEW_NUM_OF_CORES_AFTER_REBOOT=$(($NUM_OF_CORES_BEFORE_UPGRADE * 2))
		debug "Hyper threadind is enabled post upgrade"
		debug "NUM_OF_CORES_BEFORE_UPGRADE is $NUM_OF_CORES_BEFORE_UPGRADE"
		debug "NEW_NUM_OF_CORES_AFTER_REBOOT is $NEW_NUM_OF_CORES_AFTER_REBOOT"
		run_chroot "\$FWDIR/boot/fwboot bootconf set_core_override $NEW_NUM_OF_CORES_AFTER_REBOOT" 1>>$DEBUG 2>&1
	fi

	ORIG_STATIC_AFFINITY=$($BASE_CPPROD FwIsFwdStaticAffinity)
	debug "fwd static affinity: $ORIG_STATIC_AFFINITY"
	run_chroot "\$CPDIR/bin/cpprod_util FwSetFwdStaticAffinity $ORIG_STATIC_AFFINITY" 1>>$DEBUG 2>&1

	ORIG_KERN_INSTANCE_NUM=$($GREP KERN_INSTANCE_NUM $BASE_BOOT_CONF | $AWK '{print $2}')
	ORIG_DEFAULT_KERN_INSTANCE_NUM=$(executeCoreXL "corexl def_instance4_count")
	ORIG_KERN6_INSTANCE_NUM=$($GREP KERN6_INSTANCE_NUM $BASE_BOOT_CONF | $AWK '{print $2}')
	ORIG_DEFAULT_KERN6_INSTANCE_NUM=$(executeCoreXL "corexl def_instance6_count")
	CPU_ALLOWED_BY_LIC_STR=$(fw ctl get int fwlic_num_of_allowed_cpus)
	debug "CPU_ALLOWED_BY_LIC_STR: $CPU_ALLOWED_BY_LIC_STR"
	read -r NAME COMP CPU_ALLOWED_BY_LIC <<< $CPU_ALLOWED_BY_LIC_STR
	debug "the num cpu allowed by license is $CPU_ALLOWED_BY_LIC"
	#This check is needed when we have an unlimited cpu license. 100000 or higher is used to signal an unlimited cpu license
	REGEX_DIGIT='^[0-9]+$'
	if ! [[ $CPU_ALLOWED_BY_LIC =~ $REGEX_DIGIT ]] ; then
		debug "bad CPU_ALLOWED_BY_LIC given!!"
		NEW_DEF_KERN_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance4_count")
		NEW_DEF_KERN6_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance6_count")
	else
		if [ $CPU_ALLOWED_BY_LIC -ge  100000 ]; then
			NEW_DEF_KERN_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance4_count")
			NEW_DEF_KERN6_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance6_count")
		else
			NEW_DEF_KERN_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance4_count_lic $CPU_ALLOWED_BY_LIC")
			NEW_DEF_KERN6_INSTANCE_NUM_DEF=$(executeChrootCoreXL "def_instance6_count_lic $CPU_ALLOWED_BY_LIC")
		fi
	fi
	debug "NEW_DEF_KERN_INSTANCE_NUM_DEF $NEW_DEF_KERN_INSTANCE_NUM_DEF"
	debug "NEW_DEF_KERN6_INSTANCE_NUM_DEF $NEW_DEF_KERN6_INSTANCE_NUM_DEF"

	debug "ORIG_KERN_INSTANCE_NUM is $ORIG_KERN_INSTANCE_NUM"
	debug "HT_PRE_INSTALL_STATUS is $HT_PRE_INSTALL_STATUS"
	debug "ORIG_DEFAULT_KERN_INSTANCE_NUM is $ORIG_DEFAULT_KERN_INSTANCE_NUM"




	NEW_KERN_INSTANCE_NUM=$ORIG_KERN_INSTANCE_NUM
	debug "NEW_KERN_INSTANCE_NUM is $NEW_KERN_INSTANCE_NUM"

	debug "ORIG_KERN6_INSTANCE_NUM is $ORIG_KERN6_INSTANCE_NUM"
	debug "ORIG_DEFAULT_KERN6_INSTANCE_NUM is $ORIG_DEFAULT_KERN6_INSTANCE_NUM"
	NEW_KERN6_INSTANCE_NUM=$ORIG_KERN6_INSTANCE_NUM

	debug "NEW_KERN6_INSTANCE_NUM is $NEW_KERN6_INSTANCE_NUM"
	run_chroot "\$FWDIR/boot/fwboot corexl enable $NEW_KERN_INSTANCE_NUM" 1>>$DEBUG 2>&1
	if [ "X$NEW_KERN6_INSTANCE_NUM" != "X" ]; then
		ORIG_IPV6_STATE=$($FWDIR/boot/fwboot bootconf get_ipv6)
		debug "ORIG_IPV6_STATE is $ORIG_IPV6_STATE"
		run_chroot "\$FWDIR/boot/fwboot bootconf set_ipv6 $ORIG_IPV6_STATE" 1>>$DEBUG 2>&1
		run_chroot "\$FWDIR/boot/fwboot corexl enable -6 $NEW_KERN6_INSTANCE_NUM" 1>>$DEBUG 2>&1
	fi

	return 0
}

function disableCoreXL() {
	run_chroot "$FW_BOOT_DIR_COREXL/fwboot corexl disable"
	return 0
}

function setCoreXL() {
	debug "Setting CoreXL configuration"
	isNotMainVS && return 0

	IS_COREXL_INSTALLED=$($GREP COREXL_INSTALLED $BASE_BOOT_CONF | $AWK '{print $2}')
	is_scalable_platform=`/usr/bin/is_scalable_platform.sh`
	debug "IS_COREXL_INSTALLED = $IS_COREXL_INSTALLED"
	debug "is_scalable_platform = $is_scalable_platform"

	# the file doesn't contain the keyword, which means CoreXL was disabled.
	if [ "X$IS_COREXL_INSTALLED" = "X" ]; then
		debug "Can't find IS_COREXL_INSTALLED keyword, which means CoreXL was disabled"
		return 0
	fi

	# Configure CoreXL like it was before upgrade
	if [ $IS_COREXL_INSTALLED -eq 1 ] ; then
		debug "Enabling CoreXL"
		if [ "$is_scalable_platform" == "1" ] ; then
			enableCoreXLSP
		else
			enableCoreXL
		fi
	else
		debug "Disabling CoreXL"
		disableCoreXL
	fi
	return 0
}

function copyFwaffinityConf(){
	
	debug "copy fwaffinity.conf"
	
	if [ -f $BASE_FWDIR/conf/fwaffinity.conf ]; then
		$CP $BASE_FWDIR/conf/fwaffinity.conf $TARGET_FWDIR/conf/fwaffinity.conf  1>>$DEBUG 2>&1
	fi
}
function setMQ() {
	debug "Setting /sys link"
	for f in /sys/class/net/* ; do
		mkdir -p "$MNT_DIR/$f/device"
		$CP "$f/ifindex" "$MNT_DIR/$f/"
		$CP "$f/iflink" "$MNT_DIR/$f/"
		$LN -s `readlink $f/device/driver` "$MNT_DIR/$f/device/driver"
	done

	debug "Setting Multi Queue configuration"
	$CP $BASE_FWDIR/boot/mq.conf $TARGET_FW_BOOT_DIR/mq.conf 1>>$DEBUG 2>&1
	run_chroot "\$FWDIR/bin/cpmq reconfigure" 1>>$DEBUG 2>&1
	$RM $MNTDIR/sys/ -fR
}

function setCtxdbConfig() {
	debug "Modify ctxdb.C file"

	if [ $IS_VSX -ne 1 ]; then
		debug "Operation is relevant for VSX Only"
		return 0
	fi
	
	isNotMainVS && return 0
	
	CTXDB_FILE=$VAR_TARGET_CPDIR/conf/ctxdb.C
	debug "Modify the following CTXDB file ${CTXDB_FILE}"
	$SED -i.bak '/ipaddr/d' $CTXDB_FILE
	debug "${CTXDB_FILE} was modified"
}


function setSecureXL(){
	
	isNotMainVS && return 0
	debug "Configuring state of SecureXL"
	BaseSecureXLOn=$($BASE_CPPROD FwIsSecureXL) 1>>$DEBUG 2>&1
			
	run_chroot "\$CPDIR/bin/cpprod_util FwSetSecureXL $BaseSecureXLOn" 1>>$DEBUG 2>&1
		
}

function setIpV6(){
	debug "Configuring Ipv6"
	IPV6STATE=$($BASE_CPPROD FwIsFireWallIPv6) 1>>$DEBUG 2>&1

	run_chroot "\$CPDIR/bin/cpprod_util FwSetFireWallIPv6 $IPV6STATE" 1>>$DEBUG 2>&1
}

function updateCphaprobConf()
{
	debug "Modify cphaprob.conf file, remove routed 0 problem"

    if [ -f $VAR_TARGET_FWDIR/conf/cphaprob.conf ]; then
        $SED '/routed 0 problem/d' -i $VAR_TARGET_FWDIR/conf/cphaprob.conf
    fi
}

function copySAM(){

	debug "============ SAM Start ==============="
	if [ $IS_VSX -eq 1 ]; then
		debug "SAM is not supported in VSX, returning"
		return 0
	fi
	
	#create log directory at target dir
	if [ ! -d /var/log/opt/CPsuite-$VER/fw1/log/ ]; then
		debug "create log directory at target dir"
		mkdir -p "/var/log/opt/CPsuite-$VER/fw1/log/"
		RC=$?
		debug "RC=$RC"
	fi
	
	#copy SAM rules
	if [ -f $BASE_FWDIR/log/sam.dat ]; then
		debug "copy SAM file to log directory at new version"
		$CP -v $BASE_FWDIR/log/sam.dat /var/log/opt/CPsuite-$VER/fw1/log/	1>>$DEBUG 2>&1
		RC=$?
		debug "RC=$RC"
	fi

	#SAM reformat
	if [ -f $TARGET_FWDIR/bin/sam_reformat ]; then
		debug "running sam_reformat in target directory $TARGET_FWDIR/bin/sam_reformat"
		$TARGET_FWDIR/bin/sam_reformat -u 1>>$DEBUG 2>&1
		RC=$?
		debug "RC=$RC"
	else
		debug "running sam_reformat in base directory $BASE_FWDIR/bin/sam_reformat"
		$BASE_FWDIR/bin/sam_reformat -u 1>>$DEBUG 2>&1
		RC=$?
		debug "RC=$RC"
	fi
	
	
	#copy tmp file
	if [ $RC -eq 0 ]; then
		debug "sam_reformat ended with success, going to copy the temp file"
		$CP -v $BASE_FWDIR/log/sam.dat.tmp /var/log/opt/CPsuite-$VER/fw1/log/sam.dat	1>>$DEBUG 2>&1
		RC=$?
		debug "RC=$RC"
		$RM $BASE_FWDIR/log/sam.dat.tmp 1>>$DEBUG 2>&1
	else
		if [ -f $BASE_FWDIR/log/sam.dat.tmp ]; then
			debug "reformatting is not needed, going to remove temp file"
			$RM $BASE_FWDIR/log/sam.dat.tmp 1>>$DEBUG 2>&1
			RC=$?
			debug "RC=$RC"
		fi
	fi
	
	debug "============ SAM End ==============="

}

function handleNohtFlagPerAppliance()
{
    case $1 in
        P-370-00|P-230-00)
        # 13500|12600 appliances
            NOHT_FLAG="0" ;;

        *)
        # All other cases
            NOHT_FLAG="1" ;;
    esac
    $ECHO "noht flag set for appliance type $1 to $NOHT_FLAG" 1>>$DEBUG 2>&1
}

function setNohtFlag()
{
    debug "============ setNohtFlag Start ==============="
    cur_path=`$PWD`
    
    #get current appliance type from current appliance_config.xml
    cur_machine_type=`$ECHO "cat /model/@type" | $XMLLINT --shell $CUR_APPLIANCE_CONFIG | $GREP type | $AWK -F'"' '{print $2}'`
    if [ ! -d /tmp/rpm ]
    then
        $MKDIR /tmp/rpm
    else
        # remove old content
        $RM -rf /tmp/rpm/* 2>/dev/null 
    fi
    
    #extract appliance_configuration.xml from new partition
    $CP $MNT_DIR/xmlUpgrade*.rpm /tmp/rpm 1>>$DEBUG 2>&1
    $CD /tmp/rpm/
    $RPM2CPIO xmlUpgrade*.rpm | $CPIO -idv 1>>$DEBUG 2>&1
    $CD tmp/
    if [ ! -f appliance_configuration.xml ]
    then
        $ECHO "appliance_configuration.xml file does not exist in xmlUpgrade!" 1>>$DEBUG 2>&1
        return
    fi

    #check no_hyper_thread tag for current appliance in new appliance_config.xml
    #if noht_flag is set to nothing after the command, there won't be hyper-threading and we set the flag to one
    NOHT_FLAG="`$ECHO -e "ls appliance_configuration/model[@type='$cur_machine_type']/cpuconf/no_hyper_thread" | $XMLLINT --shell appliance_configuration.xml | $AWK '{print $5}'`"
    $ECHO "noht in appliance_config.xml is: $NOHT_FLAG" 1>>$DEBUG 2>&1
    if [[ -z "$NOHT_FLAG" ]]
    then
        #if <no_hyper_thread> does not exist, NOHT_FLAG will be empty and needs to be set
        handleNohtFlagPerAppliance $cur_machine_type
    fi
    $CD $cur_path
}
function copyINextAgentFiles()
{
	debug "============ copyINextAgentFiles Start ==============="

	ORCH_CONF=/etc/cp/orchestration/cp-nano-orchestration.cfg
	if [ -f $ORCH_CONF ]
	then
	  BASE_GAIA_VERSION=$(echo $BASE_FWDIR | awk -F'CPsuite-|/fw1' '{print $2}')
	  sed -i 's,'"$BASE_GAIA_VERSION"','"$VER"',g' $ORCH_CONF 1>>$DEBUG 2>&1 \
		&& echo "$ORCH_CONF updated successfully..." >> $DEBUG
	fi

	if [ -d /etc/cp ]
	then
	  $CP -R /etc/cp $MNT_DIR/etc/ 1>>$DEBUG 2>&1 && echo "/etc/cp copied successfully..." >> $DEBUG
	fi

	if [ -f /usr/sbin/cpnano ]
	then
	  $CP /usr/sbin/cpnano $MNT_DIR/usr/sbin 1>>$DEBUG 2>&1 && echo "/usr/sbin/cpnano copied successfully..." >> $DEBUG
	fi

	if [ -d /usr/lib/cpnano ]
	then
	  $CP -R /usr/lib/cpnano $MNT_DIR/usr/lib/cpnano/ 1>>$DEBUG 2>&1 && echo "/usr/lib/cpnano copied successfully..." >> $DEBUG
	fi

	debug "============ copyINextAgentFiles End ==============="
}

function copySDWANFiles()
{
	debug "============ copySDWANFiles Start ==============="

	if [ -d $BASE_FWDIR/conf/sdwan ]
	then
	  $CP -R $BASE_FWDIR/conf/sdwan $VAR_TARGET_FWDIR/conf/sdwan/ 1>>$DEBUG 2>&1 && echo "\$BASE_FWDIR/conf/sdwan copied successfully..." >> $DEBUG
	fi

	if [ -L /var/$FWDIR/state/local/SDWAN ]
	then
	  REAL_TARGET_FWDIR=$(echo $VAR_TARGET_FWDIR | sed 's,'"$MNT_DIR"',,g')
	  BASE_SDWAN_POLICY_DIR=$(readlink $BASE_FWDIR/state/local/SDWAN)
	  TARGET_SDWAN_POLICY_DIR=$(echo $BASE_SDWAN_POLICY_DIR | sed 's,'"$BASE_FWDIR"','"$REAL_TARGET_FWDIR"',g')
	  $LN -sf $TARGET_SDWAN_POLICY_DIR $VAR_TARGET_FWDIR/state/local/SDWAN 1>>$DEBUG 2>&1 && echo "\$VAR_TARGET_FWDIR/state/local/SDWAN link created successfully" >> $DEBUG
	fi

	debug "============ copySDWANFiles End ==============="
}

SVN_UP_50_SO_File_RESOLVED=`$READLINK -f ${SVN_UP_50_SO_File}`
FW1_UP_50_SO_File_RESOLVED=`$READLINK -f ${FW1_UP_50_SO_File}`

if [ ! "${SVN_UP_50_SO_File_RESOLVED}" == "${SVN_UP_50_SO_File}" ]; then
		debug "Resolved upgrade_50.conf file, original path: ${SVN_UP_50_SO_File}, resolved path ${SVN_UP_50_SO_File_RESOLVED}"
        SVN_UP_50_SO_File=$SVN_UP_50_SO_File_RESOLVED
fi

if [ ! "${FW1_UP_50_SO_File_RESOLVED}" == "${FW1_UP_50_SO_File}" ]; then
		debug "Resolved upgrade_50.conf file, original path: ${FW1_UP_50_SO_File}, resolved path ${FW1_UP_50_SO_File_RESOLVED}"
        FW1_UP_50_SO_File=$FW1_UP_50_SO_File_RESOLVED
fi


# Will be executed for all types of machine
copySnmpdConf
copyScratchpad
copySshKeys
copyINextAgentFiles

if [[ $IS_GW -eq 0 ]]; then
	exit 0
fi

# gateway only actions
if [ $IS_MGMT -eq 0 ] && [ $IS_MDS -eq 0 ]; then
	upgradeFiles
	if [[ $? -eq $ERROR_XINSTALL ]]; then
		exit $ERROR_XINSTALL
	fi	
	setCtxdbConfig
	registerFw1Components
	upgradePersistency
	fwauthdConfUpdate
	sofaWareScript
	copyHaBootConf
else
	copyDynamicObjectsDB
fi

# gateway and standalone actions
generatePolicy
addAmonConfiguration
addProductLibraries
setBootSec
copySimAffinityConfig
copyFwKernConf
copyIoportConf
copySecureidFiles
copyModprobeConf
copyClusterxlMonitorScript
copySAM
setCoreXL
copyFwaffinityConf
setSecureXL
setIpV6
setMQ
copySnmpmonitorConf
copyAdjtime
updateCphaprobConf
copySDWANFiles

exit 0

